/*
 *  Long Long Integer Maths Library, for compilers not supporting long long ints.
 *  Written by Nigel Atkinson, SGS Thomson Microelectronics Limited.
 */

/* Copyright (C) SGS-THOMSON Microelectronics Ltd, 1996 */

#include <stdio.h>
#include <mathf.h>
#include <errno.h>
#include <ctype.h>
#include "llmath.h"
#include <limits.h>

/* Useful constant values */

const int64_t ll_one_const = { 1L, 0L };
const int64_t ll_zero_const = { 0L, 0L };
const int64_t ll_max_const = { ULONG_MAX, LONG_MAX };
const int64_t ll_min_const = { 0, LONG_MIN };

/* Long long integer math functions */

/*{{{  function prototypes */
static void LongSum (unsigned long *, unsigned long *, unsigned long, unsigned long, unsigned long);
static void LongAdd (unsigned long *, unsigned long, unsigned long, unsigned long);
static void LongDiff (unsigned long *, unsigned long *, unsigned long, unsigned long, unsigned long);
static void LongProd (unsigned long *, unsigned long *, unsigned long, unsigned long, unsigned long);
static void LongDiv (unsigned long *, unsigned long *, unsigned long, unsigned long, unsigned long);
extern void ShiftLeft (unsigned long *, unsigned long *, unsigned long, unsigned long, unsigned long);
extern void ShiftRight (unsigned long *, unsigned long *, unsigned long, unsigned long, unsigned long);
extern void AShiftLeft (signed long *, signed long *, signed long, signed long, unsigned long);
extern void AShiftRight (signed long *, signed long *, signed long, signed long, unsigned long);
extern void Normalise (unsigned long *, unsigned long *, unsigned long *, unsigned long, unsigned long);
static void double2ll_x (int *, unsigned long *, unsigned long *, int, unsigned long, unsigned long);
static void ll2double_x (unsigned long *, unsigned long *, int, unsigned long, unsigned long);
static void ll_div_rem  (int64_t *, int64_t *, int64_t, int64_t);
static void ull_div  (int64_t *, int64_t *, int64_t, int64_t);
static void ll2float_x (unsigned long *, int, unsigned long, unsigned long);
/*}}}*/

/*{{{  static functions */
/*{{{  LongSum */
static void
LongSum (unsigned long *CarryOut, unsigned long *Sum, unsigned long Left, unsigned long Right, unsigned long CarryIn)
{
  CarryIn &= 1;
   
  if ((MaxUnsignedInt32 - Left) >= Right){
    *Sum = Left + Right;
    if ((CarryIn != 0) && (*Sum == MaxUnsignedInt32)){
      *Sum = 0;
      *CarryOut = 1;
    }
    else{
      *Sum = *Sum + CarryIn;
      *CarryOut = 0;
    }
  }
  else{
    *CarryOut = 1;
    *Sum = (Right - ((MaxUnsignedInt32 - Left) + 1)) + CarryIn;
  }
}
/*}}}*/


/*{{{  LongAdd */
static void
LongAdd (unsigned long *Sum, unsigned long Left, unsigned long Right, unsigned long CarryIn)
{
  *Sum = Left + (Right + (CarryIn & 1));
}  
/*}}}*/


/*{{{  LongDiff */
void
LongDiff (unsigned long *BorrowOut, unsigned long *Diff, unsigned long Left, unsigned long Right, unsigned long BorrowIn)
{
  BorrowIn &= 1;
  *BorrowOut = 0;
   
  if (Left >= Right)
    *Diff = Left - Right;
  else{
    *Diff = MaxUnsignedInt32 - Right;
    *Diff = (*Diff + Left) + 1;
    *BorrowOut = 1;
  }
   
  if (BorrowIn != 0){
    if (*Diff > 0)
      (*Diff)--;
    else{
      *Diff = MaxUnsignedInt32;
      *BorrowOut = 1;
    }
  } 
}
/*}}}*/


/*{{{  LongProd */
static void
LongProd (unsigned long *ProdHi, unsigned long *ProdLo, unsigned long Left, unsigned long Right, unsigned long CarryIn)
{
  unsigned long Temp1, Temp2, Temp3, Temp4, TempLo, TempHi, Carry;
   
  Temp1 = Left & LowBits16;          /* Left MOD 2^16 */
  Temp2 = (Left >> 16) & LowBits16;  /* Left DIV 2^16 */
  Temp3 = Right & LowBits16;         /* Right MOD 2^16 */
  Temp4 = (Right >> 16) & LowBits16; /* Right DIV 2^16 */
   
  *ProdLo = Temp1 * Temp3;
  TempLo = Temp2 * Temp3;
  TempHi = (TempLo >> 16) & LowBits16;
  TempLo = TempLo << 16;

  LongSum(ProdHi, ProdLo, *ProdLo, TempLo, 0);
   
  *ProdHi = *ProdHi + TempHi;
  TempLo = Temp1 * Temp4;
  TempHi = (TempLo >> 16) & LowBits16;
  TempLo = TempLo << 16;
    
  LongSum(&Carry, ProdLo, *ProdLo, TempLo, 0);
    
  *ProdHi = (*ProdHi + Carry) + TempHi;
    
  LongSum(&Carry, ProdLo, *ProdLo, CarryIn, 0);
    
  *ProdHi = (*ProdHi + Carry) + (Temp2 * Temp4);
}   
/*}}}*/
    

/*{{{  LongDiv */
static void
LongDiv (unsigned long *Quot, unsigned long *Rem, unsigned long DividHi, unsigned long DividLo, unsigned long Divis)
{
  if (Divis == 0){
    *Quot = 0;
    *Rem = 0;
  }
  else if ((DividHi == 0) && (DividLo == 0)){
    *Quot = 0;
    *Rem = 0;
  }
  else{
    int BitCount;
    unsigned long Places, RemHi, RemLo, QuotHi, QuotLo;
   
    RemHi = 0;
    RemLo = 0; 
    QuotHi = DividHi;
    QuotLo = DividLo;
    
    Normalise(&Places, &QuotHi, &QuotLo, QuotHi, QuotLo);
    
    BitCount = 64 - ((int) Places);
    do{
      ShiftLeft(&RemHi, &RemLo, RemHi, RemLo, 1);
    
      if ((QuotHi & SignBit32) != 0)
        RemLo++;
 
      ShiftLeft(&QuotHi, &QuotLo, QuotHi, QuotLo, 1);

      if ((RemHi > 0) || ((RemHi == 0) && (RemLo >= Divis))){
        unsigned long Borrow;
   
        LongDiff(&Borrow, &RemLo, RemLo, Divis, 0);
        LongDiff(&Borrow, &RemHi, RemHi, 0, Borrow);
        QuotLo++;
      }
   
      BitCount--;
    } while (BitCount > 0);
    
    *Quot = QuotLo;  
    *Rem = RemLo;
  }
}   
/*}}}*/


/*{{{  ShiftLeft */
extern void
ShiftLeft (unsigned long *OutHi, unsigned long *OutLo, unsigned long InHi, unsigned long InLo, unsigned long Places)
{
  if (Places >= 64){
      *OutHi = 0;
      *OutLo = 0;
  }
  else if (Places >= 32){
      *OutLo = 0;
      *OutHi = InLo << (Places - 32);
  }
  else if (Places == 0){
      *OutLo = InLo;
      *OutHi = InHi;
  } 
  else{
      *OutHi = (InHi << Places) | (InLo >> (32 - Places));
      *OutLo = InLo << Places;
  }
}   
/*}}}*/


/*{{{  ShiftRight */
extern void
ShiftRight (unsigned long *OutHi, unsigned long *OutLo, unsigned long InHi, unsigned long InLo, unsigned long Places)
{
  if (Places >= 64){
    *OutHi = 0;
    *OutLo = 0;
  }
  else if (Places >= 32){
    *OutHi = 0;
    *OutLo = InHi >> (Places - 32);
  } 
  else if (Places == 0){
    *OutLo = InLo;
    *OutHi = InHi;
  }
  else{
    *OutLo = (InLo >> Places) | (InHi << (32 - Places));
    *OutHi = InHi >> Places;
  } 
}   
/*}}}*/


/*{{{  AShiftLeft */
extern void 
AShiftLeft (signed long *OutHi, signed long *OutLo, signed long InHi, signed long InLo, unsigned long Places)
{
  if (Places >= 64){
      *OutHi = 0;
      *OutLo = 0;
  }
  else if (Places >= 32){
      *OutLo = 0;
      *OutHi = InLo << (Places - 32);
  }
  else if (Places == 0){
      *OutLo = InLo;
      *OutHi = InHi;
  }
  else{
      *OutHi = (InHi << Places) | (InLo >> (32 - Places));
      *OutLo = InLo << Places;
  }
}
/*}}}*/


/*{{{  AShiftRight */
extern void 
AShiftRight (signed long *OutHi, signed long *OutLo, signed long InHi, signed long InLo, unsigned long Places)
{
  if (Places >= 64){
    *OutHi = 0;
    *OutLo = 0;
  }
  else if (Places >= 32){
    *OutHi = 0;
    *OutLo = InHi >> (Places - 32);
  }
  else if (Places == 0){
    *OutLo = InLo;
    *OutHi = InHi;
  }
  else{
    *OutLo = (InLo >> Places) | (InHi << (32 - Places));
    *OutHi = InHi >> Places;
  }
}
/*}}}*/


/*{{{  Normalise */
extern void
Normalise (unsigned long *Places, unsigned long *OutHi, unsigned long *OutLo, unsigned long InHi, unsigned long InLo)
{  
  if ((InHi == 0) && (InLo == 0)){
    *Places = 64;
    *OutHi = 0;
    *OutLo = 0;
  } 
  else{
    *Places = 0;
    *OutHi = InHi;
    *OutLo = InLo;
    
    while ((*OutHi & SignBit32) == 0){
      ShiftLeft(OutHi, OutLo, *OutHi, *OutLo, 1);
      (*Places)++;
    }
  } 
}   
/*}}}*/
/*}}}*/

/********************** USER ROUTINES ************************/

/*{{{  integer arithmetic */
/*{{{  integer addition */
int64_t
ll_add (int64_t A,int64_t B)
{   
  int64_t Result;
  unsigned long Carry;
 
  LongSum ((unsigned long *)&Carry, (unsigned long *)&(Result.lslong), A.lslong, B.lslong, 0);
  LongAdd ((unsigned long *)&(Result.mslong), A.mslong, B.mslong, Carry);
    
  return Result;
}
/*}}}*/

/*{{{  integer subtraction */
int64_t
ll_sub (int64_t A, int64_t B)
{  
  int64_t Result;
  unsigned long Borrow;
   
  LongDiff ((unsigned long *) &Borrow, (unsigned long *) &(Result.lslong), A.lslong, B.lslong, 0);
  LongDiff ((unsigned long *) &Borrow, (unsigned long *) &(Result.mslong), A.mslong, B.mslong, Borrow);
   
  return Result;
}
/*}}}*/

/*{{{  unsigned integer division */
static void
ull_div (int64_t *Quot, int64_t *Rem, int64_t A, int64_t B)
{   
  if ((B.lslong == 0) && (B.mslong == 0))
    /* divide by 0 error */
    printf ("Divide by zero!!!\n");
  else{
    unsigned long Count, X, U[3], W[3], V[2];
    if (B.mslong == 0){
      /* double by single */ 
      Rem->mslong = 0;
      LongDiv((unsigned long *)&(Quot->mslong), (unsigned long *)&U[1], 0, A.mslong, B.lslong);
      LongDiv((unsigned long *)&(Quot->lslong), (unsigned long *)&(Rem->lslong), U[1], A.lslong, B.lslong);
    }
    else{
      /* double by double */
      /* normalise the operands */
      Normalise(&Count, &V[1], &V[0], B.mslong, B.lslong);
      if (Count == 0){
        /* already normalised */
        U[2] = 0;
        U[1] = A.mslong;
        U[0] = A.lslong;
      }
      else{
        /* shift the dividend */
        ShiftLeft(&U[2], &X, 0, A.mslong, Count);
        ShiftLeft(&U[1], &U[0], A.mslong, A.lslong, Count);
      }
      /* evaluate the quotient */  
      Quot->mslong = 0;
    
      if (U[2] == V[1])
        Quot->lslong = MaxUnsignedInt32;
      else
        LongDiv((unsigned long *)&(Quot->lslong), (unsigned long *)&X, U[2], U[1], V[1]);
      
      if (Quot->lslong != 0){
        /* evaluate and check for overflow */
        LongProd(&W[1], &W[0], Quot->lslong, V[0], 0);
        LongProd(&W[2], &W[1], Quot->lslong, V[1], W[1]);
      
        LongDiff(&X, &U[0], U[0], W[0], 0);
        LongDiff(&X, &U[1], U[1], W[1], X);
        LongDiff(&X, &U[2], U[2], W[2], X);
        
        while ((U[2] & SignBit32) != 0){
          /* add back as required */
          Quot->lslong = Quot->lslong - 1;
        
          LongSum(&X, &U[0], U[0], V[0], 0);
          LongSum(&X, &U[1], U[1], V[1], X);
          LongSum(&X, &U[2], U[2], 0, X);
        }
      }
    
      /* evaluate the remainder */
      if (Count == 0){
        Rem->lslong = U[0];
        Rem->mslong = U[1];
      }
      else
        ShiftRight((unsigned long *)&(Rem->mslong), (unsigned long *)&(Rem->lslong), U[1], U[0], Count);
    }
  }
}     
/*}}}*/

/*{{{  signed integer division (Knuth - algorithm D, section 4.3.1 (vol 2, pp 257)) */
static void
ll_div_rem (int64_t *Quot, int64_t *Rem, int64_t A, int64_t B)
{
  int64_t U, V;
  short NegRem, NegQuot;
  unsigned long Borrow;
       
  if ((A.mslong & SignBit32) != 0){
    NegRem = TRUE;
    LongDiff ((unsigned long *)&(U.mslong), (unsigned long *)&(U.lslong), 0, A.lslong, 0);
    LongDiff ((unsigned long *)&Borrow, (unsigned long *)&(U.mslong), 0, A.mslong, U.mslong);
  }
  else{
    NegRem = FALSE;
    U.lslong = A.lslong;
    U.mslong = A.mslong;
  }
  if ((B.mslong & SignBit32) != 0){
    NegQuot = !NegRem;
    LongDiff ((unsigned long *)&(V.mslong), (unsigned long *)&(V.lslong), 0, B.lslong, 0);
    LongDiff ((unsigned long *)&Borrow, (unsigned long *)&(V.mslong), 0, B.mslong, V.mslong);
  }
  else{
    NegQuot = NegRem;
    V.lslong = B.lslong;
    V.mslong = B.mslong;
  }
  
  ull_div (Quot,Rem,U,V);
  
  if (/* there is no error from the unsigned divide */ 1){
    if (((Quot->mslong & SignBit32) != 0) && (!NegQuot))
      printf ("Overflow error!!!\n");
    else{
      if (NegQuot){
        /* check if the result has a negative quotient */
        LongDiff ((unsigned long *) &Borrow, (unsigned long *) &(Quot->lslong), 0, Quot->lslong, 0);
        LongDiff ((unsigned long *) &Borrow, (unsigned long *) &(Quot->mslong), 0, Quot->mslong, Borrow);
      }
      if (NegRem){
        /* check if the result has a negative remainder */
        LongDiff ((unsigned long *) &Borrow, (unsigned long *) &(Rem->lslong), 0, Rem->lslong, 0);
        LongDiff ((unsigned long *) &Borrow, (unsigned long *) &(Rem->mslong), 0, Rem->mslong, Borrow);
      }
    }
  }
}
/*}}}*/

/*{{{  integer multiplication */
int64_t
ll_mul (int64_t A, int64_t B)
{
  int64_t Result;
  unsigned long W[4], Carry1, Carry2, Carry3, Dummy, Asign, Bsign, Borrow;
    
  Asign = A.mslong & SignBit32;
  Bsign = B.mslong & SignBit32;
        
  LongProd((unsigned long *) &Carry1, (unsigned long *) &W[0], A.lslong, B.lslong, 0);
  LongProd((unsigned long *) &Carry1, (unsigned long *) &W[1], A.mslong, B.lslong, Carry1);
  LongProd((unsigned long *) &Carry2, (unsigned long *) &W[1], A.lslong, B.mslong, W[1]);
  LongProd((unsigned long *) &Carry3, (unsigned long *) &Carry2, A.mslong, B.mslong, Carry2);
        
  LongSum((unsigned long *) &Carry1, (unsigned long *) &W[2], Carry2, Carry1, 0);
  LongSum((unsigned long *) &Dummy, (unsigned long *) &W[3], Carry3, Carry1, 0);
     
  if (Asign != 0){
    /* if A < 0 subtract B from carry part */
    LongDiff((unsigned long *) &Borrow, (unsigned long *) &W[2], W[2], B.lslong, 0);
    LongDiff((unsigned long *) &Borrow, (unsigned long *) &W[3], W[3], B.mslong, Borrow);
  }
   
  if (Bsign != 0){
    /* if B < 0 subtract A from carry part */
    LongDiff((unsigned long *) &Borrow, (unsigned long *) &W[2], W[2], A.lslong, 0);
    LongDiff((unsigned long *) &Borrow, (unsigned long *) &W[3], W[3], A.mslong, Borrow);
  }
  
  /* check for overflow */
  
  if (((signed long) W[1]) < 0){
    if ((W[2] & W[3]) != -1)
     ; /*printf ("Overflow error!!!\n");*/
  }
  else{
    if ((W[2] | W[3]) != 0)
     ; /*printf ("Overflow error!!!\n");*/
  }
  
  Result.lslong = W[0];
  Result.mslong = W[1];
  return Result;
}
/*}}}*/

/*{{{  integer remainder */
int64_t
ll_rem (int64_t A, int64_t B)
{  
  int64_t Result, Dummy;
  
  if ((B.mslong == 0) && (B.lslong == 0))
    printf ("Divide by zero error!!!\n");
  else{
    if (((B.mslong == 0) && (B.lslong == 1)) || ((B.mslong == -1) && (B.lslong == -1)))
      Result.mslong = Result.lslong = 0;
    else
      ll_div_rem (&Dummy, &Result, A, B);
  }
  return Result;
}
/*}}}*/

/*{{{  integer quotient */
int64_t
ll_div (int64_t A, int64_t B)
{  
  int64_t Result, Dummy;

  if ((B.mslong == 0) && (B.lslong == 0))
    printf ("Divide by zero error!!!\n");
  else{
    if (((A.mslong == MOSTNEG_INT32) && (A.lslong == 0)) && ((B.mslong == -1) && (B.lslong == -1)))
     ; /*printf ("Arithmetic overflow error!!!\n");*/
    else
      ll_div_rem (&Result, &Dummy, A, B);
  }
  return Result;
}
/*}}}*/
/*}}}*/

/*{{{  bitwise logic */
/*{{{  bitwise AND */
int64_t
ll_and (int64_t A, int64_t B)
{ 
  int64_t Result;
   
  Result.lslong = (A.lslong) & (B.lslong);
  Result.mslong = (A.mslong) & (B.mslong);

  return Result;
}
/*}}}*/

/*{{{  bitwise NOT */
int64_t
ll_not (int64_t A)
{  
  int64_t Result;
   
  Result.lslong = ~(A.lslong);
  Result.mslong = ~(A.mslong);

  return Result;
}
/*}}}*/

/*{{{  bitwise OR */
int64_t
ll_or (int64_t A, int64_t B)
{
  int64_t Result;
    
  Result.lslong = (A.lslong) | (B.lslong);
  Result.mslong = (A.mslong) | (B.mslong);
   
  return Result;
}
/*}}}*/

/*{{{  logical shift left */
int64_t
ll_ushl (int64_t A, unsigned long n)
{
  int64_t Result;

  ShiftLeft ((unsigned long *) &(Result.mslong),(unsigned long *) &(Result.lslong),A.mslong,A.lslong,n);
 
  return Result;
}
/*}}}*/

/*{{{  logical shift right */
int64_t
ll_ushr (int64_t A, unsigned long n)
{  
  int64_t Result;
 
  ShiftRight((unsigned long *) &(Result.mslong), (unsigned long *) &(Result.lslong), A.mslong, A.lslong, n);
   
  return Result; 
}
/*}}}*/

/*{{{  arithmetic shift left */
int64_t
ll_shl (int64_t A, unsigned long n)
{
  int64_t Result;

  AShiftLeft ((signed long *) &(Result.mslong),(signed long *) &(Result.lslong),(signed long)A.mslong,(signed long)A.lslong,n);

  return Result;
}
/*}}}*/

/*{{{  arithmetic shift right */
int64_t
ll_shr (int64_t A, unsigned long n)
{
  int64_t Result;

  AShiftRight((signed long *) &(Result.mslong), (signed long *) &(Result.lslong), (signed long)A.mslong, (signed long)A.lslong, n);

  return Result;
}
/*}}}*/

/*{{{  bitwise XOR */
int64_t
ll_xor (int64_t A, int64_t B)
{
  int64_t Result;
 
  Result.lslong = A.lslong ^ B.lslong;
  Result.mslong = A.mslong ^ B.mslong;
 
  return Result;
}
/*}}}*/
/*}}}*/

/*{{{  type conversions */
/*{{{  long int -> long long int */
int64_t
int2ll (long A)
{
  int64_t Result;
    
  Result.lslong = A;
  if (A < 0)
    Result.mslong = -1;
  else
    Result.mslong = 0;

  return Result;
}
/*}}}*/

/*{{{  unsigned long int -> long long int */
int64_t 
uint2ll (unsigned long A)
{   
  int64_t Result;
    
  Result.lslong = A;
  Result.mslong = 0;

  return Result;
}
/*}}}*/

/*{{{  float -> long long int */
int64_t
float2ll (float A)
{
  int64_t Result;
  float B;
  
  B = floorf (A);
  if ((float)(A - B) < 0.5)
    Result.lslong = (unsigned long)B;
  else 
    Result.lslong = (unsigned long)ceilf(A);
 
  if (A < 0)
    Result.mslong = -1L;
  else
    Result.mslong = 0L;
    
  return Result;
}
/*}}}*/

/*{{{  long long int -> float */
/*{{{  ll2float */
float  
ll2float (int64_t A)
{
  float Result;  
  
  ll2float_x ((unsigned long *)&Result, Nearest, A.mslong, A.lslong);
  
  return Result;
}
/*}}}*/

/*{{{  ll2float_x */
static void
ll2float_x (unsigned long *Result, int Mode, unsigned long AHi, unsigned long ALo)
{
  unsigned long B[2], Xfrac, Xexp, Ans, Carry, Places;
 
  if ((AHi & SignBit32) != 0){
    LongDiff ((unsigned long *)&Carry, (unsigned long *)&B[0], 0, ALo, 0);
    LongDiff ((unsigned long *)&Carry, (unsigned long *)&B[1], 0, AHi, Carry);
    Ans = SignBit32;
  } 
  else{
    B[1] = AHi;
    B[0] = ALo;
    Ans = 0;
  }    

  Normalise ((unsigned long *)&Places, (unsigned long *) &B[1], (unsigned long *)&B[0], B[1], B[0]);
  
  if (Places == (2 * BitsPerWord32))
    *Result = 0;
  else{
    Xexp = (((2 * BitsPerWord32) - 1) - Places) + RealXcess;
 
    Places = B[1] & RealExp;

    ShiftRight ((unsigned long *)&Xexp, (unsigned long *)&Xfrac, Xexp, B[1] << 1, (unsigned long) (RealShift + 1));
    if ((Mode == Truncate) || ((Places & RealRBit) == 0))
      ;
    else if (((B[0] | (Places & RealXcess)) | (Xfrac & 1)) == 0)
      ;
    else
      Xfrac = Xfrac + 1;
    
    *Result = Ans | Xfrac;
  }
}
/*}}}*/
/*}}}*/

/*{{{  long long int -> double */
/*{{{  ll2double */
double
ll2double (int64_t A)
{ 
  double Result;
  unsigned long *Hi, *Lo;
 
  Lo = (unsigned long *)&Result;
  Hi = (unsigned long *)(&(((unsigned long *)&Result)[1]));
  ll2double_x (Hi, Lo, Nearest, A.mslong, A.lslong);
    
  return Result;
}
/*}}}*/

/*{{{  ll2double_x */
static void ll2double_x (unsigned long *XHi, unsigned long *XLo, int Mode, unsigned long NHi, unsigned long NLo)
{
  if ((NHi | NLo) == 0){
    *XHi = 0;
    *XLo = 0;
  }
  else{
    unsigned long Carry, Places, Xexp, Xfrac[2], Xint[2], Xsign;
 
    if ((NHi & SignBit32) != 0){
      Xsign = SignBit32;
      LongDiff((unsigned long *) &Carry, (unsigned long *) &Xint[0], 0, NLo, 0);
      LongDiff((unsigned long *) &Carry, (unsigned long *) &Xint[1], 0, NHi, Carry);
    }
    else{
      Xsign = 0;
      Xint[1] = NHi;
      Xint[0] = NLo;
    }
    Normalise((unsigned long *) &Places, (unsigned long *) &Xint[1], (unsigned long *) &Xint[0], Xint[1], Xint[0]);
    
    Xexp = (DRealXcess - 1) + ((2 * BitsPerWord32) - Places);
 
    ShiftLeft((unsigned long *) &Xint[1], (unsigned long *) &Xint[0], Xint[1], Xint[0], 1);
    ShiftRight((unsigned long *) &Xexp, (unsigned long *) &Xfrac[1], Xexp, Xint[1], (unsigned long)(DRealShift + 1));
    ShiftRight((unsigned long *) &Xexp, (unsigned long *) &Xfrac[0], Xint[1], Xint[0], (unsigned long) (DRealShift + 1));
  
    if (Places > (BitsPerWord32 + DRealShift))
      ;
    else
    {  
      Xint[0] = Xint[0] & DRealExp;
      if ((Mode == Truncate) || ((Xint[0] & DRealRBit) == 0))
        ;
      else if (((Xint[0] & DRealXcess) | (Xfrac[0] & 1)) == 0)
        ;
      else
      {
        LongSum((unsigned long *) &Carry, (unsigned long *) &Xfrac[0], Xfrac[0], 1, 0);
        LongSum((unsigned long *) &Carry, (unsigned long *) &Xfrac[1], Xfrac[1], 0, Carry);
      }
    }
    *XHi = Xsign | Xfrac[1];
    *XLo = Xfrac[0];
  } 
}
/*}}}*/
/*}}}*/

/*{{{  double -> long long int */
/*{{{  double2ll */
int64_t
double2ll (double A)
{   
  int64_t Result;
  int Error;
  unsigned long XHi, XLo;
    
  XLo = ((unsigned long *)&A)[0];
  XHi = ((unsigned long *)&A)[1];
  double2ll_x ((int *)&Error, (unsigned long *)&Result.mslong, (unsigned long *)&Result.lslong, Nearest, XHi, XLo);

  return Result;
}
/*}}}*/

/*{{{  double2ll_x */
static void double2ll_x (int *Error, unsigned long *NHi, unsigned long *NLo, int Mode,
                         unsigned long XHi, unsigned long XLo)
{
  unsigned long Carry, Dump, Guard, Xexp, Xfrac[2], Xsign;
  *Error = FALSE;
 
  ShiftLeft((unsigned long *) &Xexp, (unsigned long *) &Dump, 0, XHi & (~SignBit32), (unsigned long) (DRealShift + 1));
  ShiftLeft((unsigned long *) &Xfrac[1], (unsigned long *) &Xfrac[0], XHi, XLo, (unsigned long) (DRealShift + 1));
   
  if (Xexp == DRealExp)
    *Error = TRUE;
  else if ((XHi == REAL64MININT64Hi) && (XLo == 0)){
    *NHi = SignBit32;
    *NLo = 0;
  }
  else if (Xexp < (DRealXcess - 1)){
    *NHi = 0;
    *NLo = 0;
  }
  else if (Xexp < DRealXcess){
    if ((Mode == Truncate) || ((Xfrac[1] | Xfrac[0]) == 0)){
      *NHi = 0; 
      *NLo = 0;
    }
    else if ((XHi & SignBit32) == 0){
      *NHi = 0;
      *NLo = 1;
    }
    else{
      *NHi = -1;
      *NLo = -1;
    }
  }
  else{
    Xsign = XHi & SignBit32;
    Xexp = Xexp - DRealXcess;
    
    if (Xexp < BitsPerWord32){
      Carry = Xfrac[0];
      Xfrac[0] = Xfrac[1];
      Xfrac[1] = 1;
      Guard = 0;
    }
    else{
      Xexp = Xexp - BitsPerWord32;
      Carry = 0;
      Guard = 1;
    }
    
    if (Xexp > (BitsPerWord32 - 2))
      *Error = TRUE;
     
    ShiftLeft((unsigned long *) &Guard, (unsigned long *) &Dump, Guard, Xfrac[1], Xexp);
    ShiftLeft((unsigned long *) &Xfrac[1], (unsigned long *) &Dump, Xfrac[1], Xfrac[0], Xexp);
    ShiftLeft((unsigned long *) &Xfrac[0], (unsigned long *) &Carry, Xfrac[0], Carry, Xexp);
     
    if ((Mode == Truncate) || ((Xfrac[0] & SignBit32) == 0))
      ;
    else if ((((Xfrac[0] & (~SignBit32)) | Carry) | (Xfrac[1] & 1)) == 0)
      ;
    else{
      LongSum((unsigned long *) &Carry, (unsigned long *) &Xfrac[1], Xfrac[1], 1, 0);
      LongSum((unsigned long *) &Carry, (unsigned long *) &Guard, Guard, 0, Carry);
    }
    *NHi = Guard & (~SignBit32);
    *NLo = Xfrac[1];
    
    if (Xsign == 0)
      ;
    else{
      LongDiff((unsigned long *) &Carry, NLo, 0, *NLo, 0);
      LongDiff((unsigned long *) &Carry, NHi, 0, *NHi, Carry);
    }
  }  
}
/*}}}*/
/*}}}*/

/*{{{  string -> long long int */
int64_t
atoll (const char *nptr)
{
  int negative;
  register int64_t cutoff;
  register int64_t cutlim;
  register int64_t i;
  register const char *s;
  register unsigned char c;
  const char *save;
  int overflow;
  int base;

  base = 10;
   
  s = nptr;
   
  /* Skip white space.  */
  while (isspace (*s))
    ++s;
  if (*s == '\0')
    goto noconv;

  /* Check for a sign.  */
  if (*s == '-'){
    negative = 1;
    ++s;
  }
  else if (*s == '+'){
    negative = 0;
    ++s;
  }
  else
    negative = 0;  
  
  /* Save the pointer so we can check later if anything happened.  */
  save = s;
      
  cutoff = ll_div ((int64_t)ll_max_const , int2ll (base));
  cutlim = ll_rem ((int64_t)ll_max_const , int2ll (base));

  overflow = 0;    
  i = ll_zero_const;
  for (c = *s; c != '\0'; c = *++s){
      if (isdigit (c))
        c -= '0';
      else
        break;
      if (c >= base)
        break;
      /* Check for overflow.  */
      if (ll_gt(i, cutoff) || (ll_eq(i, cutoff) && ll_gt(int2ll(c), cutlim)))
        overflow = 1;
      else{
        i = ll_mul (int2ll(base), i);
        i = ll_add (int2ll(c), i);
      }
  }
  
  /* Check if anything actually happened.  */
  if (s == save)
    goto noconv;

  if (overflow){
    errno = ERANGE;
    return negative ? ll_min_const : ll_max_const;
  }

  /* Return the result of the appropriate sign.  */
  return (negative ? ll_neg(i) : i);

noconv:
  /* There was no number to convert.  */
  return ll_zero_const;
}
/*}}}*/
/*}}}*/

/*{{{  comparison operators */
/*{{{  equality */
int
ll_eq (int64_t A, int64_t B)
{
  return ((A.mslong == B.mslong) && (A.lslong == B.lslong));
}
/*}}}*/

/*{{{  inequality */
int
ll_ne (int64_t A, int64_t B)
{
  return ~((A.mslong == B.mslong) && (A.lslong == B.lslong));
}  
/*}}}*/
    
/*{{{  gt or eq */
int 
ll_ge (int64_t A, int64_t B)
{   
  return (A.mslong > B.mslong) ? 1 : (A.mslong == B.mslong ? (A.lslong >= B.lslong) : 0);
}  
/*}}}*/

/*{{{  lt or eq */
int
ll_le (int64_t A, int64_t B)
{  
  return (A.mslong < B.mslong) ? 1 : (A.mslong == B.mslong ? (A.lslong <= B.lslong) : 0);
}  
/*}}}*/

/*{{{  less than */
int
ll_lt (int64_t A, int64_t B)
{  
  return (A.mslong < B.mslong) ? 1 : (A.lslong < B.lslong);
}   
/*}}}*/

/*{{{  greater than */
int
ll_gt (int64_t A, int64_t B)
{  
  return (A.mslong > B.mslong) ? 1 : (A.lslong > B.lslong);
}
/*}}}*/
/*}}}*/
