/*{{{  Header*/
/* monitor_md.c - Machine dependent monitor calls for cm10
 *
 * The monitor system has been designed using semaphores and channel waits.
 * Java monitors have the following features: When a thread enters a monitor
 * it will either take the monitor, assuming that it: i) is already owned by
 * the current thread, in which case the thread now holds the monitor n + 1 times.
 *                 ii) is not currently held
 * by any thread.
 * If the monitor is already held by another thread then the current thread will
 * be queued (in a semaphore) until the monitor is released and there are no
 * other threads in front of it in the queue.
 * Once a thread owns a monitor it may wait for some length of time or until it is
 * "woken" by another thread. When this happens the current thread gives up the
 * monitor and is placed in the wait queue, until it wakes up or is woken. It
 * then waits for the monitor to become free in the usual way.
 */

/* Copyright (C) SGS-THOMSON Microelectronics Ltd, 1996 */
/*}}}  */

/*{{{  Includes*/
#include <stdio.h>
#include <unistd.h>
#include <exec/chExec.h>
#include <sync/chSem.h> /* STREAM Semaphores in here */
#include "cm10misc.h"
#include "monitor.h"
#include "monitor_md.h"
/*}}}  */

/*{{{  sysMonitorInit()*/
/* initialise monitor */

int sysMonitorInit(sys_mon_t *mid, bool_t in_cache)
{
  mid->monitor_owner = 0;
  mid->depth = 0;

  semInit(&mid->monitor_waitq,1);

  mid->wait_list_head = NULL;
  mid->wait_list_tail = NULL;
  return SYS_OK;
}
/*}}}  */
/*{{{  sysMonitorEnter()*/
/*
 * Take ownership of monitor. This can be called before threads have
 * been initialized, in which case we do nothing since locked monitors
 * are not yet needed.
 */

int sysMonitorEnter(sys_mon_t *mon)
{
  /* Check if there is anything in the queue, if there is
   * add this thread to the back of the queue.
   * If there is nothing in the queue, check for anything in the
   * suspended queue or the wait queue, if not take the monitor
   * else add this thread to the back of the queue.
  */

  sys_thread_t *self = sysThreadSelf();
  /*printf("sysMonitorEnter sysThreadSelf() call OK\n");*/

  /* If threads have not been initialised we do not need to do anything */
  if (self == 0) {
    return ThreadsInitialized ? SYS_ERR : SYS_OK;
  }
  /* If we are a valid thread, but threads have not been initialised, we are
     in trouble! */
  sysAssert(ThreadsInitialized);

  self->state = MONITOR_WAIT;
  if (mon->monitor_owner == self) {
    sysAssert(mon->depth > 0);
    /* printf ("Thread: %p already owns monitor %d times\n",self,mon->depth);*/
    mon->depth++;
    self->state = RUNNABLE;
  }
  else{
    /*printf ("Thread: %p Entering monitor: %p\n",self,mon); */
    semP(&mon->monitor_waitq,K_NOTIMEOUT);
    self->state = RUNNABLE;
    mon->depth=1;
    mon->monitor_owner=self;
    /* printf ("Thread: %p Got monitor\n",self); */
  }
  return SYS_OK;
}
/*}}}  */
/*{{{  sysMonitorEntered()*/
/*
 * Return TRUE if this thread currently owns the monitor. This can be
 * called before threads have been initialized, in which case we always
 * return TRUE.
 */

bool_t sysMonitorEntered(sys_mon_t *mid)
{
  sys_thread_t *self = sysThreadSelf();
  return !ThreadsInitialized || self != 0 && mid->monitor_owner == self;
}
/*}}}  */
/*{{{  sysMonitorExit()*/
/*
 * Release ownership of monitor. This can be called before threads have
 * been initialized, in which case we do nothing as locked monitors are
 * not yet needed.
 */

int sysMonitorExit(sys_mon_t *mon)
{
  sys_thread_t *self = sysThreadSelf();

  if (self == 0)
    return ThreadsInitialized ? SYS_ERR : SYS_OK;
  sysAssert(ThreadsInitialized);
  if (mon->monitor_owner != self){
    printf ("sysMonitorExit: ERROR: Thread does not own this monitor!\n");
    return SYS_ERR;
  }
  sysAssert(mon->depth > 0);
  /* give up the monitor and notify other waiting threads */
  if (--mon->depth == 0) {
    /* printf ("Thread: %p Exiting monitor: %p\n",self,mon); */
    /* Final release */
    mon->monitor_owner=NULL;
    if (mon->wait_list_head == NULL){
      semV(&mon->monitor_waitq);
      /* printf ("Thread: %p monitor exited: %p\n semaphore: %d\n",
               self,mon,mon->monitor_waitq->__count); */
      return SYS_DESTROY;
    }
    semV(&mon->monitor_waitq);
    /* printf ("Thread: %p monitor exited: %p\n semaphore: %d\n",
             self,mon,mon->monitor_waitq->__count); */
  }
  return SYS_OK;
}
/*}}}  */
/*{{{  sysMonitorNotify()*/
/*
 * Notify single thread waiting on condition variable.
 */

int sysMonitorNotify(sys_mon_t *mid)
{
  sys_thread_t *next = mid->wait_list_head;

  if (!sysMonitorEntered(mid)) {
    return SYS_ERR;
  }
  if (next == 0)
    return SYS_OK;

  semV(&next->wait);
  return SYS_OK;
}
/*}}}  */
/*{{{  sysMonitorDestroy()*/
/*   
 * If monitor owned by thread then release monitor and notify all other
 * waiting threads. If no other threads are waiting for the monitor then
 * return SYS_DESTROY so that the monitor can be freed up from the cache.
 */

int sysMonitorDestroy(sys_mon_t *mid, sys_thread_t *tid)
{
  sys_thread_t *next;

  if (mid->monitor_owner == tid) {
    mid->monitor_owner = 0;
    mid->depth = 0;
    next = mid->wait_list_head;
    if (next){
      while (next){
        printf ("doing chanout %p\n",mid);
        semV(&next->wait);
        next = next->next_waiting_thread;
      }
      mid->wait_list_head = mid->wait_list_tail = 0;
      semV(&mid->monitor_waitq);
      return SYS_OK;
    }
    semV(&mid->monitor_waitq);
    return SYS_DESTROY;
  }
  return SYS_OK;
}
/*}}}  */
/*{{{  sysMonitorDumpInfo()*/
/*
 * Dump some info about the monitor
 */

void sysMonitorDumpInfo(sys_mon_t *mon)
{
  if (mon->monitor_owner != 0){
    fprintf (stderr, "monitor: %p, owned by thread: %p\n",mon,mon->monitor_owner);
  }
  else
    fprintf (stderr, "monitor: %p, unowned\n",mon);
  if (mon->wait_list_head)
    fprintf(stderr, "        threads waiting\n");
}
/*}}}  */
/*{{{  sysMonitorNotifyAll()*/
/* 
 * notify all threads waiting on a condition variable
 */

int sysMonitorNotifyAll(sys_mon_t *mid)
{
  sys_thread_t *next = mid->wait_list_head;

  if (!sysMonitorEntered(mid)) {
    return SYS_ERR;
  }

  while (next != NULL){
    semV(&next->wait);
    next = next->next_waiting_thread;
  }
  return SYS_OK;
}
/*}}}  */
/*{{{  sysMonitorSizeof()*/
int sysMonitorSizeof(void)
{
  return sizeof (struct sys_mon);
}
/*}}}  */
/*{{{  sysMonitorWait()*/
/*
 * Atomically drop mutex and wait for notification.
 */

int sysMonitorWait(sys_mon_t *mid, int millis)
{
  /* add this thread the wait queue and notify waiting monitors */
  sys_thread_t *self = sysThreadSelf();
  int depth = mid->depth;

  if (self == 0) {
    return SYS_ERR;
  }
  sysAssert(ThreadsInitialized);
  if (mid->monitor_owner != self) {
    return SYS_ERR;
  }
  mid->monitor_owner = 0;
  mid->depth = 0;
  self->state = CONDVAR_WAIT;
  semV(&mid->monitor_waitq);  /* release monitor */

  /*
   * wait list maintenance
   */

  /* adding entries */

  if (mid->wait_list_head == NULL)  /* empty list */
    mid->wait_list_head = self;

  /* add self to back of list */  
  if (mid->wait_list_tail != NULL)  /* at least one item */
    mid->wait_list_tail->next_waiting_thread = self;

  /* set backpointer */
  mid->wait_list_tail = self;

  if (millis == SYS_TIMEOUT_INFINITY) {
    semP(&self->wait,K_NOTIMEOUT);
  }
  else{
    KnTimeVal wait;
    wait.tmSec=millis/1000;
    wait.tmNSec=millis*1000000-((millis/1000)*1000000000);

    semP(&self->wait,&wait);
  }

  /* removing entries */

  /* set head pointer to next element */
  mid->wait_list_head = mid->wait_list_head->next_waiting_thread;

  /* if this is the end of the list set the tail pointer to NULL */
  if (mid->wait_list_head == NULL)
    mid->wait_list_tail = NULL;

  /* set the next pointer of the removed element to NULL */
  self->next_waiting_thread = NULL;

  /* wait to be given back the monitor */
  semP (&mid->monitor_waitq,K_NOTIMEOUT);
  self->state = RUNNABLE;
  mid->depth = depth;
  mid->monitor_owner = self;

  return SYS_OK;
}
/*}}}  */
