/*{{{  Header*/
/*
 * @(#)classresolver.c  1.52 96/01/10  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*-
 *      Routines for loading and resolving class definitions.
 *      These routines should not depending up on the interpreter or
 *      garbage collector.
 *
 */
/*}}}  */

/*{{{  Includes*/
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>

#include "timeval.h"
#include "oobj.h"
#include "interpreter.h"
#include "signature.h"
#include "tree.h"
#include "monitor.h"
#include "bool.h"
#include "path.h"
#include "sys_api.h"
/*}}}  */

ClassClass *classJavaLangObject;
ClassClass *classJavaLangClass;
ClassClass *classJavaLangString;
ClassClass *classJavaLangThrowable;
ClassClass *classJavaLangException;
ClassClass *classJavaLangRuntimeException;
ClassClass *classJavaLangThreadDeath;

ClassClass *interfaceJavaLangCloneable;


char *RuntimeInitClass(struct execenv *, ClassClass *);

static ClassClass *Locked_FindClassFromClass(struct execenv *ee, char *name, bool_t resolve, ClassClass *from);
static char *Locked_InitializeClass(ClassClass * cb, char **detail);
static char *Locked_ResolveClass(ClassClass * cb, char **detail);

unsigned NameAndTypeToHash(char *name, char *type);

extern void lock_classes(void), unlock_classes(void);
static sys_mon_t *_class_lock;
static int class_lock_depth = 0;


/*{{{  addslots(struct fieldblock ** fb, ClassClass * clb)*/
static struct fieldblock **
addslots(struct fieldblock ** fb, ClassClass * clb)
{
    long n = clb->fields_count;
    struct fieldblock *sfb = cbFields(clb);
    if (cbSuperclass(clb))
	fb = addslots(fb, unhand(cbSuperclass(clb)));
    while (--n >= 0) {
	*fb++ = sfb;
	sfb++;
    }
    return fb;
}
/*}}}  */
/*{{{  makeslottable(ClassClass * clb)*/
void
makeslottable(ClassClass * clb)
{
    ClassClass *sclb;
    int         nslots = 0;
    if (cbSlotTable(clb)) {
	return;
    }
    sclb = clb;
    while (sclb) {
	long        n = sclb->fields_count;
	struct fieldblock *fb = cbFields(sclb);
	while (--n >= 0) {
	    nslots++;
	    fb++;
	}
	if (cbSuperclass(sclb) == 0) {
	    break;
	}
	sclb = unhand(cbSuperclass(sclb));
    }
    cbSlotTableSize(clb) = nslots;
    if (nslots == 0) {
	nslots++;
    }
    cbSlotTable(clb) = (struct fieldblock **)
	sysMalloc(nslots * sizeof(struct fieldblock *));
    if (cbSlotTable(clb) == 0) {
	out_of_memory();
	return;
    }
    addslots(cbSlotTable(clb), clb);
}
/*}}}  */
/*{{{  copyclassname(char *src, char *dst)*/
#if 0
static char *
copyclassname(char *src, char *dst)
{
    sysAssert(*src == SIGNATURE_CLASS);
    src++;
    while (*src && *src != SIGNATURE_ENDCLASS)
	*dst++ = *src++;
    *dst = 0;
    return src;
}
#endif
/*}}}  */
/*{{{  ResolveFields(ClassClass *cb, unsigned slot)*/
static void
ResolveFields(ClassClass *cb, unsigned slot)
{
    struct fieldblock *fb;
    int thishash = 0;
    char *s;
    int c;
    int size;

    fb = cbFields(cb);
    for (size = 0; size < (int) cb->fields_count; size++, fb++) {
	char *signature = fieldsig(fb);
	int size = (((signature[0] == SIGNATURE_LONG || 
		      signature[0] == SIGNATURE_DOUBLE)) ? 2 : 1);
	fb->ID = NameAndTypeToHash(fb->name, signature);
	if (fb->access & ACC_STATIC) {
	    /* Do nothing.  Handled when the class is loaded. */
	} else {
	    fb->u.offset = slot;
	    slot += size * sizeof(OBJECT);
	}
	if ((fb->access & (ACC_STATIC | ACC_TRANSIENT)) == 0) {
	    for (s = fieldname(fb); (c = *s++) != 0;)
		thishash = thishash * 7 + c;
	    for (s = fieldsig(fb); (c = *s++) != 0;)
		thishash = thishash * 7 + c;
	}
    }
    cbInstanceSize(cb) = slot;
    cbThisHash(cb) = thishash;
    if (cbSuperclass(cb))
	cbTotalHash(cb) = thishash - cbTotalHash(unhand(cbSuperclass(cb)));
    else
	cbTotalHash(cb) = thishash;
    if (cbTotalHash(cb) < N_TYPECODES)
	cbTotalHash(cb) += N_TYPECODES;
}
/*}}}  */
/*{{{  ResolveMethods(ClassClass *cb)*/
char *
ResolveMethods(ClassClass *cb)
{
    struct methodblock *mb;
    int size;
    struct methodtable *new_table;
    struct methodblock **super_methods;
    unsigned mslot, super_methods_count;
    void *ptr;
    static unsigned finalizerID = 0;

    if (finalizerID == 0)
	finalizerID = NameAndTypeToHash(FINALIZER_METHOD_NAME,
					FINALIZER_METHOD_SIGNATURE);
    if (cbSuperclass(cb) != NULL) { 
	ClassClass *super = unhand(cbSuperclass(cb));
	mslot = cbMethodTableSize(super);
	super_methods = cbMethodTable(super)->methods;
	super_methods_count = cbMethodTableSize(super);
	/* Inherit one's parent's finalizer, if it has one */
	cb->finalizer = unhand(cb->superclass)->finalizer;
    } else { 
	mslot = 1;
	super_methods = NULL;
	super_methods_count = 0;
	cb->finalizer = NULL;
    }

    mb = cbMethods(cb);
    for (size = 0; size < (int) cb->methods_count; size++, mb++) {
	unsigned long ID = NameAndTypeToHash(mb->fb.name, mb->fb.signature);
	struct methodblock **super_methods_p;
	int count;

	mb->fb.ID = ID;
	mb->fb.u.offset = 0;

	if (mb->fb.access & ACC_STATIC) 
	    continue;
	if (strcmp(mb->fb.name, "<init>") == 0)
	    continue;

	/* If this item has its own finalizer method, grab it */
	if (mb->fb.ID == finalizerID && cb != classJavaLangObject) 
	    cb->finalizer = mb;

	for (super_methods_p = super_methods, count = super_methods_count;
	       count > 0;
	       super_methods_p++, count--) {
	    if ((*super_methods_p != NULL) && ((*super_methods_p)->fb.ID == ID)) { 
		mb->fb.u.offset = (*super_methods_p)->fb.u.offset;
		break;
	    }
	}

	if (mb->fb.u.offset == 0) { 
	    mb->fb.u.offset = mslot;
	    mslot++;
	}
    }

    /*
	This should be the only place that method tables are
	allocated.  We allocate more than we need here and mask the
	resulting pointer because the methodtable pointer field in
	object handles is overloaded and sometimes hold array types
	and array lengths.  We must ensure that method table pointers
	are allocated on an appropriate boundary so we don't lose any
	address bits when we mask off the type portion of the pointer.
    */
    ptr = sysMalloc(sizeof(struct methodtable)
		       + (mslot - 1)* sizeof(struct methodblock *)
		       + FLAG_MASK);
    if (ptr == NULL) {
	CCSet(cb, Resolved);
	return "OutOfMemoryError";
    }
    new_table = (struct methodtable *)((((long)ptr) + FLAG_MASK) & LENGTH_MASK);
    new_table->classdescriptor = cb;
    memset((char *)new_table->methods, 0, mslot * sizeof(struct methodblock *));
    if (super_methods) 
	memcpy((char *) new_table->methods,
	       (char *) super_methods,
	       super_methods_count * sizeof(struct methodblock *));
    mb = cbMethods(cb);
    mb = cbMethods(cb);
    for (size = 0; size < (int) cb->methods_count; size++, mb++) {
	int offset = (int)mb->fb.u.offset;
	if (offset > 0) { 
	    sysAssert(offset < mslot);
	    mt_slot(new_table, offset) = mb;
	}
    }

    cbMethodTable(cb) = new_table;
    cbMethodTableSize(cb) = mslot;

    return 0;
}
/*}}}  */
/*{{{  InitializeClass(ClassClass * cb, char **detail)*/
char *
InitializeClass(ClassClass * cb, char **detail) 
{ 
    char *result;
    monitorEnter(obj_monitor(cb));
    result = Locked_InitializeClass(cb, detail);
    monitorExit(obj_monitor(cb));
    return result;
}
/*}}}  */
/*{{{  ResolveClass(ClassClass *cb, char **detail)*/
char *
ResolveClass(ClassClass *cb, char **detail) 
{
    char *result;
    monitorEnter(obj_monitor(cb));
    result = Locked_ResolveClass(cb, detail);
    monitorExit(obj_monitor(cb));
    return result;
}
/*}}}  */
/*{{{  Locked_InitializeClass(ClassClass * cb, char **detail)*/
static char *
Locked_InitializeClass(ClassClass * cb, char **detail)
{
    ClassClass *super = 0;
    char *ret = 0;

    if (CCIs(cb, Initialized))
	return 0;

#ifndef TRIMMED
    if (verbose)
	fprintf(stderr, "[Initializing %s]\n", classname(cb));
#endif

    if (cb->fields_count > 2000) {
	return "ClassFormatError";
    }
    if (strcmp(classname(cb), CLS_RESLV_INIT_CLASS) == 0) {
	classJavaLangClass = cb;
	classJavaLangString = FindClass(NULL, JAVAPKG "String", TRUE);

	classJavaLangThreadDeath = FindClass(NULL, JAVAPKG "ThreadDeath", TRUE);
	classJavaLangThrowable = FindClass(NULL, JAVAPKG "Throwable", TRUE);
	classJavaLangException = FindClass(NULL, JAVAPKG "Exception", TRUE);
	classJavaLangRuntimeException = 
			 FindClass(NULL, JAVAPKG "RuntimeException", TRUE);
	interfaceJavaLangCloneable = FindClass(NULL, JAVAPKG "Cloneable", TRUE);
    } else if (strcmp(classname(cb), CLS_RESLV_INIT_OBJECT) == 0) {
	classJavaLangObject = cb;
    }

    cbClassnameArray(cb) = MakeString(classname(cb), strlen(classname(cb)));

    if (cbHandle(cb) == 0) {
	cbHandle(cb) = 
	    (HClass *) AllocHandle(mkatype(T_BYTE, 1), (ClassObject *) cb);
	if (cbHandle(cb) == 0)
	    return "OutOfMemoryError";
    }
    if (strcmp(classname(cb), CLS_RESLV_INIT_REF) == 0)
	CCSet(cb, SoftRef);
    if (cbSuperclass(cb) == 0) {
	if (cb->super_name) {
	    if ((super = FindClassFromClass(0, classsupername(cb), FALSE, cb))) {
		sysAssert(CCIs(super, Initialized));
		cbSuperclass(cb) = cbHandle(super);
		if (CCIs(super, SoftRef))
		    CCSet(cb, SoftRef);
	    } else {
		ret = "NoClassDefFoundError";
		*detail = classsupername(cb);
		cbSuperclass(cb) = 0;
	    } 
	} else if (cb == classJavaLangObject) {
	    cbSuperclass(cb) = 0;           
	} else {
	    *detail = classname(cb);
	    return "ClassFormatException";
	}
    }
    CCSet(cb, Initialized);

    /* Make sure we know what classJavaLangClass is, and that it's method table
     * is filled in.
     */
    if (classJavaLangClass == 0) {
	classJavaLangClass = 
	      FindClassFromClass(0, CLS_RESLV_INIT_CLASS, TRUE, cb);
	if (classJavaLangClass == 0) {
	    return "NoClassDefFoundError";
	}
    }

    /* The following may not do anything useful if classJavaLangClass hasn't
     * been completely resolved.  But we clean up in ResolveClass 
     */

    cbHandle(cb)->methods = cbMethodTable(classJavaLangClass);
    return ret;
}
/*}}}  */
/*{{{  Locked_ResolveClass(ClassClass * cb, char **detail)*/
static char *
Locked_ResolveClass(ClassClass * cb, char **detail)
{
    ClassClass *super = 0;
    unsigned slot = 0;
    char *ret = 0;

    if (CCIs(cb, Error)) {
	return "NoClassDefFoundError";
    }

    sysAssert(CCIs(cb, Initialized));

    if (CCIs(cb, Resolved)) {
	return 0;
    }

    cbInstanceSize(cb) = -1;

    if (cbSuperclass(cb)) {
	/* If this object has a superclass. . . */
	super = unhand(cbSuperclass(cb));
	if (!CCIs(super, Resolved)) {
	    if ((ret = ResolveClass(super, detail)) != 0) {
		CCSet(cb, Error);
		return ret;
	    }
	}
	sysAssert(CCIs(super, Resolved));
	slot = cbInstanceSize(super);
	if ((int) slot < 0) {
#ifndef TRIMMED
	    fprintf(stderr, "%s: Can't resolve class because of interdependant static initializers\n", classname(cb));
#endif
	    CCSet(cb, Error);
	    return "NoClassDefFoundError";
	}
    }

#ifndef TRIMMED
    if (verbose)
	fprintf(stderr, "[Resolving %s]\n", classname(cb));
#endif
    CCSet(cb, Resolved);
    ResolveFields(cb, slot);
    ResolveMethods(cb);

    /*  Perform runtime-specific initialization on the class.  The
	definition of this routine if different depending on if we
	are compiling the interpreter or not.
    */
    if ((ret = RuntimeInitClass(0, cb))) {
	*detail = cb->name;
	CCClear(cb, Resolved);
	CCSet(cb, Error);
	return ret;
    }

    /* We need this for bootstrapping.  We can't set the Handle's class block
     * pointer in Object or Class until Class has been resolved.
     */
    if (cb == classJavaLangClass) {
	int         j;
	ClassClass **pcb;
	for (j = nbinclasses, pcb = binclasses; --j >= 0; pcb++) {
	    cbHandle(*pcb)->methods = cbMethodTable(cb);
	}
    }
    return 0;
}

/*}}}  */
/*{{{  FindLoadedClass(char *name, bool_t resolve, HObject *loader)*/
/* Find an already loaded class with the given name.  If no such class exists
 * then return 0.
 *
 * We attempt to resolve it, also, if the "resolve" argument is true.
 * Otherwise, we only perform a minimal initialization on it, such that
 * it points to its superclass, which points to its superclass. . . .
 */

#ifdef DEBUG
void
PrintLoadedClasses()
{
    int         j;
    ClassClass **pcb, *cb;
    pcb = binclasses;
    for (j = nbinclasses; --j >= 0; pcb++) {
	cb = *pcb;
	printf("class=%s, 0x%x\n", classname(cb), cb->loader);
    }
}
#endif

static ClassClass *
FindLoadedClass(char *name, bool_t resolve, HObject *loader)
{
    int         j;
    ClassClass **pcb, *cb;

    sysAssert(class_lock_depth > 0);

    pcb = binclasses;
    for (j = nbinclasses; --j >= 0; pcb++) {
	cb = *pcb;
	if ((cb->loader == loader) && (strcmp(name, classname(cb)) == 0)) {
	    char *err, *detail;
	    if (!CCIs(cb, Initialized)) {
		if ((err = InitializeClass(cb, &detail))) {
		    SignalError(0, err, detail);
		    return 0;
		}
	    }
	    if (resolve) {
		unlock_classes();
		err = ResolveClass(cb, &detail);
		lock_classes();
		if (err) {
		    SignalError(0, err, detail);
		    return 0;
		}
	    }
	    /*
	    if (resolve && !CCIs(cb, Resolved)) {
		if ((err = ResolveClass(cb, &detail))) {
		    SignalError(0, err, detail);
		    return 0;
		}
	    }
	    */
	    return cb;
	}
    }
    return 0;
}
/*}}}  */
/*{{{  FindClass(struct execenv *ee, char *name, bool_t resolve)*/
/* Find the class with the given name.  
 * If "resolve" is true, then we completely resolve the class.  Otherwise,
 * we only make sure that it points to its superclass, which points to its
 * superclass, . . . .
 */
ClassClass *(*class_loader_func)(HObject *loader, char *name, long resolve) = 0;

ClassClass *
FindClass(struct execenv *ee, char *name, bool_t resolve)
{
    return FindClassFromClass(ee, name, resolve,
	    (ee && ee->current_frame && ee->current_frame->current_method) ?
		  fieldclass(&ee->current_frame->current_method->fb) : 0);
}
/*}}}  */
/*{{{  lock_classes()*/
static ClassClass *
FindClassLoaderArrayClass(struct execenv *ee, char *name, 
			  bool_t resolve, ClassClass *from);


void
lock_classes()
{
    static bool_t initialized = FALSE;
    if (!initialized) {
	_class_lock = (sys_mon_t *) sysMalloc(sysMonitorSizeof());
	memset((char *) _class_lock, 0, sysMonitorSizeof());
	monitorRegister(_class_lock, "Class lock");
	initialized = TRUE;
    }
    sysMonitorEnter(_class_lock);
    class_lock_depth++;
}
/*}}}  */
/*{{{  unlock_classes()*/
void
unlock_classes()
{
    class_lock_depth--;
    sysMonitorExit(_class_lock);
}
/*}}}  */
/*{{{  Locked_FindClassFromClass(struct execenv *ee, char *name, bool_t resolve, ClassClass *from)*/
static ClassClass *
Locked_FindClassFromClass(struct execenv *ee, char *name, bool_t resolve, ClassClass *from)
{
    ClassClass *cb = 0;

    /*printf("In Locked_FindClassFromClass()\n");*/

    sysAssert(class_lock_depth > 0);

    if (from && from->loader && class_loader_func) {

	/*printf("Doing first if\n");*/

	if (name[0] == SIGNATURE_ARRAY) 
	    cb = FindClassLoaderArrayClass(ee, name, resolve, from);
	else {
	    /* unlock the classes while the class is located by
	     * the class loader. The class loader should do this
	     * in a synchronized method
	     */
	    unlock_classes();
	    cb = (*class_loader_func)(from->loader, name, resolve);
	    lock_classes();
	}
    }

    /*printf("First if overwith\n");*/

    if (cb == 0) {
	/*printf("Doing second if\n");*/
	cb = FindLoadedClass(name, resolve, 0);
	if (cb == 0) {
	    /*printf("Doing DoImport\n");*/
	    DoImport(name, 0);
	    cb = FindLoadedClass(name, resolve, 0);
	}
    }

    /*printf("Second if overwith\n");*/

    return cb;
}
/*}}}  */
/*{{{  FindClassFromClass(struct execenv *ee, char *name, bool_t resolve, ClassClass *from) {*/
ClassClass *
FindClassFromClass(struct execenv *ee, char *name, bool_t resolve, ClassClass *from) {
    ClassClass *cb;
    lock_classes();
    cb = Locked_FindClassFromClass(ee, name, resolve, from);
    unlock_classes();
    return cb;
}
/*}}}  */
/*{{{  FindClassLoaderArrayClass(struct execenv *ee, char *name,*/
static ClassClass *
FindClassLoaderArrayClass(struct execenv *ee, char *name, 
			  bool_t resolve, ClassClass *from)
{
    ClassClass *cb;
    char *ptr;
    HObject *loader = from->loader;
    if ((cb = FindLoadedClass(name, resolve, loader)) != NULL)
	return cb;
    for (ptr = name; *ptr == SIGNATURE_ARRAY; ptr++);
    if (*ptr != SIGNATURE_CLASS) {
	return NULL;
    } else {
	ClassClass *inner_cb;
	char buffer[256];

	jio_snprintf(buffer, sizeof(buffer), ptr+1);
	buffer[strlen(buffer) - 1] = '\0';

	inner_cb = FindClassFromClass(ee, buffer, FALSE, from);
	if (inner_cb == NULL || inner_cb->loader != loader) 
	    return NULL;
	cb = (ClassClass *)createFakeArrayClass(name);
	cb->loader = loader;
	AddBinClass(cb);
	/* this initializes */
	return FindLoadedClass(name, resolve, loader) ;
    }
}       
/*}}}  */
/*{{{  unsigned NameAndTypeToHash(char *name, char *type)*/
unsigned NameAndTypeToHash(char *name, char *type)
{
    extern struct StrIDhash *interfaceHash;
    unsigned name_key = Str2ID(&interfaceHash, name, 0, FALSE);
    unsigned type_key = Str2ID(&interfaceHash, type, 0, FALSE);
    return (name_key << 16) + type_key;
}
/*}}}  */
