/*
 * @(#)BreakpointHandler.java	1.12 95/10/04
 *
 * Copyright (c) 1995 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.tools.debug;

import java.util.*;
import java.io.IOException;

class BreakpointHandler extends Thread implements AgentConstants {
    private Agent		agent;
    private Hashtable		bkptHash;
    private BreakpointQueue	bkptQ;
    private Hashtable		catchHash;
    private Vector		skipLines;

    native int setBreakpoint(int pc) throws sun.tools.debug.InvalidPCException;
    native void clrBreakpoint(int pc, int oldOpcode) throws sun.tools.debug.InvalidPCException;

    synchronized void addBreakpoint(Class classObj, int pc,
				    int type, byte cond[]) throws IOException {
	BreakpointSet	bkpt;
	int		opcode;

	bkpt = (BreakpointSet)bkptHash.get(new Integer(pc));
	if (bkpt == null) {
	    Agent.message("Adding breakpoint for pc " + pc);
	    if (classObj == null) {
		Agent.message("Invalid classObj in addBreakpoint");
		return;
	    }
	    try {
	        opcode = setBreakpoint(pc);
	    } catch (InvalidPCException ex) {
		throw new IOException("invalid PC");
	    }
	    bkpt = new BreakpointSet(pc, opcode, classObj, type, cond);
	    bkptHash.put(new Integer(pc), bkpt);
	    if (cond == null || cond.length == 0) {
		bkpt.setUnconditional();
	    } else {
		bkpt.addCondition(cond);
	    }
	}
    }

    synchronized void deleteBreakpoint(Class classObj, int pc) throws IOException {
	BreakpointSet	bkpt;

	bkpt = (BreakpointSet)bkptHash.get(new Integer(pc));
	if (bkpt == null) {
	    return;
	}
	try {
	    clrBreakpoint(pc, bkpt.opcode);
	} catch (InvalidPCException ex) {
	    throw new IOException("invalid PC");
	}
	bkptHash.remove(new Integer(pc));
    }

    synchronized BreakpointSet listBreakpoints()[] {
	BreakpointSet ret[] = new BreakpointSet[bkptHash.size()];
	Enumeration e = bkptHash.elements();
	for (int i = 0; i < ret.length; i++) {
	    try {
		ret[i] = (BreakpointSet)e.nextElement();
	    } catch (NoSuchElementException nse) {
		// Should never happen!
		Agent.error("Internal error" + agent.exceptionStackTrace(nse));
		return null;
	    }
	}
	return ret;
    }

    synchronized void catchExceptionClass(Class c) {
	if (catchHash.get(c) == null) {
	    catchHash.put(c, c);
	}
    }

    synchronized void ignoreExceptionClass(Class c) {
	catchHash.remove(c);
    }

    Class getCatchList()[] {
	Class list[] = new Class[catchHash.size()];
	Enumeration enum = catchHash.elements();
	for (int i = 0; i < list.length; i++) {
	    list[i] = (Class)enum.nextElement();
	}
	return list;
    }

    void addSkipLine(LineNumber line) {
	agent.message("addSkipLine: thread=" + line.thread.getName() +
		      ", startPC=" + line.startPC + ", endPC = " + line.endPC);
	skipLines.addElement(line);
    }

    void removeSkipLine(LineNumber line) {
	skipLines.removeElement(line);
    }

    boolean ignoreSingleStep(Thread thread, int pc) {
	agent.message("ignoreSingleStep(" + thread.getName() + ", " + pc + ")");
	Enumeration e = skipLines.elements();
	while (e.hasMoreElements()) {
	    LineNumber ln = (LineNumber)e.nextElement();
	    agent.message("   ln: thread=" + ln.thread.getName() +
			  ", startPC=" + ln.startPC + ", endPC = " + ln.endPC);
	    if (ln.thread == thread) {
		if (pc >= ln.startPC && pc <= ln.endPC) {
		    return true;
		} else {
		    // Only one skip line per thread.
		    removeSkipLine(ln);
		}
	    }
	}
	agent.message("   no match.");
	return false;
    }

    BreakpointHandler(Agent ag) {
	super("Breakpoint handler");
	bkptQ = new BreakpointQueue();
	bkptQ.reset();
	bkptQ.the_bkptQ = bkptQ;
	bkptHash = new Hashtable();
	catchHash = new Hashtable();
	skipLines = new Vector();
	agent = ag;
    }

    public void run() {
	try {
	    BreakpointSet bkpt;

	    while (bkptQ.nextEvent()) {

		if (agent.asyncOutputStream == null) {
		    continue;
		}

		if (bkptQ.the_bkptQ.exception != null) {
		    if (bkptQ.the_bkptQ.exception instanceof ThreadDeath) {
			Agent.message("Received ThreadDeath event" +
				", pc=" + bkptQ.the_bkptQ.pc);
                        agent.asyncOutputStream.write(CMD_THREADDEATH_NOTIFY);
                        agent.writeObject(bkptQ.the_bkptQ.thread,
                                          agent.asyncOutputStream);
                        agent.asyncOutputStream.flush();
		    } else {
			Agent.message("Received exception " +
				bkptQ.the_bkptQ.exception.toString() +
				", pc=" + bkptQ.the_bkptQ.pc);
			Class excClass = bkptQ.the_bkptQ.exception.getClass();
			if (catchHash.containsKey(excClass) ||
			    bkptQ.the_bkptQ.catch_pc == 0) {

			    agent.suspendAllThreads();
			    String errorText = (bkptQ.the_bkptQ.catch_pc == 0) ?
				"Uncaught exception: " : "Exception: ";
			    String excText = agent.exceptionStackTrace(
				bkptQ.the_bkptQ.exception);
			    errorText = errorText.concat(excText);

			    agent.asyncOutputStream.write(CMD_EXCEPTION_NOTIFY);
			    agent.writeObject(bkptQ.the_bkptQ.thread,
					      agent.asyncOutputStream);
			    agent.asyncOutputStream.writeUTF(errorText);
			    agent.asyncOutputStream.flush();
			}
		    }
		} else if (bkptQ.the_bkptQ.opcode != -1) {
		    addBreakpoint(/*bkptQ.the_bkptQ.clazz*/ null,
				  bkptQ.the_bkptQ.pc, BKPT_NORMAL, null);
		} else {
		    bkpt = (BreakpointSet)bkptHash.get(
			new Integer(bkptQ.the_bkptQ.pc));
		    boolean ignoreBreakpoint = false;
		    if (bkpt == null) {
			Agent.message("Single step event: pc=" +
				      bkptQ.the_bkptQ.pc);
			if (ignoreSingleStep(bkptQ.the_bkptQ.thread,
					     bkptQ.the_bkptQ.pc)) {
			    continue;
			}
			agent.setSingleStep(bkptQ.the_bkptQ.thread, false);
			bkptQ.the_bkptQ.opcode = -1;
		    } else {
			Agent.message("Received a breakpoint event: pc=" +
				      bkptQ.the_bkptQ.pc);
			bkptQ.the_bkptQ.opcode = bkpt.opcode;
		    }
		    agent.suspendAllThreads();

		    if (bkpt != null && bkpt.type == BKPT_ONESHOT) {
			deleteBreakpoint(bkpt.clazz, bkpt.pc);
		    } else {
			agent.asyncOutputStream.write(CMD_BRKPT_NOTIFY);
			agent.writeObject(bkptQ.the_bkptQ.thread,
					      agent.asyncOutputStream);
			agent.asyncOutputStream.flush();
		    }
		}
	    }
	} catch (ThreadDeath td) {
	    Agent.message("BreakpointHandler: ThreadDeath received.");
	    /* Clear all breakpoints before exiting. */
	    Enumeration enum = bkptHash.elements();
	    while (enum.hasMoreElements()) {
		BreakpointSet bkpt = (BreakpointSet)enum.nextElement();
	        try {
		    clrBreakpoint(bkpt.pc, bkpt.opcode);
		} catch (Exception x) {
		    // Ignore exceptions here
		}
		bkptHash.remove(new Integer(bkpt.pc));
	    }
	    throw td;
	} catch (Exception ex) {
	    Agent.error(agent.exceptionStackTrace(ex));
	    Enumeration enum = bkptHash.elements();
	    while (enum.hasMoreElements()) {
		BreakpointSet bkpt = (BreakpointSet)enum.nextElement();
	        try {
		    clrBreakpoint(bkpt.pc, bkpt.opcode);
		} catch (Exception x) {
		    // Ignore exceptions here
		}
		bkptHash.remove(new Integer(bkpt.pc));
	    }
	    return;
	}
    }
}
