/*
 * @(#)ZipInputStream.java	1.3 95/11/11 David Connelly
 *
 * Copyright (c) 1995 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

package sun.tools.zip;

import java.io.InputStream;
import java.io.IOException;

/**
 * Zip file input stream, constructed from a zip file entry.
 *
 * @see ZipFile
 * @see ZipEntry
 * @version	1.3, 11/11/95
 * @author	David Connelly
 */
public
class ZipInputStream extends InputStream implements ZipConstants {
    private ZipEntry entry;	// Zip file entry for this stream
    private long pos;		// Current position within zip file
    private long count;		// Number of bytes left in entry data

    /**
     * Creates an input stream for the specified zip file entry.
     * @param entry the zip file entry
     * @exception ZipFormatException If the zip file format is invalid.
     * @exception IOException If an I/O error has occurred.
     */
    public ZipInputStream(ZipEntry entry) throws ZipFormatException,
	                                         IOException {
      this.entry = entry;
      init();
    }

    /**
     * Reads zip file entry data into an array of bytes. This method
     * will block until some input is available.
     * @param	b the buffer into which the data is read
     * @param	off the start offset of the data
     * @param	len the maximum number of bytes to read
     * @return	the actual number of bytes read, or -1 if the end of
     * 		the stream has been reached.
     * @exception ZipFormatException If the zip file format is invalid.
     * @exception IOException If an I/O error has occurred.
     */
    public synchronized int read(byte b[], int off, int len)
    	    throws ZipFormatException, IOException {
        if (count == 0) {
	    return -1;	// at end of file
	}
	if (len > count) {
	    len = (int)count;
	}
	int n = entry.zip.read(pos, b, off, len);
	if (n == -1) {
	    throw new ZipFormatException("Invalid LOC header");
	}
	pos += n;
	count -= n;
	return n;
    }

    /**
     * Reads a byte of data. This method will block until some input
     * is available.
     * @return	the byte read, or -1 if the end of the stream has been
     *		reached.
     * @exception ZipFormatException If the zip file format is invalid.
     * @exception IOException If an I/O error has occurred.
     */
    public synchronized int read() throws ZipFormatException, IOException {
	if (count == 0) {
	    return -1;	// at end of file
	}
	int n = entry.zip.read();
	if (n == -1) {
	    throw new ZipFormatException("Invalid LOC header");
	}
	pos += 1;
	count -= 1;
	return n;
    }

    /**
     * Skips n bytes of input.
     * @param n	the number of bytes to skip
     * @return	the actual number of bytes skipped
     */
    public synchronized long skip(long n) {
	if (n > count) {
	    n = count;
	}
	pos += n;
	count -= n;
	return n;
    }

    /*
     * Read LOC header for entry and position input stream at beginning
     * of entry data.
     */
    private void init() throws ZipFormatException, IOException {
	// Read LOC header and check signature
	byte locbuf[] = new byte[LOCHDRSIZ];
	entry.zip.read(entry.locpos, locbuf, 0, LOCHDRSIZ);
	if (!ZipFile.sigMatch(locbuf, LOCSIG)) {
	    throw new ZipFormatException("Invalid LOC header signature");
	}
	// Make sure entry is not compressed
	if (ZipFile.get16(locbuf, LOCHOW) != 0) {
	    throw new ZipFormatException("Compressed entry not supported");
	}
	// Make sure entry is not encrypted
	if ((ZipFile.get16(locbuf, LOCFLG) & 1) == 1) {
	    throw new ZipFormatException("Encrypted entry not supported");
	}
	// Get length and position of entry data
	count = entry.length;
	pos = entry.locpos + LOCHDRSIZ;
	pos += ZipFile.get16(locbuf, LOCNAM) + ZipFile.get16(locbuf, LOCEXT);
	if (pos + count > entry.zip.cenpos) {
	    throw new ZipFormatException("Invalid LOC header format");
	}
    }
}
