/*
 * @(#)context.c	1.43 96/01/11
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Initialize the platform dependent parts of a thread context.
 */

#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <unistd.h>
#include <memory.h>

#include "sys_api.h"
#include "threads_md.h"
#include "interrupt.h"
#include "itimer_md.h"
#include "sparc_md.h"
#include "context.h"
#include "log.h"
#include "schedule.h"

#include "sys_api.h"

/*
 * The list on which we put freed thread stacks pending reclamation.
 */
typedef struct free_stack_t {
    struct free_stack_t *next;
    long   stack_size;
} free_stack_t;

static free_stack_t *stackFreeList = (free_stack_t *) 0;
static int stackFreeCount = 0;

static int redZoneSize = 0;	/* One page */

/*
 * Free tid and stack space for dead threads.  We do this whenever
 * we do sysThreadCreate() and there are more than FREESTACKTHRESHOLD
 * free stacks.  We only free stacks in excess of FREESTACKTHRESHOLD.
 * Note that we have to play some games here because we need to make
 * stackFreeList and stackFreeCount access Green thread-safe, but can't
 * call free() with SCHED_LOCKED().
 *
 * The value for FREESTACKTHRESHOLD probably isn't too critical, at
 * least for current uses of Java in HotJava.  The choice of 5 means
 * that the majority of stack allocations when browsing HTML pages
 * including applets will be satisfied from the free lists, and
 * only when pages are left are there sometimes enough stacks freed
 * to flush them back to the O/S.  This may need to be more sophisti-
 * cated in the future.
 */

#define FREESTACKTHRESHOLD 5

static void freeStacks() 
{
    free_stack_t* stack = 0;
    free_stack_t* next;
    long size;

    if (stackFreeCount > FREESTACKTHRESHOLD) {
	SCHED_LOCK();
	/* Set up to stacks in excess of FREESTACKTHRESHOLD */
	while (stackFreeCount > FREESTACKTHRESHOLD) {
	    free_stack_t *freeStack;
	    freeStack = stackFreeList;
	    stackFreeList = freeStack->next;
	    freeStack->next = stack;
	    stack = freeStack;
	    stackFreeCount--;
	}
	SCHED_UNLOCK();
	while (stack) {
	    long junk;
	    next = stack->next;
	    Log3(2,
    "Over free stack threshold: freeing (0x%x:0x%x), new free count %d\n",
		 (char *) stack+stack->stack_size,
		 stack->stack_size, stackFreeCount);
	    /* Don't forget the redzone... */
	    size = stack->stack_size + redZoneSize;
	    /* REMIND: Ignoring return value here... */
	    sysUnmapMem((char *)stack - redZoneSize, size, &junk);
	    stack = next;
	}
    }
}

/*
 * Allocate a single thread stack of specified size.
 *
 * Note that "stack base" here and elsewhere means the base of the
 * logical stack, growing from high addresses to low addresses, and
 * *not* the way you'd think of, say, a malloc'ed chunk.
 *
 * We let the low level stack allocator take care of any rounding or
 * other system-specific sizing of the stack request.  But if no stack
 * size is given we take the default one. 
 */

static int 
allocateStack(gstack_t *stack, long size)
{
    extern int ProcStackSize;
    free_stack_t *freeStack;
    free_stack_t **freeStackPrev;

    freeStacks();		/* Free any stacks for dead threads */

    if (size == 0) {
        /* Allocate conservatively if we don't know better */
        size = ProcStackSize;
    }

    /* Take stack off the stackFreeList if one's available */
    SCHED_LOCK();
    freeStack = stackFreeList;
    freeStackPrev = &stackFreeList;
    while (freeStack && freeStack->stack_size != size) {
	freeStackPrev = &freeStack;
	freeStack = freeStack->next;
    }

    if (freeStack) {		/* Got one */
	*freeStackPrev = freeStack->next;
	stackFreeCount--;
	SCHED_UNLOCK();
	stack->base = (stackp_t) ((char *) freeStack + freeStack->stack_size);
	stack->size = freeStack->stack_size;
	Log3(2, "New stack from free list (0x%x:0x%x), new free count %d\n",
	     stack->base, stack->size, stackFreeCount);
	return 1;
    } else {			/* Didn't get one */
	SCHED_UNLOCK();		/* Can't allocate with SCHED_LOCKED() */
	if (redZoneSize == 0) {
	    redZoneSize = sysconf(_SC_PAGESIZE);	/* one page */
	}
	size += redZoneSize;	/* For redzone */
	stack->base = (stackp_t) sysMapMem(size, &size);
	/*
	 * On Solaris we want to demand-page stack memory, so we
	 * don't sysCommitMem() here...
	 */
	if (!stack->base) {
	    return 0;		/* Allocation failed */
	} else {
	    if (mprotect((char *)stack->base, redZoneSize, PROT_NONE)) {
		return 0;
	    }
	    stack->base = (stackp_t) ((char *) stack->base + size);
	    /* gstack_t.size does NOT include the redzone */
	    stack->size = size - redZoneSize;
	    Log3(2, "New stack (0x%x:0x%x), free count %d\n",
		 stack->base, stack->size, stackFreeCount);
	    return 1;
	}
    }
}

/*
 * Returns the stack pointer for any threads but the current one.
 */
void *
currentStackPointer(sys_thread_t *tp)
{
    context_t   *context = &tp->mdcontext;
    ucontext_t	*uc = &(context->ucontext);
#ifdef i386
    return (void *) (uc->uc_mcontext.gregs[USP]);
#else  /* Presumed to be SPARC */
    return (void *) (uc->uc_mcontext.gregs[REG_O6]);
#endif /* i386 */

}

#ifdef i386

/* Solaris/x86 version */

/*
 * This is the stack layout, as defined by the Intel ABI
 *
 * In the Intel world, a stack frame looks like this:
 *
 * %fp0->|                               |
 *       |-------------------------------|
 *       |  Args to next subroutine      |
 *       |-------------------------------|-\
 * %sp0->|  One word struct-ret address  | |
 *       |-------------------------------|  > minimum stack frame (8 bytes)
 *       |  Previous frame pointer (%fp0)| |
 * %fp1->|-------------------------------|-/
 *       |  Local variables              |
 * %sp1->|-------------------------------|
 */

void
initContext(sys_thread_t *tid, unsigned int pc,
         void (*death_func)(), unsigned int arg)
{
    context_t *mdcontext = CONTEXT(tid);
    ucontext_t *uc = &(mdcontext->ucontext);
    unsigned char *sp;
    int *p;

    getcontext(uc);
    sp = (unsigned char *) tid->stack_base;
    /* Note: WINDOWSIZE stuff is for SPARC, not really needed for x86 */
    sp -= (sizeof(sys_thread_t) + WINDOWSIZE + SA(MINFRAME));
    /* Force sp to be double aligned! */
    sp = (unsigned char *)((unsigned long)(sp) & 0xfffffff8);
    /* for x86 the args to death_func are passed on the stack */
    p = (int *)sp;
    *p-- = (int ) pc;
    *p-- = (int ) arg;
    *p-- = (int ) 1; /* dummy arg - for compatibility with SPARC version */

    uc->uc_mcontext.gregs[USP] = (unsigned int) p;
    uc->uc_mcontext.gregs[PC] = (unsigned int) death_func;
}

#else

/* Solaris/SPARC version */
 
/*
 * This is the stack layout, as defined by the SPARC ABI:
 *
 *      HIGH ADDRESSES
 *	|-----------------------|
 *	| ThreadPrivate data	|
 *      |-----------------------|
 *      | 16 words for in and   |  death_func() stack frame
 *      | local save area       |
 *      |-----------------------|  --------
 *      | arguments past the 6th|
 *      |-----------------------|  startup func stack frame
 *	| 6 words for locals    |
 *	|-----------------------|
 *      | 1 word for aggregate  |
 *      | return value          |
 *      |-----------------------|
 *      | %i7 = startaddr       |
 *      | %i6 = sp              |
 *      | %i0-%i1 --> args 1-5  |
 *      |-----------------------|
 *      | 8 words for saved %l's|
 *      |-----------------------|       <--- sp(double aligned)
 *      LOW ADDRESSES
 */
void
initContext(sys_thread_t *tid, unsigned int pc,
	     void (*death_func)(), unsigned int arg)
{
    unsigned char *sp;
    ucontext_t *uc; 
    context_t *mdcontext = CONTEXT(tid);
    machdep_t *machdepP = MACHDEP(tid);

    uc = &(mdcontext->ucontext);
    sp = (unsigned char *) tid->stack_base;
    sysAssert(sp == ((unsigned char *)tid+sizeof(sys_thread_t)));

    sp -= (sizeof(sys_thread_t) + WINDOWSIZE + SA(MINFRAME));

    /* Force sp to be double aligned! */
    sp = (unsigned char *)((unsigned long)(sp) & 0xfffffff8); 

    /* clear the context structure */
    memset(uc, 0, sizeof(ucontext_t)); 
    uc->uc_stack.ss_sp = (char *) sp;
    uc->uc_stack.ss_size = tid->stack_size;
    uc->uc_stack.ss_flags = 0; /* ? */

    /* initialially all signals are enabled */
    /* FIXME change to be initially disabled? */
    memset((char *)&(uc->uc_sigmask), 0, sizeof(sigset_t));
    machdepP->intrLockCount = 0;

    uc->uc_mcontext.gregs[REG_SP] = (unsigned int) sp;
    uc->uc_mcontext.gregs[REG_PC] = (unsigned int) death_func;
    uc->uc_mcontext.gregs[REG_nPC] = (unsigned int) death_func+4;
    uc->uc_mcontext.gregs[REG_O1] = arg; 
    uc->uc_mcontext.gregs[REG_O2] = (int) pc;

    uc->uc_flags = UC_ALL;
    return;
}

#endif /* i386 */

/*
 * This routine allocates enought memory for the thread context
 * and the stack.  This implementation puts the thread context
 * at the top of the stack.
 *
 * Must zero memory and initialize tid->PrivateInfo and
 * tid->PrivateInfo->stack.
 */
int
allocateContextAndStack(sys_thread_t **ptid, long requested_size)
{
    gstack_t stack;
    char *sp;

    if (!allocateStack(&stack, requested_size)) {
	return 0;
    }

    sp = (char *) stack.base;
    sp -= sizeof(sys_thread_t);
    memset(sp, 0, sizeof(sys_thread_t));

    *ptid = (sys_thread_t *) sp;
    (*ptid)->stack_base = stack.base;
    (*ptid)->stack_size = stack.size;	/* Actual size allocated */
    return 1;
}

/*
 * Delete the specified context, and the stack associated with it.
 * This does *not* free the memory associated with the stack: that
 * is handled elsewhere!
 *
 * WARNING: Called when SCHED_LOCKED(), so may not enter any monitors!
 */
void
deleteContextAndStack(sys_thread_t *tid)
{
    /*
     * If tid->primordial_thread (Unix main()), we don't want to free
     * the C stack, but we have malloc'ed the sys_thread_t.  This is
     * NOT the normal case!  But we don't want this going back to the
     * stack allocator, so we drop the sys_thread_t on the floor here.
     */
    if (!tid->primordial_thread) {
        free_stack_t *tmp;
	/*
	 * The normal case: queue the stack and the thread private data
	 * to be freed when safe.
	 */
	sysAssert(0 <= tid->stack_size);
	/* Note that free stacks are pointed to *above* the redzone */
	tmp = (free_stack_t *)((char *) tid->stack_base - tid->stack_size);
	tmp->stack_size = tid->stack_size;
	tmp->next = stackFreeList;
	stackFreeList = tmp;
	stackFreeCount++;
	Log3(2, "Freed stack (0x%x:0x%x), new free count %d\n",
	     tid->stack_base, tid->stack_size, stackFreeCount);
    }
    return;
}
