/*
 * @(#)io_md.c	1.8 95/11/29  Pavani Diwanji
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Solaris green threads dependent I/O:
 * This file contains all the io methods that are called by the java.io package.
 *
 * For native share libraries making direct system io calls, overriding wrappers
 * are provided in the file iomgr.c
 *
 * Files io_md.c and iomgr.c are related - io_md.c methods are called from
 * java land and hence they are passes the filedescriptor object. 
 * iomgr.c methods are called from C (by shared libraries), so they directly 
 * deal with the integer fd.
 *.
 * Soln for solving the net/io race conditions:
 * 	read(fdobj) {
 *        fd = incr(fdobj); 
 *	  read(fd);
 *	  decr(fdobj); 
 *	}
 *      incr() { lock; refcnt++; unlock }
 *      decr() { 
 *         lock; // lock is per fd lock; 
 *         refcnt--; 
 *         if (refcnt==0 and fdclosed) system_close();
 *         unlock 
 *      }
 *	close() {
 *	  lock();
 *	  if (refcnt>0) { 
 *              fdclosed = true;
 *		<interrupt any ongoing io>
 *	  }
 *        else system_close();
 *	  unlock();
 *	}
 * NOTE: 
 * If we are doing non-blocking/async-io the step <interrupt any ongoing io>
 * is easy as its just equivalent to signalling on per fd monitor.
 * But platform where we do blocking io, we need to keep track of 
 * all the thread ids doing blocking io and somehow have a way of
 * kicking the thread out of ongoing io call with an error.
 */

#include <errno.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/filio.h>
#include <sys/socket.h>
#include <poll.h>

#include "threads.h"
#include "threads_md.h"
#include "monitor_md.h"
#include "itimer_md.h"
#include "iomgr.h"
#include "log.h"
#include "fd_md.h"

extern int pendingException();
extern boolean_t initialize_monitors(int);
extern int system_close(int);

/* initialize the stdin, stdout, stderr */
void sysInitFD(Classjava_io_FileDescriptor *fdptr, int desc) {
  fdptr->fd = desc+1;
}

/* since fd=0 means uninitialized fd, we bump up fd before
returning */

int
sysOpenFD(Classjava_io_FileDescriptor *fdobj, const char *fname, int flags, 
int mode)
{
    register int fd;
    boolean_t ret = B_TRUE;

    IO_LOCK();

    while(1) {
	fd = (*systable[SYS_OPEN].addr)(fname, flags, mode);
	if ((fd != -1) || pendingException() ||  
	    ((errno != EAGAIN) && (errno != EINTR)))
	    break;      
    }
 
    if (fd != -1) {
	Log1(1, "Opened fd: %d\n", fd);
	ret = initialize_monitors(fd);
	fdobj->fd = fd+1;
    } else {
	Log1(1, "open error: %d\n", errno);
    }

    IO_UNLOCK();

    sysAssert(fd != 0 && ret != B_FALSE);
    return (fd!=-1)?fd+1:fd;
}

int sysCloseFD(Classjava_io_FileDescriptor *fdobj) {
    sys_mon_t *mon; 
    int ret = -1;
    int fd = fdobj->fd-1;

    if (fd < 0) {
        Log1(1, "Closing bad fd: %d?\n", fd);
        return ret;
    }

    mon = fdmon[fd];
    sysAssert(mon!=NULL);

    sysMonitorEnter(mon);

    fd = fdobj->fd-1;
    if (fd>=0) {
      if (fd_ref[fd] == 0) {
	if (!closed_fd(fd)) 
          ret = system_close(fd);
	  fdobj->fd = -1;
      } else {
	if (!closed_fd(fd)) {
	  fd_flags[fd] |= FD_CLOSED;
	  sysMonitorNotifyAll(mon);
	  sysMonitorNotifyAll(asyncMon(SYS_ASYNC_MON_IO));
	  Log1(1, "Marked Closed fd: %d\n", fd);
	}
	ret = 0;
      }
    }

    sysMonitorExit(mon);

    return ret;

}


int sysReadFD(Classjava_io_FileDescriptor *fdobj, char *buf, int nbytes) {
    sys_mon_t *mon; 
    register int nread;
    int fd = fdobj->fd-1;

    if (fd < 0) {
      Log1(1, "Read on closed fd: %d?\n", fd);
      return -1;
    }

    mon = fdmon[fd];
    sysAssert(mon!=NULL);

    check_single_fd(fd);

    sysMonitorEnter(mon);
    fd = fdobj->fd-1;
    if (fd>=0) {
        fd_ref[fd]++;

        while(!pendingException() && !closed_fd(fd)) {
	    nread = (*systable[SYS_READ].addr)(fd, buf, nbytes);
	    if ((nread != -1) || pendingException() ||  
		((errno != EAGAIN) && (errno != EINTR)))
	      break;
	    if (errno == EAGAIN) 
	      sysMonitorWait(mon, READ_TIMEOUT);
	}

	fd_ref[fd]--;
	/* If I am the last person to leave and fd has been closed
	   call system close on the fd */
	if (fd_ref[fd]==0 && closed_fd(fd)) system_close(fd); 

    }
    Log2(1, "Read for %d bytes finished on fd: %d\n", nread, fd);
    sysMonitorExit(mon);
    return nread;

}


int sysWriteFD(Classjava_io_FileDescriptor *fdobj, char *buf, int nbytes) {
    register int nwrote;
    register caddr_t next = (caddr_t)buf;
    register int total_written = 0;
    sys_mon_t	*mon;
    int fd = fdobj->fd-1;

    if (fd < 0) {
      Log1(1, "Write on closed fd: %d?\n", fd);
      return -1;
    }

    mon = fdmon[fd];
    sysAssert(mon!=NULL);

    sysMonitorEnter(mon); 
    fd = fdobj->fd-1;
    if (fd>=0) {
        fd_ref[fd]++;

	while(!pendingException() && !closed_fd(fd)) {
	    nwrote = (*systable[SYS_WRITE].addr)(fd,
			 next + total_written,
			 nbytes - total_written);
	    if ((nwrote==-1) && (errno != EAGAIN && errno != EINTR)) {
	         total_written = -1;
		 break;
	    }
	    if (nwrote==-1) { 
	      /* with errno== EAGAIN or EINTR */

	        if (errno == EAGAIN)
		  sysMonitorWait(mon, WRITE_TIMEOUT);
		continue;
	    }
	    total_written += nwrote;
	    if (total_written >= nbytes) break;
	} /* end of while */

	fd_ref[fd]--;
	/* If I am the last person to leave and fd has been closed
	   call system close on the fd */
	if (fd_ref[fd]==0 && closed_fd(fd)) system_close(fd); 
    }

    Log2(1, "Write for %d bytes finished on fd: %d\n", nwrote, fd);
    sysMonitorExit(mon);

    return total_written;
}


/* for the following routines we hold the lock for the duration
of blocked system/library call */
int
sysAvailableFD(Classjava_io_FileDescriptor * fdobj, long *pbytes)
{
    off_t cur, end;
    struct stat stbuf;
    int fd = fdobj->fd-1;
    int ret = 1; /* success */
    sys_mon_t	*mon;

    if (fd < 0) {
      Log1(1, "available operation on closed fd: %d?\n", fd);
      return 0;
    }

    mon = fdmon[fd];
    sysAssert(mon!=NULL);

    sysMonitorEnter(mon);     

    fd = fdobj->fd-1;
    if (fd>=0) {
      if (fstat(fd, &stbuf) >= 0) {
          int mode = stbuf.st_mode;
	  if (S_ISCHR(mode) || S_ISFIFO(mode) || S_ISSOCK(mode)) {
              if (ioctl(fd, FIONREAD, pbytes) >= 0) {
                  return 1;
	      }
	  }
      }
      if ((cur = lseek(fd, 0L, SEEK_CUR)) == -1) {
	  ret = 0;
      } else if ((end = lseek(fd, 0L, SEEK_END)) == -1) {
	  ret = 0;
      } else if (lseek(fd, cur, SEEK_SET) == -1) {
	  ret = 0;
      }
      *pbytes = end - cur;
    }

    sysMonitorExit(mon); 
  
    return ret;
}

long
sysLseekFD(Classjava_io_FileDescriptor *fdobj, long offset, long whence) {
    int ret = -1;
    int fd = fdobj->fd-1;
    sys_mon_t	*mon;

    if (fd<=0) {
      Log1(1, "lseek operation on closed fd: %d?\n", fd);
      return ret;
    }

    mon = fdmon[fd];
    sysAssert(mon!=NULL);

    sysMonitorEnter(mon);  
    fd = fdobj->fd-1;
    if (fd>=0) 
      ret = lseek(fd, offset, whence);
    sysMonitorExit(mon); 

    return ret;
}

/******************** Socket related calls ************************/

void sysSocketInitializeFD(Classjava_io_FileDescriptor *fdptr, int infd) {
    sys_mon_t *mon = fdmon[infd];
    sysAssert(mon!=NULL);

    sysMonitorEnter(mon);    

    fdptr->fd = infd+1;

    /* make the socket address immediatly available for reuse */
    setsockopt(infd, SOL_SOCKET, SO_REUSEADDR, "yes", 4);
#ifndef NATIVE
    nonblock_io(infd, IO_DONTBLOCK);
#endif

    sysMonitorExit(mon);    
}

int 
sysListenFD(Classjava_io_FileDescriptor *fdobj, long count) {
   int fd = fdobj->fd - 1;
   int ret = -1;
   sys_mon_t *mon;

   if (fd < 0) {
     Log1(1, "listen operation on closed fd: %d?\n", fd);
     return ret;
   }

   mon = fdmon[fd];
   sysAssert(mon!=NULL);

   sysMonitorEnter(mon);  
   fd = fdobj->fd - 1;
   if (fd>=0) ret = listen(fd, count);
   sysMonitorExit(mon);

   return ret;
}

int 
sysAcceptFD(Classjava_io_FileDescriptor *fdobj, struct sockaddr *him, 
int *len) {

   int fd = fdobj->fd - 1;
   int ret = -1;
   sys_mon_t *mon;

   if (fd < 0) {
     Log1(1, "listen operation on closed fd: %d?\n", fd);
     return ret;
   }

   mon = fdmon[fd];
   sysAssert(mon!=NULL);

   sysMonitorEnter(mon);  
   fd = fdobj->fd - 1;
   if (fd>=0) ret = accept(fd, him, len);
   sysMonitorExit(mon);

   return ret;
}

int
sysConnectFD(Classjava_io_FileDescriptor *fdobj, struct sockaddr *him, int len)
{
    int	   result=-1;
    int	   delay = 50;	/* sleep 50 ms and try again */
    fd_set write_fds;
    struct timeval  poll;
    sys_mon_t *mon;
    extern int _select (int width, fd_set *, fd_set *, fd_set *, 
			struct timeval *);
    int s = fdobj->fd-1;

    if (s < 0) {
        Log1(1, "connect operation on closed fd: %d?\n", s);
        return result;
    }

    mon = fdmon[s];
    sysAssert(mon!=NULL);

    sysMonitorEnter(mon);
    s = fdobj->fd-1;
    if (s>=0) {
        FD_ZERO(&write_fds);
        FD_SET(s, &write_fds);
	poll.tv_sec = poll.tv_usec = 0;

	result = connect(s, him, len);
	while (result < 0 && !pendingException()) {
	   if (errno == EINPROGRESS) {
	     int	cnt = _select(s+1, 0, &write_fds, 0, &poll);

	     FD_SET(s, &write_fds);
	     if (cnt == 1) {
	       int bytes;

	       if (ioctl(s, FIONREAD, &bytes) == -1) {
		 if (errno == EPIPE) {
		   errno = ECONNREFUSED;
		 }
		 /* If this fails with EPIPE, then we assume
		    connection is refused - who knows what's
		    right?  Not I. */
		 sysAssert(result < 0);
		 break;
	       }
	       result = 0;
	       break;
	     }
	     sysMonitorWait(mon, delay);
	   } else {
	     break;
	   }
	} /* end of while */
    }
    sysMonitorExit(mon);
    return result;
}


int sysRecvFD(Classjava_io_FileDescriptor *fdobj, char *buf, int nbytes, 
int flags) {
    return  sysReadFD(fdobj, buf, nbytes);
}


int sysSendFD(Classjava_io_FileDescriptor *fdobj, char *buf, int nbytes, int flags) {
    return sysWriteFD(fdobj, buf, nbytes);
}

long
sysSocketAvailableFD(Classjava_io_FileDescriptor * fdobj, long *pbytes)
{
    int fd = fdobj->fd - 1;
    sys_mon_t	*mon;
    int ret = 1;


    if (fd < 0) {
      Log1(1, "socket available operation on closed fd: %d?\n", fd);
      return 0;
    }

    mon = fdmon[fd];
    sysAssert(mon!=NULL);

    sysMonitorEnter(mon);     

    fd = fdobj->fd-1;
    if (fd < 0
	|| ioctl(fd, FIONREAD, pbytes) < 0) {
      ret = 0;
    } 

    sysMonitorExit(mon); 

    return ret;
}


/* Routines to do datagram */
ssize_t
sysSendtoFD(Classjava_io_FileDescriptor * fdobj, char *buf, int len, int flags, struct sockaddr *to, int tolen)
{
    int fd = fdobj->fd - 1;
    sys_mon_t	*mon;
    int retval=0;
 
    if (fd < 0) {
      Log1(1, "socket sendto operation on closed fd: %d?\n", fd);
      return retval;
    }

    mon = fdmon[fd];
    sysAssert(mon!=NULL);

    check_single_fd(fd);

    sysMonitorEnter(mon);
    fd = fdobj->fd-1;
    if (fd>=0) {
        fd_ref[fd]++;
	while(!pendingException() && !closed_fd(fd)) {
	  retval = sendto(fd, buf, len, flags, to, tolen);
	  if ((retval != -1) || pendingException() ||  
	      ((errno != EAGAIN) && (errno != EINTR)))
	    break;
	  if (errno == EAGAIN)
	    sysMonitorWait(mon, SYS_TIMEOUT_INFINITY);
	}
	fd_ref[fd]--;
	/* If I am the last person to leave and fd has been closed
	   call system close on the fd */
	if (fd_ref[fd]==0 && closed_fd(fd)) system_close(fd); 
    }

    Log2(1, "sendto done on fd: %d, retval %d\n", fd, retval);
    sysMonitorExit(mon);
    return retval;
}


/* TBD: WE ARE MAKING A COPY OF FD HERE - and not bothering to lock as
we do a poll and then dont do recvfrom unless we know that data is ready.
There is a potential race here - if another thread closes the fd after poll
and before recvfrom - I need to visit this later -pavani*/

ssize_t
sysRecvfromFD(Classjava_io_FileDescriptor * fdobj, char *buf, int nbytes, int flags, 
struct sockaddr *from, int *fromlen)
{
   
    sys_mon_t *mon; 
    register int nread = -1;
    struct pollfd polltbl[1];
    int fd = fdobj->fd-1;
    int status;

    if (fd < 0) {
      Log1(1, "Recfrom on closed fd: %d?\n", fd);
      return -1;
    }

    check_single_fd(fd);

    /* Since the recvfrom is not working in nonblocking manner,
       we poll on the fd until there is data. Poll succeeds
       when client sends data. So then do a blocking recvfrom */

    polltbl[0].fd = fd;
    polltbl[0].events = POLLIN;
    polltbl[0].revents = 0;
    if ((status = poll(polltbl, 1, -1)) <=0) {
	Log2(1, "polling before recvfrom failed with status %d, errno %d\n", 
	     status, errno);
    } else {
	nread = recvfrom(fd, buf, nbytes, flags, from, fromlen);
	Log2(1, "Recvfrom for %d bytes finished on fd: %d\n", nread, fd);
    }
    return nread;    
}





