/*
 * @(#)iomgr.c	1.55 95/11/29  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * I/O management, including non-blocking io.
 *
 * Eventually this module needs to have a formal interface, and
 * be replacable (e.g. it needs to be a porting layer).
 */

#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/fcntl.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/socket.h>
#include <errno.h>
#include <sys/filio.h>
#include <sys/ioctl.h>
#include <sys/uio.h>
#include <sys/syscall.h>
#include <limits_md.h>
#include <stdio.h>
#include <sys/resource.h>
#include <dlfcn.h>

#include <stropts.h>
#include <sys/conf.h>
#include <poll.h>

#include "threads.h"
#include "threads_md.h"
#include "monitor_md.h"
#include "itimer_md.h"
#include "clock.h"
#include "timeval.h"
#include "iomgr.h"
#include "bool.h"
#include "log.h"
#include "sys_api.h"

/*
 Pavani's Async IO notes:


 Because we have our own thread library we do not want
 the process to block. So code in here in conjuction with
 code in signals.c is used to provide wrapper to convert
 all the calls to be asynchronous io.

 For native share libraries making direct system io calls, 
 interposing wrappers are provided in this file. io_md.c contains 
 all the io methods that are called by the java.io package. SO if 
 you fix a bug in iomgr.c, make sure you look in io_md.c to see if
 a similar problem exists there.

 fd_flags is an array indexed by the file descriptor used
 to keep information about the state of the fd i.e. if it has 
 been made nonblocking or whether its marked to be closed.

 When fd is opened, the per fd monitors are allocated (i.e. if
 they have not been preallocated at init time. On the first
 read/write/read/writev call the fd is marked nonblocking.
 On close, the state for fd is cleared. The monitors are
 not freed as they can be reused for subsequesnt opens refering
 to same fd.

 After converting to the nonblocking mode, the io calls which return
 EAGAIN need a condition variable/monitor to wait on. We provide 
 a monitor per fd for this purpose.

 There is also a default async IO monitor which serves as condition 
 variable or monitor for calls like poll and select which deal
 with fd sets.

 Also a per fd reference count is maintained as to who is currently 
 in the io manager doing io or waiting for an io signal. Now on close
 the state of the fd is marked closed and if ref count is zero i.e. no
 one else is in middle of io, then only system_close is called on fd.
 If ref count is greater than zero, everyone waiting on the fd monitor
 is woken up and the last thread to leave the io manager closes the fd.

 Thus per fd monitor guards two structures: fd_flags and fd_ref 

 */

char *fd_flags;
int *fd_ref;

/*

 io_lock guards the state of the io control data structures:
 fdmon: per fd monitor
 fdMap: structure taht maps fd->index into poll table
 PollTable: used in poll call in the SIGIO signal handler 
            to decide whihc fd the io occured on.

 For more details look at comments in signals.c

*/
sys_mon_t _io_lock;
sys_mon_t **fdmon;


/*
The variable which records maximum number of files that
can be open at a time. Its initialized at init time and
is also the size of structures like fdmon_, fdMap and 
fd_flags.
*/

static int max_files = 0;
#define PREALLOCATE_FDS 16

extern void asyncIOActivateFD(int fd);
extern void asyncIODeactivateFD(int fd);

#define IO_OPTIMIZE

syscall_t systable[] = {
	{ "write", 0 },
	{ "writev", 0 },
	{ "read", 0 },
	{ "readv", 0 },
	{ "getmsg", 0 },
	{ "putmsg", 0 },
	{ "poll", 0 },
	{ "open", 0 },
	{ "close", 0 },
	{ "pipe", 0 },
	{ "fcntl", 0},
	{ "dup", 0 },
	{ "creat", 0 },
	{ (char *) 0, 0 }
};

extern int errno;

/* This isn't in syscall.h, for some reason. */
int syscall (int, ...);
int _select (int width, fd_set *, fd_set *, fd_set *, struct timeval *);

/* macro to copy fd_set. fd_set is an array of fd_mask */
/* NFDBITS is defined in select.h to be bits per fd_mask */
#define COPY_SET(_to, _from, _width)              \
        memcpy(_to, _from,                        \
               ((_width + NFDBITS-1) / (NFDBITS)) \
                * sizeof(fd_mask)                 \
               );

static void
initSyscalls()
{
    syscall_t *tp = systable;
    
    while (tp->sym != (char *) 0) {
	tp->addr = (sysfunc_t) dlsym(RTLD_NEXT, tp->sym);
	tp++;
    }
}

/*
 * Returns non-zero if an exception is pending for this ee.
 */
int pendingException() { 
    struct execenv *ee = EE(); 
    return (ee && exceptionOccurred(ee));
}

boolean_t
initialize_monitors(int desc)
{
    if (fdmon[desc] == 0) {
    	if ((fdmon[desc] =
		(sys_mon_t *) calloc(1, sizeof(sys_mon_t)))==NULL) {
	    return B_FALSE;
	}
	sysMonitorInit(fdmon[desc], FALSE);
	
    }

    return B_TRUE;
}

/* External routine:
   DO all the necessary nonblocking io initialization.
   Called at startup time from javai main routine.
*/

void
InitializeAsyncIO()
{
    extern void InitializeAsyncMonitors(int);
    struct rlimit nbr_files;
    int i;

    /* set the file descriptor limit to OPEN_MAX_FD */
    getrlimit(RLIMIT_NOFILE, &nbr_files);
    max_files = nbr_files.rlim_cur;
    nbr_files.rlim_cur =
	OPEN_MAX_FD<nbr_files.rlim_max?OPEN_MAX_FD:nbr_files.rlim_max;
    if (!setrlimit(RLIMIT_NOFILE, &nbr_files)) {
	max_files = nbr_files.rlim_cur;
    }

    InitializeAsyncMonitors(max_files);
    IO_LOCK_INIT();  

    fdmon = (sys_mon_t **) calloc(max_files, sizeof(sys_mon_t *));
    if (fdmon == 0) {
	out_of_memory();		/* fatal error */
    }

    for (i=0; i< PREALLOCATE_FDS; ++i) {
	if (initialize_monitors(i) == B_FALSE) {
	    out_of_memory();		/* fatal error */
	}
    }

    fd_flags = (char *) calloc(max_files, sizeof(char));
    fd_ref = (int *) calloc(max_files, sizeof(int));
    if ((fd_flags == 0) || (fd_ref == 0)) {
	out_of_memory();		/* fatal error */
    }

    /* If the fd is stdin, stderr or stdout we do not want to
       make the fd nonblocking otherwise when you exit java
       with ^c the shell in which you were running java chokes.
       Basically shells cannot deal with nonblocking stdin,stderr,stdout.
       Registering an atexit handler to make fds blocking on exit does
       not help in case java terminates through signals - so its 
       best to leave these fds alone
       So we leave fds 0, 1, 2 falsely marking them already nonblocking,
       so tha tin future check_single_fd routine will never call nonblocking_io
       routine for these fds. */

       fd_flags[0] = FD_NBINIT; 
       fd_flags[1] = FD_NBINIT; 
       fd_flags[2] = FD_NBINIT;

    initSyscalls();
    return;
}


/*
 *  If the second argument is true the fd is put into non-blocking
 *  mode.  That is,  a read when there is no data will not stop
 *  the entire heavyweight process.
 *  Frankly I can't imagine a situation where blocking IO is what is
 *  desired.  Should all fd's opened in a multithreaded environment be
 *  nonblocking? -csw
 *
 * The routine basically makes the fd blocking/noblocking as desired.
 * Blocing io is currently being used at least one place - makeTTYSane
 * routine. - pavani
 */

int
nonblock_io(int desc, int onoff)
{
    long flgs;
 
    sysAssert(desc < max_files);
    sysAssert(fdmon[desc] != NULL);

    sysMonitorEnter(fdmon[desc]); 
    IO_LOCK(); /* Activate/Deactivate routine need this lock held */

    /* If you donot want to block and fd is nonblocking - do nothing
       otherwise make fd nonbloking */
    if ((onoff == IO_DONTBLOCK) && !(fd_flags[desc] & FD_NBINIT)) {
	/* if it's already non-blocking, do nothing... */
	if ((flgs = fcntl(desc, F_GETFL, 0)) == -1)
	    {  IO_UNLOCK(); sysMonitorExit(fdmon[desc]);  return -1;}

	fcntl(desc, F_SETFL, flgs | O_NONBLOCK);
	ioctl(desc, I_SETSIG, S_INPUT | S_OUTPUT);


	asyncIOActivateFD(desc);
	fd_flags[desc] |= FD_NBINIT;

	Log1(1, "Made %d nonblocking\n", desc);

    } else if ((onoff == IO_BLOCK) && (fd_flags[desc] & FD_NBINIT)) {
 	/* if it's already blocking, do nothing... 
	   otherwise make fd blocking*/

	if ((flgs = fcntl(desc, F_GETFL, 0)) == -1)
	    { IO_UNLOCK(); sysMonitorExit(fdmon[desc]);  return -1; }

	fcntl(desc, F_SETFL, flgs & ~O_NONBLOCK);
	ioctl(desc, I_SETSIG, 0);

	fd_flags[desc] &= ~FD_NBINIT;

	Log1(1, "Made %d blocking\n", desc);

    }
    sysMonitorExit(fdmon[desc]);
    IO_UNLOCK(); 
    return 0;
}

/* Hold the per fd lock for the duration of system close */
int
system_close(int fd) {
int ret = -1;

    IO_LOCK();
    if (fd_flags[fd] & FD_NBINIT) { 
        fd_flags[fd] = 0; 
        asyncIODeactivateFD(fd);
    }

    ret = (*systable[SYS_CLOSE].addr)(fd);
    Log1(1, "Closed fd: %d\n", fd);

    IO_UNLOCK();
    return ret;
}


int
close(int fd)
{
    int ret = -1;

    /* We seem to call close on fd multiple times for
       some reason - to be fixed in upper layer */
    if (fd <= 0) {
        Log1(1, "Closing bad fd: %d?\n", fd);
        return ret;
    }

    sysMonitorEnter(fdmon[fd]);

    if (fd_ref[fd] == 0) {
      if (!closed_fd(fd)) 
          ret = system_close(fd);
    } else {
      if (!closed_fd(fd)) {
	fd_flags[fd] |= FD_CLOSED;
	sysMonitorNotifyAll(fdmon[fd]);
	sysMonitorNotifyAll(asyncMon(SYS_ASYNC_MON_IO));
	Log1(1, "Marked Closed fd: %d\n", fd);
      }
      ret = 0;
    }

    sysMonitorExit(fdmon[fd]);

    return ret;
}


/* This routine does nto queue up the write request but
   instead returns the corresponding error back.
   Seems to be used at only one place: by the audio.c
*/
int
write_noblock(int fd, caddr_t buf, int nbytes)
{
    register int nwrote;

    if (nbytes < 0)
	return nbytes;
    check_single_fd(fd);
    while ((nwrote = (*systable[SYS_WRITE].addr)(fd, buf, nbytes)) < 0) {
	if (errno == EAGAIN)
	    return 0;
	if (errno != EINTR)
	    return -1;
    }
    return nwrote;
}

ssize_t
recv(int fd, char *buf, int nbytes, int flags)
{
    return read(fd, (void *) buf, (uint) nbytes);
}


ssize_t
send(int fd, const char *buf, int nbytes, int flags)
{
    return write(fd, buf, (uint) nbytes);
}

/* routines to do datagram */
ssize_t
syssendto(int fd, char *buf, int len, int flags, struct sockaddr *to, int tolen)
{
    int retval;
    sys_mon_t *mon = fdmon[fd];

    sysAssert(mon!=NULL);

    check_single_fd(fd);

    sysMonitorEnter(mon);
    while(1) {
      retval = sendto(fd, buf, len, flags, to, tolen);
      if ((retval != -1) || pendingException() ||  
	  ((errno != EAGAIN) && (errno != EINTR)))
	  break;
	if (errno == EAGAIN)
	    sysMonitorWait(mon, SYS_TIMEOUT_INFINITY);
    }

    Log2(1, "sendto done on fd: %d, retval %d\n", fd, retval);
    sysMonitorExit(mon);
    return retval;
}


ssize_t
sysrecvfrom(int fd, char *buf, int nbytes, int flags, struct sockaddr *from, int *fromlen)
{
    register int nread = -1;
    struct pollfd polltbl[1];
    int status;

    check_single_fd(fd);

    /* Since the recvfrom is not working in nonblocking manner,
       we poll on the fd until there is data. Poll succeeds
       when client sends data. So then do a blocking recvfrom */

    polltbl[0].fd = fd;
    polltbl[0].events = POLLIN;
    polltbl[0].revents = 0;
    if ((status = poll(polltbl, 1, -1)) <=0) {
	Log2(1, "polling before recvfrom failed with status %d, errno %d\n", 
	     status, errno);
    } else {
	nread = recvfrom(fd, buf, nbytes, flags, from, fromlen);
	Log2(1, "Recvfrom for %d bytes finished on fd: %d\n", nread, fd);
    }
    return nread;    
}


ssize_t
read(int fd, void *buf, unsigned int nbytes)
{
    register int nread;
    sys_mon_t *mon = fdmon[fd];

    sysAssert(mon!=NULL);

    check_single_fd(fd);

#ifdef IO_OPTIMIZE
    nread = (*systable[SYS_READ].addr)(fd, buf, nbytes);
    if ((nread != -1) || pendingException() ||  
	(errno != EAGAIN && errno != EINTR)) {
      Log2(1, "Read for %d bytes finished (did not block) on fd: %d\n", \
	   nread, fd);
      return nread;
    }
#endif

    sysMonitorEnter(mon);
    fd_ref[fd]++;

    while(!pendingException() && !closed_fd(fd)) {
      nread = (*systable[SYS_READ].addr)(fd, buf, nbytes);
      if ((nread != -1) || pendingException() ||  
	  ((errno != EAGAIN) && (errno != EINTR)))
	  break;
      if (errno == EAGAIN) 
	  sysMonitorWait(mon, READ_TIMEOUT);
    }

    fd_ref[fd]--;
    /* If I am the last person to leave and fd has been closed
       call system close on the fd */
    if (fd_ref[fd]==0 && closed_fd(fd)) system_close(fd); 

    Log2(1, "Read for %d bytes finished on fd: %d\n", nread, fd);
    sysMonitorExit(mon);
    return nread;
}


ssize_t
write(int fd, const void *buf, unsigned int nbytes)
{
    register int nwrote;
    register caddr_t next = (caddr_t)buf;
    register int total_written = 0;
    sys_mon_t	*mon = fdmon[fd];

    sysAssert(mon!=NULL);

    check_single_fd(fd);

#ifdef IO_OPTIMIZE
    while (total_written < nbytes && !pendingException()) {
	nwrote = (*systable[SYS_WRITE].addr)(fd,
			 next + total_written,
			 nbytes - total_written);
	if (nwrote == -1)  break;
	total_written += nwrote;
    }	

    /* If done - return; basically whole write succeeded immidiately */
    /* If error was other than EINTR and EAGAIN - return immidiately */
    /* if there was a pending exception return immidiately */
    if ((nwrote!=-1 && (total_written == nbytes)) ||
       ((nwrote==-1) && (errno != EAGAIN && errno != EINTR)) ||
	pendingException()) {
      Log2(1, "Write for %d bytes finished (did not block) on fd: %d\n", \
	   nwrote, fd);
       return ((nwrote!=-1)?total_written:-1);
     }
   /* else take the slow path */
#endif

    sysMonitorEnter(mon); 
    fd_ref[fd]++;

    while(!pendingException() && !closed_fd(fd)) {
	nwrote = (*systable[SYS_WRITE].addr)(fd,
			 next + total_written,
			 nbytes - total_written);
	if ((nwrote==-1) && (errno != EAGAIN && errno != EINTR)) {
         	total_written = -1;
		break;
        }
	if (nwrote==-1) { 
	  /* with errno== EAGAIN or EINTR */
	  /* TBD: WRITE_TIMEOUT should be replaced by SYS_TIMEOUT_INIFINITY */
	  if (errno == EAGAIN)
	    sysMonitorWait(mon, WRITE_TIMEOUT);
	  continue;
	}
        total_written += nwrote;
        if (total_written >= nbytes) break;
    } /* end of while */

    fd_ref[fd]--;
    /* If I am the last person to leave and fd has been closed
       call system close on the fd */
    if (fd_ref[fd]==0 && closed_fd(fd)) system_close(fd); 

    Log2(1, "Write for %d bytes finished on fd: %d\n", nwrote, fd);
    sysMonitorExit(mon);

    return total_written;
}



/* AVH -- added the writev system call, it is used
 *        by some libraries like X11.
 */

int
writev(int fd, const struct iovec *iovec, int iovcnt)
{
    int total_written = 0;
    int nwrote, ndone = 0;
    struct iovec *iov = (struct iovec *)iovec;
    sys_mon_t *mon = fdmon[fd];

    sysAssert(mon!=NULL);

    check_single_fd(fd);

    sysMonitorEnter(mon);
    fd_ref[fd]++;

    while (iovcnt > 0 && !pendingException()  && !closed_fd(fd)) {
	iov->iov_len -= ndone;
	iov->iov_base += ndone;
	nwrote = (*systable[SYS_WRITEV].addr)(fd, iov, iovcnt);

	while (nwrote == -1) {
	    if (errno == EAGAIN) {
	    /* TBD: WRITE_TIMEOUT should be replaced by 
	       SYS_TIMEOUT_INIFINITY */
		sysMonitorWait(mon, WRITE_TIMEOUT);
	    } else if (errno != EINTR) {
		total_written = -1;
		goto unlock;
	    }
	    nwrote = (*systable[SYS_WRITEV].addr)(fd, iov, iovcnt);
	}
	total_written += nwrote;
	iov->iov_len += ndone;
	iov->iov_base -= ndone;
	ndone += nwrote;
	for (; (iovcnt > 0) && (ndone >= iov->iov_len) ; iovcnt--, iov++) {
	    ndone -= iov->iov_len;
	}
    }
unlock:
    fd_ref[fd]--;
    /* If I am the last person to leave and fd has been closed
       call system close on the fd */
    if (fd_ref[fd]==0 && closed_fd(fd)) system_close(fd); 

    Log2(1, "Writev for %d bytes finished on fd: %d\n", total_written, fd);
    sysMonitorExit(mon);
    return total_written;
}

/* AVH -- added the readv system call, it is used
 *        by some libraries like X11.
 */
/* TBD: optimize like read routine - pavani */
int
readv(int fd, struct iovec *iov, int iovcnt)
{
    int nread;
    sys_mon_t *mon = fdmon[fd];

    sysAssert(mon!=NULL);

    check_single_fd(fd);

    sysMonitorEnter(mon);
    fd_ref[fd]++;

    while(!pendingException() && !closed_fd(fd)) {
      nread = (*systable[SYS_READV].addr)(fd, iov, iovcnt);
      if ((nread != -1) || pendingException() ||  
	  ((errno != EAGAIN) && (errno != EINTR)))
	  break;
	if (errno == EAGAIN) {
	    sysMonitorWait(mon, READ_TIMEOUT);
	}
    }

    fd_ref[fd]--;
    /* If I am the last person to leave and fd has been closed
       call system close on the fd */
    if (fd_ref[fd]==0 && closed_fd(fd)) system_close(fd); 

    Log2(1, "Readv for %d bytes finished on fd: %d\n", nread, fd);
    sysMonitorExit(mon);
    return nread;
}

/*
 we do not initialize the fd to be nonblocking for getmsg and 
 putmsg i.e. no calls to check_single_fd - due to a bug
 in solaris 2.3 or so I am told by jpayne. I need to confirm
 and revisit later - but for now this works as before - pavani. 
*/
int
getmsg(int fd, struct strbuf *ctlptr, struct strbuf *dataptr, int *flags)
{
    int retval;
    sys_mon_t *mon = fdmon[fd];

    sysAssert(mon!=NULL);

    /* check_single_fd(fd); */

    sysMonitorEnter(mon);
 
    fd_ref[fd]++;

    while(!pendingException() && !closed_fd(fd)) {
      retval = (*systable[SYS_GETMSG].addr)(fd, ctlptr, dataptr, flags);
      if ((retval != -1) || pendingException() ||  
	  ((errno != EAGAIN) && (errno != EINTR)))
	  break;
      if (errno == EAGAIN)
	  sysMonitorWait(mon, SYS_TIMEOUT_INFINITY);
    }

    fd_ref[fd]--;
    /* If I am the last person to leave and fd has been closed
       call system close on the fd */
    if (fd_ref[fd]==0 && closed_fd(fd)) system_close(fd); 

    Log1(1, "getmsg on fd: %d\n", fd);
    sysMonitorExit(mon);
    return retval;
}

int
putmsg(int fd, const struct strbuf *ctlptr,
       const struct strbuf *dataptr, int flags)
{
    int retval;
    sys_mon_t *mon = fdmon[fd];

    sysAssert(mon!=NULL);

    /* check_single_fd(fd); */

    sysMonitorEnter(mon);
    fd_ref[fd]++;

    while(!pendingException() && !closed_fd(fd)) {
      retval = (*systable[SYS_PUTMSG].addr)(fd, ctlptr, dataptr, flags);
      if ((retval != -1) || pendingException() ||  
	  ((errno != EAGAIN) && (errno != EINTR)))
	  break;
	if (errno == EAGAIN)
	    sysMonitorWait(mon, SYS_TIMEOUT_INFINITY);
    }

    fd_ref[fd]--;
    /* If I am the last person to leave and fd has been closed
       call system close on the fd */
    if (fd_ref[fd]==0 && closed_fd(fd)) system_close(fd); 

    Log1(1, "putmsg done on fd: %d\n", fd);
    sysMonitorExit(mon);
    return retval;
}


/*
 * On SunOS 2.3 the connect routine (really a library call) can
 * sometimes return an error and errno == EPROTO.  in this case it
 * is pointless to restart the connect call as it will just fail
 * again with EPROTO.  I don't know what causes this but it seem
 * like for now that the only way to fix the problem is to shut the
 * socket down and open it up again.  We report the error to Java and
 * it can try to restart the connection.  Also, if the connection is
 * refused, it seems that select() returns 1, as if everything is
 * OK, but there's no way to find out that the connection failed
 * until you try something!  So, now we try something immediately,
 * and see what we get.  What this boils down to is socket "system
 * calls" do not work when the socket is in nonblocking I/O mode.
 */

int
sysConnect(int s, struct sockaddr *him, int len)
{
    int	   result;
    int	   delay = 50;	/* sleep 50 ms and try again */
    fd_set write_fds;
    struct timeval  poll;
    extern int _select (int width, fd_set *, fd_set *, fd_set *, struct timeval *);

    FD_ZERO(&write_fds);
    FD_SET(s, &write_fds);
    poll.tv_sec = poll.tv_usec = 0;

    sysMonitorEnter(asyncMon(SYS_ASYNC_MON_IO));
    result = connect(s, him, len);
    while (result < 0 && !pendingException()) {
	if (errno == EINPROGRESS) {
	    int	cnt = _select(s+1, 0, &write_fds, 0, &poll);

	    FD_SET(s, &write_fds);
	    if (cnt == 1) {
		int bytes;

		if (ioctl(s, FIONREAD, &bytes) == -1) {
		    if (errno == EPIPE) {
			errno = ECONNREFUSED;
		    }
		    /* If this fails with EPIPE, then we assume
		       connection is refused - who knows what's
		       right?  Not I. */
		    sysAssert(result < 0);
		    break;
		}
		result = 0;
		break;
	    }
	    sysMonitorWait(asyncMon(SYS_ASYNC_MON_IO), delay);
	} else {
	    break;
	}
    }
    sysMonitorExit(asyncMon(SYS_ASYNC_MON_IO));
    return result;
}


/* Poll system call */

int
poll(struct pollfd *fds, unsigned long nfds, int timeout)
{
    timeval_t time_limit;
    int	    poll_timeout = 0;
    int	    retval;
    int	    i, fd;

    for (i = nfds; --i >= 0; ) {
	if ((fd = fds[i].fd) < max_files) check_single_fd(fd);
    }

    if (timeout > 0) {
	timeval_t tv;
	tv.tv_sec = timeout / 1000;
	tv.tv_usec = timeout % 1000;
	time_limit = currentTime();
	timerAdd(&time_limit, (timeval_t *) &tv);
    }

    sysMonitorEnter(asyncMon(SYS_ASYNC_MON_IO));
    while (!pendingException()) {
	timeval_t delay;

	errno = 0;
	retval = (*systable[SYS_POLL].addr)(fds, nfds, poll_timeout);

	if (retval == -1)
	    Log1(1, "poll => -1; errno = %d\n", errno);

	if (retval != 0)
	    break;
	if (timeout > 0) {
	    /* would have blocked, see whether timer has expired */
	    timeval_t now = currentTime();

	    if (timerCmp(&now, &time_limit, >) || timerEq(&now, &time_limit)) {
		retval = 0;
		break;
	    }
	    delay = time_limit;
	    timerSub(&delay, &now);
	    timeout = (delay.tv_sec * 1000) + (delay.tv_usec / 1000);
	}
	sysMonitorWait(asyncMon(SYS_ASYNC_MON_IO), timeout > 0 ? timeout : SYS_TIMEOUT_INFINITY);
    }
    sysMonitorExit(asyncMon(SYS_ASYNC_MON_IO));
    return retval;
}

/* Select system call in a multi-threaded environment.

   This works by affecting a poll.  If _select returns
   non-negative, then we return that value.  Otherwise, we block
   until *any* IO is ready, and then try again.  If we still come
   up empty then we block for a smaller amount of time this time.
   If we reach the time limit, then we just return 0, as select
   normally would. */

int
select(int width, fd_set *readfds, fd_set *writefds,
       fd_set *exceptfds, struct timeval *tv)
{
    struct timeval  poll;
    int             nfds = -1;
    timeval_t       time_limit;
    fd_set          reads, writes, excepts;
    int		    i;

    for (i = 0; i < width; i++) {
	if (i >= OPEN_MAX)
	    break;
	if ((readfds != 0 && FD_ISSET(i, readfds))
	    || (writefds != 0 && FD_ISSET(i, writefds))
	    || (exceptfds != 0 && FD_ISSET(i, exceptfds))) {
	    check_single_fd(i);
	}
    }

    poll.tv_sec = poll.tv_usec = 0;
    if (tv != 0) {
	time_limit = currentTime();
	timerAdd(&time_limit, (timeval_t*)tv);
    }

    
    if (readfds != 0)
	COPY_SET(&reads, readfds, width);
    if (writefds != 0)
	COPY_SET(&writes, writefds, width);
    if (exceptfds != 0)
	COPY_SET(&excepts, exceptfds, width);

    sysMonitorEnter(asyncMon(SYS_ASYNC_MON_IO));
    while (!pendingException()) {
	timeval_t delay;
	int timeout = SYS_TIMEOUT_INFINITY;

	nfds = _select(width, readfds, writefds, exceptfds, &poll);

	if (nfds != 0) {
	    break;
	}
	if (tv != 0) {
	    timeval_t now = currentTime();

	    if (timerCmp(&now, &time_limit, >) || timerEq(&now, &time_limit)) {
		nfds = 0;
		break;
	    }
	    delay = time_limit;
	    timerSub(&delay, &now);
	    timeout = (delay.tv_sec * 1000) + (delay.tv_usec / 1000);
	}
	sysMonitorWait(asyncMon(SYS_ASYNC_MON_IO), timeout);
	if (readfds != 0)
	    COPY_SET(readfds, &reads, width);
	if (writefds != 0)
	    COPY_SET(writefds, &writes, width);
	if (exceptfds != 0)
	    COPY_SET(exceptfds, &excepts, width);
    }
    sysMonitorExit(asyncMon(SYS_ASYNC_MON_IO));
    return nfds;
}

/*
 We need to override every call through which we can get a new file descriptor
and make sure we are initializing the monitors for the new fds before we return 
control. Known calls that are already taken care of are: open, fcntl, pipe
dup and creat */

/*
 * Although we set the SA_RESTART flag on all our signal handlers we
 * still see open() occasionally return with errno == EINTR, at
 * least on SunOS 2.3.  This routine is an attempt to get aorund
 * this bug by making sure open doesnot return until it either
 * succeeds or errno != EINTR.
 */

int
open(const char *fname, int flags, ...)
{
    va_list ap;
    register int fd;
    unsigned int mode;
    boolean_t ret = B_TRUE;

    IO_LOCK();
    va_start(ap, flags);
    if (flags & O_CREAT) {
	mode = va_arg(ap, unsigned int);
    } else {
	mode = 0;
    }
    va_end(ap);
    while(1) {
	fd = (*systable[SYS_OPEN].addr)(fname, flags, mode);
	if ((fd != -1) || pendingException() ||  
	    ((errno != EAGAIN) && (errno != EINTR)))
	    break;      
    }
 
    if (fd != -1) {
	Log1(1, "Opened fd: %d\n", fd);
	ret = initialize_monitors(fd);
    } else {
	Log1(1, "open error: %d\n", errno);
    }

    IO_UNLOCK();

    sysAssert(fd != 0 && ret != B_FALSE);
    return fd;
}


/*
 * Although we set the SA_RESTART flag on all our signal handlers we
 * still see open() occasionally return with errno == EINTR, at
 * least on SunOS 2.3.  This routine is an attempt to get aorund
 * this bug by making sure open doesnot return until it either
 * succeeds or errno != EINTR.
 */
int
pipe(int fds[])
{
    long long sys_ret = 0;
    int ret;

    IO_LOCK();

    while(1) {
	/*
	 * As unbelievable as it may seem, you cannot simply simulate
	 * the pipe() system call with a simple
	 * syscall(SYS_pipe,...).  SYS_pipe returns two values in
	 * %o0 and %o1, effectively returning a "long long", but the
	 * prototype for syscall only returns an int!
	 */
	sys_ret = (*systable[SYS_PIPE].addr)(fds);
	  if ((sys_ret != -1) || pendingException() ||  
	      ((errno != EAGAIN) && (errno != EINTR)))
	      break;      
    }
 
    if (sys_ret != -1) {
	Log2(1, "Opened pipe fd0: %d fd1: %d\n", fds[0], fds[1]);
	ret = initialize_monitors(fds[0]);
	ret = (ret == B_FALSE ? ret : initialize_monitors(fds[1]));
    } else {
	Log1(1, "pipe error: %d\n", errno);
    }

    IO_UNLOCK();

    sysAssert(sys_ret == 0 && ret != B_FALSE);
    return sys_ret;
}

/* The following wrapper is necessary because nis+ library calls dup,
as a means of getting new fds - and we need to initialize monitors
for the newly opened fds before they can be used.
Note dup2 call uses fcntl internally - so the following wrapper
takes care of duplicating fds through fcntl or dup2
*/
int fcntl (int fildes, int cmd, ...)
{
    register int fd;
    va_list ap;
    int ret;
    int arg;  /* optional third argument */

    va_start(ap, cmd);
    arg = va_arg(ap, int);
    IO_LOCK();
    while (1) {
	fd = (*systable[SYS_FCNTL].addr)(fildes, cmd, arg);
	if ((fd != -1) || pendingException() ||  
	      ((errno != EAGAIN) && (errno != EINTR)))
	      break;      
    }
    va_end(ap);

    if (cmd == F_DUPFD) {
	if (fd != -1) {

	    Log1(1, "fcntl dup opened fd: %d\n", fd);
	    ret = initialize_monitors(fd);
	    sysAssert(ret != B_FALSE);
	} else {
	    Log1(1, "fcntl dup error: %d\n", errno);
	}
    }
    IO_UNLOCK();
    return fd;
}

int dup(int fildes)
{
int fd;
int ret;

  IO_LOCK();
  while (1) {
      fd = (*systable[SYS_DUP].addr)(fildes);
      if ((fd != -1) || pendingException() ||  
	  ((errno != EAGAIN) && (errno != EINTR)))
	  break;      
  }
  if (fd != -1) {
      Log2(1, "Dup fd: %d to newfd: %d\n", fildes, fd);
      ret = initialize_monitors(fd);
  } else {
      Log1(1, "Dup error: %d\n", errno);
  }

  IO_UNLOCK();
  return fd;
}

int creat(const char *path, mode_t mode)
{
int fd;
int ret;

  IO_LOCK();
  while (1) {
      fd = (*systable[SYS_CREAT].addr)(path, mode);
      if ((fd != -1) || pendingException() ||  
	  ((errno != EAGAIN) && (errno != EINTR)))
	  break;      
  }
  if (fd != -1) {
      Log1(1, "Creat fd: %d \n", fd);
      ret = initialize_monitors(fd);
  } else {
      Log1(1, "Creat error: %d\n", errno);
  }

  IO_UNLOCK();
  return fd;
}

