/*
 * @(#)signals.c	1.39 95/12/01  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * A spl and interrupt implementation to run on top of Unix.
 */

/*
 * Header files.
 */
#include <ucontext.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>

#include "sys_api.h"
#include "threads_md.h"
#include "monitor_md.h"
#include "interrupt.h"
#include "ucontext.h"
#include "log.h"
#include "itimer_md.h"
#include "schedule.h"
#include "monitor.h" 		/* For monitorRegister() */
#include "iomgr.h"      	/* For io lock */

void out_of_memory(void);

/*
 * Local variables.
 */
static sys_mon_t	*handlerMonitors[SYS_ASYNC_MON_MAX];

sys_mon_t *PendingNotifyQ;

/*
 * Local routine prototypes.
 */

/*
 * intrNotifyHandler() --
 *	Interrim routine that delivers signals in BKM's original fashion.
 */
static int intrNotifyHandler(int interrupt, void *arg, void *siginfo);


#include <limits_md.h>
#include <poll.h>

/* The polling fd table.  This table is contains all the file
   descriptors which are currently attempting i/o but have gotten
   EAGAIN indicating the operation would block.  When a sigio comes
   in, this polling structure is used to figure out which file
   descriptor(s) is ready to do i/o.  This might seem like overkill,
   but for applications which have 50 or 100 or more open files, e.g.,
   network connections, the old way of waking up *all* the threads
   blocked in i/o is just too slow!  Imagine 100 threads attempting to
   reread their connection before the right one is tried.  Way too
   slow.

   This array is always packed.  That is, there are no holes in it. 
   The goal is to pass the minimum amount of fd's to the kernel for
   examination, and to only pass valid fds.  Hopefully this isn't more
   work than it's worth. */

static struct pollfd *pollTable;

/* This is the number of fds to examine in fdsToPoll */
static int	fdCount = -1;

/* FdMap is indexed by file descriptor.  It maps from a file
   descriptor to a position in the pollTable.  A pollOffset of -1
   indicates fd is not allocated into pollTable. */

static struct FdEntry {
    short	pollOffset;	/* offset into the pollTable */
    short	flags;		/* bits POLLIN and POLLOUT */
} *fdMap;

/*
 * External routines.
 */

static void asyncIORegister(void);

void
InitializeAsyncMonitors(int max_files)
{
    int i;

    /* Async monitor numbering starts at 1 */
    for (i = 1; i < SYS_ASYNC_MON_MAX; i++) {
	handlerMonitors[i] = (sys_mon_t *) malloc(sysMonitorSizeof());
	if (handlerMonitors[i] == 0) {
	    out_of_memory();		/* fatal error */
	}
	memset((char *) handlerMonitors[i], 0, sysMonitorSizeof());
	switch (i) {	/* Have to keep in sync with the enum... */
	    case SYS_ASYNC_MON_ALARM:
		monitorRegister(handlerMonitors[i], "Alarm monitor");
		break;
	    case SYS_ASYNC_MON_IO:
		monitorRegister(handlerMonitors[i], "I/O monitor");
		break;
	    case SYS_ASYNC_MON_EVENT:
		monitorRegister(handlerMonitors[i], "Event monitor");
		break;
	    case SYS_ASYNC_MON_CHILD:
		monitorRegister(handlerMonitors[i], "Child death monitor");
		break;
	    default:
		monitorRegister(handlerMonitors[i], "Unknown monitor");
		break;
	}
    }

    asyncIORegister();

    if (fdCount == -1) {
        struct FdEntry *entry;
        int slot;

	Log1(1, "Initializing fdmap enteries = %d\n", max_files);
	fdCount = 0;
	slot = max_files;
        fdMap = (struct FdEntry *)calloc(max_files, sizeof(struct FdEntry));
	if (fdMap == 0) {
	    out_of_memory();		/* fatal error */
	}
	for (entry = fdMap; --slot >= 0; entry += 1) {
	    entry->pollOffset = -1;
	    entry->flags = 0;
	}
        pollTable = (struct pollfd *)calloc(max_files, sizeof(struct pollfd));
	if (pollTable == 0) {
	    out_of_memory();
	}
    }
}

/*
 * Thread all outstanding CV notification.  Don't mess with the currently
 * running context, because we have no idea what state we are in.
 */
static int
processPendingNotification()
{
    sys_mon_t *mid;
    sys_mon_t *midq;
    int need_to_reschedule = 0;

    /*
     * Perform the notifyall's that couldn't be done because of locking.
     * Pull all of the pending CV notifies off of the queue, and process
     * them.  The notification is completely normal, except for the locking.
     * If a yield is necessary, it is done by the caller.
     */
    sysAssert(PendingNotifyQ);
    midq = PendingNotifyQ;
    while (midq) {
	mid = midq;
	midq = midq->pendingq;
	mid->pendingq = SYS_MID_NULL;
	mid->flags &= ~SYS_MON_PENDING_NOTIFICATION;

	/* See asyncEventNotify for explanation of this. */
	if (mid->monitor_owner != SYS_THREAD_NULL) {
	    mid->flags |= SYS_MON_STICKY_NOTIFICATION;
	}

	/* Now notify all of the waiters. */
	queueBroadcast(mid);

	/* If the monitor isn't locked, give it to the first waiter */
	if (!mid->monitor_owner && mid->monitor_waitq) {
	    need_to_reschedule = queueSignal(mid, &mid->monitor_waitq);
	}
    }
    PendingNotifyQ = SYS_MID_NULL;
    return need_to_reschedule;
}

/*
 * This function is used to return to any context that was interrupted
 * by a signal delivery.  We assume that the signal trampoline has saved
 * all state, and all we need to do is execute a signal cleanup (the
 * backend of the signal trampoline) to return control to the interrup-
 * ted thread.
 */
void
fullSwitchContext(context_t *c)
{
    /*
     * This thread was interrupted, which means that it was NOT inside
     * a critical section.  This also means that the thread does not know
     * that it needs to clear the critical section lock when control
     * returns to it.  We have to clear the lock for it, and there may
     * be pending interrupts.
     */

    /*
     * If the scheduler was interrupted, we process the events, and call
     * it with interrupts disabled.  This prevents recursion.
     */
    /* intrLock(); - removed by csw */
    if (PendingNotifyQ) {
	if (processPendingNotification()) {
	    sys_thread_t    *self = sysThreadSelf();

	    if (self->state == RUNNABLE) {
		queueInsert(&runnable_queue, self);
	    }
	    reschedule();
	}
    }
    /* intrUnlock(); - removed by csw */

    /*
     * Don't have to call return_from_trap instead of switch_context
     * if preemption is set because we are continuing the thread that
     * was interrupted when the signal came in and the signal trampoline
     * will restore the original signal mask.
     */
    sysThreadSelf()->full_switch = 0;
    switchContext(c);
    /* NOTREACHED */
}

/*
 * return a registered asynchronous notifier
 */
sys_mon_t *
asyncMon(async_mon_key_t key)
{
    if (key >= SYS_ASYNC_MON_MAX || key == 0) {
	return SYS_MID_NULL;
    } else {
	return handlerMonitors[key];
    }
}

static int asyncKeyToSig[SYS_ASYNC_MON_MAX] = {
        0,		/* Placeholder.  async_mon_key_t starts at 1 */
	SIGALRM,	/* SYS_ASYNC_MON_ALARM */
	SIGIO,		/* SYS_ASYNC_MON_IO */
	SIGUSR1,	/* SYS_ASYNC_MON_EVENT */
	SIGCHLD,	/* SYS_ASYNC_MON_CHILD */
};

/*
 * This interface is used by client programs to register monitors that
 * they will recieve asynchronous notification on.  The async monitor key
 * is translated into a machine dependent Unix signal number.
 */
int
asyncNotifier(async_action_t action, async_mon_key_t key)
{
    int sig;
    sys_mon_t *mid;

    if (key >= SYS_ASYNC_MON_MAX) {
	return -1;
    }

    sig = asyncKeyToSig[key];
    mid = handlerMonitors[key];
    memset((void *)mid, 0, sizeof (*mid));

    if (action == ASYNC_REGISTER) {
	mid->flags |= SYS_MON_HAS_HANDLER;
	/* Mustn't be a cache monitor: see monitor_md.h */
	sysAssert((mid->flags & SYS_MON_IN_CACHE) == 0);

	/* Register our interrrupt handler */
	intrRegister(sig, intrNotifyHandler, handlerMonitors[key]);
    } else {
	intrUnregister(sig, intrNotifyHandler, handlerMonitors[key]);
    }

    return 0;
}

void
asyncIOActivateFD(int fd)
{
    struct FdEntry *entry;
    struct pollfd *pollfd;
    int	slot;

    /* Here we need to mask SIGIO signal so that the signal handler 
       does not run & use the poll table while we are constructing it.
       This is necessary as poll table is shared data structure and we
       cannot use the locks in the signal handler.
       NOTE: The assumption here is sighold and sigrelse apply process
       wide - TBD: Confirm this is true! If NOT then we need to grab the 
       SCHED_LOCK so that thread cannot context switch here -
       Well just to be on safe side we grab SCHED_LOCK as well - pavani*/

    SCHED_LOCK();
    sighold(SIGIO); 

    entry = &fdMap[fd];
    if ((slot = entry->pollOffset) == -1) {
	entry->pollOffset = fdCount++;
	entry->flags = 0;
    }
    entry->flags |= (POLLOUT|POLLIN);

    pollfd = &pollTable[entry->pollOffset];
    pollfd->fd = fd;		/* may be redundant */
    pollfd->events = entry->flags;
    Log2(1, "Activating fd=%d onto slot=%d\n", fd, slot);
    sigrelse(SIGIO);
    SCHED_UNLOCK();
}

void
asyncIODeactivateFD(int fd)
{
    struct FdEntry *entry;
    int	slot;

    SCHED_LOCK();
    sighold(SIGIO);

    entry = &fdMap[fd];
    if ((slot = entry->pollOffset) != -1) {
	/* Now copy the highest slot in the poll array into this slot,
	   and adjust the fdMap to point at this new location.  But
	   only if we're not already in the last slot. */
	Log2(1, "Deactivating fd=%d from slot=%d\n", fd, slot);

	entry->flags &= ~(POLLOUT|POLLIN);
	if (entry->flags == 0) {	/* not reading or writing this
					   fd anymore */
	    entry->pollOffset = -1; 	/* this one's no longer occupied */
	    if (slot != --fdCount) {	/* not last entry in pollTable */
		int movefd = pollTable[fdCount].fd;

		pollTable[slot] = pollTable[fdCount];
		fdMap[movefd].pollOffset = slot;
		Log3(1, "\t and moving fd=%d from slot %d to %d\n",
		     movefd, fdCount, slot);
	    }
	} else {
	    pollTable[slot].events = entry->flags;
	}
    }
    sigrelse(SIGIO);
    SCHED_UNLOCK();
}

/*
 * Asynchronous notification of a condition variable
 *
 * NotifyAll the condition variable, and perform any scheduling updates
 * required.   Return TRUE if the caller needs to immediately
 * reschedule.
 *
 * ASSUMPTION: the current threads state is already saved, and we
 * are free to do anything we wish.
 */
static int
asyncEventNotify(sys_mon_t *mid)
{
    int need_to_switch = 0;

    sysAssert(mid!=NULL);
    if (SCHED_LOCKED()) {
	/*
	 * Queue this notification
	 */
	if ((mid->flags & SYS_MON_PENDING_NOTIFICATION) == 0) {
	    sysAssert(mid->pendingq == SYS_MID_NULL);
	    mid->pendingq = PendingNotifyQ;
	    PendingNotifyQ = mid;
	}
	mid->flags |= SYS_MON_PENDING_NOTIFICATION;
    } else {
	/*
	 * If a thread is currently inside the monitor, there is a potential
	 * race condition.  It may have already checked its test condition,
	 * and decided it wants to wait.  It will miss the notification,
	 * assuming that none can arrive while it owns the monitor.
	 *
	 * Solve this race by setting the stick_notification boolean, 
	 * indicating that a CV notify was done while some thread owned
	 * the monitor.  monitorExit() and monitorWait() are responsible
	 * for clearing this boolean.  monitorWait() should also fail
	 * to wait if the boolean is set.
	 */
	if (mid->monitor_owner != SYS_THREAD_NULL) {
	    mid->flags |= SYS_MON_STICKY_NOTIFICATION;
	}

	/* Actually handle the notification. */
	queueBroadcast(mid);

	/* If the monitor isn't locked, give it to the first waiter */
	if (!mid->monitor_owner && mid->monitor_waitq) {
	    need_to_switch = queueSignal(mid, &mid->monitor_waitq);
	}
    }
    return need_to_switch;
}

static int
sigioNotifier(int interrupt, void *arg, void *siginfo)
{
    extern sys_mon_t **fdmon;
    int resched = 0;
    int i, nfds, count;

    nfds = fdCount;
    /* Optimization: when there's only one fd to choose from it must have
       been the one that delivered the sigio, right? */
    if (nfds == 1) {
	Log1(1, "Notify single fd=%d\n", pollTable[0].fd);
	resched |= asyncEventNotify(fdmon[pollTable[0].fd]);
    } else if (nfds > 0) {
	count = (*systable[SYS_POLL].addr)((int)pollTable, nfds, 0);
	Log1(1, "Poll returns %d fds to look at\n", count);
	for (i = 0; (i < nfds) && (count>0); i++) {
	    if ( ((pollTable[i].revents & (POLLIN)) != 0) ||
		 ((pollTable[i].revents & (POLLOUT)) != 0) ) {
		Log2(1, "Notify fd=%d on mon for revents %d \n", 
		     pollTable[i].fd, pollTable[i].revents);
		pollTable[i].events = (POLLIN|POLLOUT);
		resched |= asyncEventNotify(fdmon[pollTable[i].fd]);
		 --count;
	    } 
	}
    }

    /* And for good measure... */
    /* This monitor is the default io monitor mainly used for 
       thread manipulating the group fds to wait on i.e when 
       a poll/select happens */
    resched |= asyncEventNotify(handlerMonitors[SYS_ASYNC_MON_IO]);
 
    if (resched) {
	_needReschedule = 1;
    }

    return 1;
}

/* register the signal handler for asynchronous io */
static void
asyncIORegister()
{
    intrRegister(SIGIO, sigioNotifier, 0);
}

/*
 * intrNotifyHandler() --
 *	Handle a unix signal.  This is called from the interrupt
 *	routine, which saves all pending thread state.  If there is
 *	a registered monitor, notify it (and potentially yield).
 *
 *	There can be no signal which needs both async and sync
 *	notification.
 */
static int
intrNotifyHandler(int interrupt, void *arg, void *siginfo)
{
    sys_mon_t	*mon = (sys_mon_t *)arg;

    Log2(2, "Received signal %d interrupting sys_thread_t 0x%x\n", interrupt,
	 sysThreadSelf());

    if ((mon->flags & SYS_MON_HAS_HANDLER) != 0
	&& asyncEventNotify(mon)) {
	Log(2, " (Not Locked)\n");
	_needReschedule = 1;
    }
#ifdef DEBUG
    else {
	Log1(2, " (Locked: %d)\n", _scheduling_lock);
    }
#endif
    return (1);
}


/*
 * Thread all of the pending condition variable notifications that
 * couldn't be handled at interrupt time.
 *
 * ASSUMPTION: this is being called from a thread context.
 */
void
_sched_unlock()
{
    int need_to_reschedule;

restart:
    intrLock();
    SCHED_LOCK();

    /*
     * Perform the notifyall's that couldn't be done because of locking.
     * The notification is completely normal, except for the locking.
     */
    if (PendingNotifyQ) {
	need_to_reschedule = processPendingNotification();

	intrUnlock();
	if (need_to_reschedule) {
	    YIELD();
	}

	/*
	 * We may have gotten an interrupt in between the last two
	 * statements (intrUnlock & YIELD).  Check, and if so, 
	 * loop.  This also handles the case were we really do yield
	 * to somebody else, and they get an interrupt while yielding
	 * back to us.
	 *
	 * This is gross, but it keeps this thread from recursing in
	 * this routine.
	 */
	_scheduling_lock--;
	if (PendingNotifyQ) {
	    goto restart;
	}

    } else {
	_scheduling_lock--;
	intrUnlock();
    }

    return;
}
