/*
 * @(#)threads_md.c	1.70 95/11/29
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Implementation of Java threads HPI on top of Green threads
 */

#include "sys_api.h"
#include "threads_md.h"
#include "internal_md.h"
#include "queue.h"
#include "schedule.h"
#include "itimer_md.h"
#include "threads.h"	/* For THR_USER, TID... */
#include "monitor.h"
#include "monitor_cache.h"
#include "clock.h"
#include "async_gc.h"

sys_thread_t *ThreadQueue;
static char *currentThreadStackBase;

/*
 * threads_initialized simplifies the check that has to be done in
 * sysThreadCheckStack().  Otherwise, before sysThreadInit() is called
 * on the primordial thread, we need to handle there being no current
 * thread at all, and there being one with a 0 stack_base.
 */
static int threads_initialized = 0;

/*
 * Thread C stack overflow check
 */
int sysInterruptsPending()
{
    return (int) (PendingNotifyQ != 0);
}

/*
 * Thread C stack overflow check
 */

/*
 * sysThreadCheckStack() being a function call is unfortunate, as it
 * takes stack space to do, but that is weakly accounted for by the
 * previous stack redzone.  In general, where we can't predict stack
 * use by C, thread stack overflow checking doesn't really work.
 */ 

#define STACK_REDZONE 4096

int
sysThreadCheckStack() 
{
    sys_thread_t *tid = currentThread();    /* cf. sysThreadSelf() */

    /* Stacks grow toward lower addresses on Solaris... */
    if (!threads_initialized ||
	(char *)(tid)->stack_base - (char *)&(tid) + STACK_REDZONE <
	    tid->stack_size) {
	return 1;
    } else {
	return 0;
    }
}

void
sysThreadSetBackPtr(sys_thread_t *t, void *p)
{
    t->cookie = p;
}

void *
sysThreadGetBackPtr(sys_thread_t *t)
{
    return t->cookie;
}

/*
 * sysThreadBootstrap() bootstraps the UNIX process running from main()
 * into a first primordial thread.  This thread is distinguishable because
 * it uniquely has the primordial_thread flag on in its private data.
 * However, so that we have to special-case it as little as possible, we
 * set it up to look as closely as possible to a thread that we created.
 * One difference is that its stack is *not* known to us.
 */
int
sysThreadBootstrap(sys_thread_t **ptid)
{
    sys_thread_t *tid;

    /* Allocate and initialize the ThreadPrivate structure */
    tid = (sys_thread_t *) malloc(sizeof(*tid));
    if (tid == 0) {
	out_of_memory();
    }
    memset((char *)tid, 0, sizeof(*tid));

    /* 
     * The stack size of the primordial thread is is set to an impossibly
     * large value.  This causes the thread stack overflow check in the
     * interpreter to always succeed for the primordial thread.  We
     * could conceivably model C stack expansion with available swap
     * in mind, but that would be pretty painful.
     */

#define HUGE_C_STACK_SIZE 0x70000000

    tid->stack_size = HUGE_C_STACK_SIZE;
    tid->system_thread = 0;     /* Primordial thread considered user */
    tid->primordial_thread = 1;
    
    /* Record whether this thread is system or user */
    tid->system_thread = 0;

    QUEUE_LOCK();
    ActiveThreadCount++;        /* Count the main thread */
    tid->next = ThreadQueue;    /* Chain all threads */
    ThreadQueue = tid;
    if (tid->system_thread == 0) {
        UserThreadCount++;	/* The main thread considered user */
    }    
    QUEUE_UNLOCK();

    tid->state = RUNNABLE;
    setCurrentThread(tid);

    *ptid = tid;
    return SYS_OK;
}

/*
 * Create system dependent threads.
 */
void
sysThreadInitializeSystemThreads()
{
    extern int noasyncgc;
    extern void InitializeFinalizerThread(void);

    /* Initialize the clock */
    InitializeClock();

    /* Create the idle thread */
    InitializeIdle();

    if (noasyncgc == 0) {
	InitializeGCThread();
    }

    /* Create the finalization thread */
    InitializeFinalizerThread();
}

/*
 * Access to the thread stack pointer of an arbitrary thread (for GC).
 */
void *
sysThreadStackPointer(sys_thread_t * tid)
{
    extern void *currentStackPointer(sys_thread_t *);

    return currentStackPointer(tid);
}

/*
 * Access to the thread stack base of an arbitrary thread.
 */
stackp_t
sysThreadStackBase(sys_thread_t * tid)
{
    return tid->stack_base;	/* Used to be 0 for primordial thread */
}

static void removeFromActiveQ(sys_thread_t *p);

static void start_func(int dummy, int arg, void (*func)()) {
    SCHED_UNLOCK();
    (*func)(arg);
    panic("returned from start_func\n");

}

int
sysThreadCreate(long stack_size, unsigned int flags,
		void *(*start)(void *), sys_thread_t **ptid, void * cookie)
{ 
    sysAssert(cookie != 0);

    /*
     * Allocate thread private memory and stack.
     */
    if (!allocateContextAndStack(ptid, stack_size)) {
	/* Exception thrown by caller if necessary. */
	return SYS_ERR;		/* What exactly can happen?  No mem? */
    }

    /*
     * All threads are created in a suspened (non-runnable) state.
     */
    (*ptid)->state = SUSPENDED;
    initContext(*ptid, (unsigned int) start, start_func, (unsigned int) cookie);

    /* Record whether this thread is system or user */
    if (flags == THR_USER) {
        (*ptid)->system_thread = 0;
    } else {
        (*ptid)->system_thread = 1;
    }   

    QUEUE_LOCK();
    ActiveThreadCount++;        /* Global thread count */
    (*ptid)->next = ThreadQueue;    /* Chain all threads */
    ThreadQueue = *ptid;
    /* Record whether this thread is system or user */
    if ((*ptid)->system_thread == 0) {
        UserThreadCount++;
    }
    QUEUE_UNLOCK();

    return SYS_OK;
}

/*
 * sysThreadInit() must be called from within the context of the
 * newly running thread.
 */
void
sysThreadInit(sys_thread_t * tid, stackp_t stack)
{
    if (tid->primordial_thread) {
	tid->stack_base = stack;
	threads_initialized = 1;
    }
}

/*
 * Destroy the specified thread.  If a thread is destroying itself,
 * this function will never return.
 *
 * NOTE: there is a big assumption in this code that the current thread
 * can destroy itself.  I.e. there are some strong order dependencies
 * in the code.  E.g. the stack can be nuked, and the thread can keep
 * running.  This works under Unix because freed memory is still valid
 * memory.  May need to revisit.
 *
 * This function gets called on initial thread creation, when the
 * the thread returns from its initial function we destroy it.  We don't
 * want to actually destroy a thread Object, but we do want to clean up
 * all of its private state.
 *
 * We also want to check the waiters queue to see if anybody is
 * waiting on the thread to die.
 */

#ifdef DEBUG
static void threadReallyDead(sys_thread_t *tid);
#endif

void 
sysThreadExit()
{
    extern int greenMonitorExitLocked(sys_mon_t *);
    extern void monitorCleanup(TID);

    int resched;
    sys_thread_t *tid;
    TID thread_handle;
    sys_mon_t *mid;
    monitor_t *monitor;
    ExecEnv *ee;

    tid = sysThreadSelf();

    CACHE_LOCK();	/* Make sure we have monitor when we need it */
    REGISTRY_LOCK();	/* Make sure we have monitor when we need it */

    removeFromActiveQ(tid);

    /*
     * WARNING: DO NOT ATTEMPT TO ENTER ANY MONITORS AFTER THIS POINT
     * (unless you are reentering something you are *guaranteed* to
     * already own).
     */
    SCHED_LOCK();

    /*
     * Free the Java stack.  Free the thread stack and sys_thread_t,
     * and zero the thread object's PrivateInfo field.  The issue
     * here is we need to do everything that requires any kind of
     * monitor locking, including calls to malloc and free, before
     * we do the monitor cleanup code below.  The reason is any of
     * those calls into monitors *could* result in this process that
     * we're trying to kill being placed on some queue.  NOTE: This
     * can only happen if we're trying to kill the current thread,
     * but we take these precautions all the time since it does no
     * harm.
     *
     * We continue to use the "freed" stack, but we don't actually
     * reclaim the stack until later, when we are guaranteed to be
     * done with it.
     */

    /* Put the Java stack segments on to-be-freed list (don't free yet) */
    thread_handle = (TID) tid->cookie;
    ee = (ExecEnv *) THREAD(thread_handle)->eetop;
    if (ee) {		/* ee == 0 does happen... */
	DeleteExecEnv(ee, (JHandle *) thread_handle);	/* zeros eetop */
    }

    /* Put the sys_thread_t and stack on to-be-freed list (don't free yet) */
    deleteContextAndStack(tid);

    /*
     * This notifies anybody who is waiting for us to die.  Note that
     * we notify *everybody* on the queue.
     */
    monitor = lookupMonitor(obj_monitor((tid->cookie)));
    if (monitor) {
	mid = (sys_mon_t *) &monitor->mid;
	if (mid->condvar_waitq) {
	    queueBroadcast(mid);

	    /* If the monitor isn't locked, give it to first waiter */
	    if (!mid->monitor_owner && mid->monitor_waitq) {
		resched = queueSignal(mid, &mid->monitor_waitq);
	    }
	}
    }

    /*
     * As threads only exit now, the exiting thread can't be waiting
     * on a monitor, so we don't have to clean up the mon_wait monitor.
     */
    sysAssert(tid->mon_wait == 0);

    /*
     * Clean up all monitors that we currently own.  monitorCleanup()
     * exits _moncache_lock but not _registry_lock
     */
    monitorCleanup(thread_handle);
    resched |= greenMonitorExitLocked(_registry_lock);

    /* Remove the dying thread from threadAlarmQ */
    cleanupPendingAlarm(tid);

    THREAD(thread_handle)->PrivateInfo = 0;

#ifdef DEBUG
    threadReallyDead(tid);
#endif

    reschedule();	 /* Call with SCHED_LOCKED() */
    /* NOTREACHED */
    panic("Return from sysThreadExit(): a zombie is awake!\n");
}

/*
 * Current thread yield control
 *
 * Green threads originally supported forcing another thread to yield...
 */
void
sysThreadYield()
{
    SCHED_LOCK();
    if (runnable_queue->priority == sysThreadSelf()->priority) {
	YIELD();
    }
    SCHED_UNLOCK();
    return;
}

/*
 * Get the scheduling priority of a specified thread
 */
int
sysThreadGetPriority(sys_thread_t * tid, int *priority)
{
    *priority = tid->priority;
   
    return SYS_OK;
}

/*
 * Set the scheduling priority of a specified thread
 */
int
sysThreadSetPriority(sys_thread_t *tid, int priority)
{
    sys_mon_t *mid;
    sys_thread_t *self;

    self = sysThreadSelf();

    SCHED_LOCK();
    if (tid == self) {
        self->priority = priority;
    } else {
	/*
	 * The priority is changing, which means we have to reorder the
	 * queue that the thread is queued upon.
	 */
	tid->priority = priority;
	switch(tid->state) {
	    case RUNNABLE:
		queueRemove(&runnable_queue, tid);
		queueInsert(&runnable_queue, tid);
		break;

	    case MONITOR_WAIT:
		mid = (sys_mon_t *) tid->mon_wait;
		queueRemove(&mid->monitor_waitq, tid);
		queueInsert(&mid->monitor_waitq, tid);
		break;

	    case CONDVAR_WAIT:
		mid = (sys_mon_t *) tid->mon_wait;
		queueRemove(&mid->condvar_waitq, tid);
		queueInsert(&mid->condvar_waitq, tid);
		break;

	    case MONITOR_SUSPENDED:
		mid = (sys_mon_t *) tid->mon_wait;
		queueRemove(&mid->suspend_waitq, tid);
		queueInsert(&mid->suspend_waitq, tid);
		break;

	    case SUSPENDED:
		/* The thread is not on any queue when it is suspended. */
		/* so we don't have to do anything, right, or not? */
		break;

	    default:
		/* REMIND: Should throw an exception! */
		fprintf(stderr, "Tried to set priority of a thread in unknown state\n");
		fprintf(stderr, "See %s, line %d\n", __FILE__, __LINE__);
		sysAbort();

	}
    }

    if (runnable_queue && runnable_queue->priority > priority) {
	YIELD();
    }

    SCHED_UNLOCK();
    return SYS_OK;
}

/* Resume execution of the specified thread */
int
sysThreadResume(sys_thread_t *tid)
{
    sys_mon_t *mid;

    SCHED_LOCK();
    tid->pending_suspend = 0;

    switch(tid->state) {
	case SUSPENDED:
	    tid->state = RUNNABLE;
	    queueInsert(&runnable_queue, tid);
	    if (tid->priority > sysThreadSelf()->priority) {
		YIELD();
	    }
	    break;
	    
	case MONITOR_SUSPENDED:
	    mid = (sys_mon_t *) tid->mon_wait;
	    sysAssert(mid);
	    tid->state = MONITOR_WAIT;
	    queueRemove(&mid->suspend_waitq, tid);
	    queueInsert(&mid->monitor_waitq, tid);

	    /* AVH: Signal if this thread was put in the wait queue
	     * and the monitor is not blocked.
	     */
	    if ((mid->monitor_waitq == tid) && (mid->monitor_owner == 0)) {
		queueSignal(mid, &mid->monitor_waitq);
	    }
	    break;
    }

    SCHED_UNLOCK();
    return SYS_OK;
}

/* Suspend execution of the specified thread */
int
sysThreadSuspend(sys_thread_t *tid)
{
    int err = 0;
    sys_thread_t *self = sysThreadSelf();

    SCHED_LOCK();

    if (tid == self) {
	self->state = SUSPENDED;
	yieldContext(CONTEXT(self));
    } else {
	switch(tid->state) {
	    case RUNNABLE:
		queueRemove(&runnable_queue, tid);
		tid->state = SUSPENDED;
		break;

	    case MONITOR_WAIT:
	    case CONDVAR_WAIT:
		tid->pending_suspend = 1;
		break;

	    case SUSPENDED:
	    case MONITOR_SUSPENDED:
		err = -1;
		break;
	}
    }

    SCHED_UNLOCK();
    return (err == 0 ? SYS_OK : SYS_ERR);
}

/*
 * The next set of global variables and the two routines that follow
 * are used by routines like the garbage collector to lock out other
 * threads for some period of time.  See the extensive comments in
 * interrupt_md.c for a complete explanation.
 */

/*
 * Stop all threads other than the current one.  The stopped threads
 * may be restarted with sysThreadMulti(); the operation of this
 * routine is atomic; it either stops all java threads or it stops
 * none of them.  Upon success (all threads stopped) this routine
 * returns SYS_OK, otherwise SYS_ERR.
 *
 * In general, sysThreadSingle() should take care of anything below
 * the HPI that needs to be done to safely run single-threaded.
 */
int
sysThreadSingle()
{
    SCHED_LOCK();

    return SYS_OK;
}

/*
 * Allow multi threaded execution to resume after a
 * sysThreadSingle() call.
 *
 * Note: When this routine is called the scheduler should already
 * have been locked by sysThreadSingle().  Each sysThreadSingle()
 * call should have a matching sysThreadMulti().
 */
void sysThreadMulti() 
{
    sysAssert(SCHED_LOCKED());

    SCHED_UNLOCK();
}

/* Set the current thread, and also set currentThreadStackBase to to
 * thread's base.  This global allows quick stack overflow checks.
 */
sys_thread_t *_CurrentThread;
 
sys_thread_t *
sysThreadSelf()
{
    return currentThread();
}

/*
 * Enumerate over all threads, calling a function for each one.  A
 * paranoid helper function would be prepared to deal with threads
 * that have not been created by Java.
 */
int
sysThreadEnumerateOver(int (*func)(sys_thread_t *, void *), void *arg)
{
    sys_thread_t *tid;
    int err = SYS_OK;
    int i;

    QUEUE_LOCK();
    tid = ThreadQueue;
    for (i = 0; i < ActiveThreadCount && tid != 0; i++) {
	if ((err = (*func)(tid, arg)) != SYS_OK)
	    break;
        tid = tid->next;
    }
    QUEUE_UNLOCK();
    return err;
}

/* 
 * Dump system-specific stuff about a thread, e.g. its native thread ID.
 */
void
sysThreadDumpInfo(sys_thread_t *tid)
{
    /* Nothing to do for Solaris/Green threads */
}

/*
 * Miscellaneous stuff in support of HPI for Green threads
 */
void
setCurrentThread(sys_thread_t *tid)
{
    _CurrentThread = tid;
    if (tid->primordial_thread) {
        currentThreadStackBase = 0;     /* infinite, essentially */
    } else {
        currentThreadStackBase = (char *) &tid->stack_base;
    }
}

/*
 * Remove this thread from the list of Active threads.
 */
static void
removeFromActiveQ(sys_thread_t * t) 
{
    sys_thread_t *prev;
    sys_thread_t *tid;

    QUEUE_LOCK();
    ActiveThreadCount--;
    if (t->system_thread == 0) {
        UserThreadCount--;
	QUEUE_NOTIFY();
    }

    prev = 0;
    tid = ThreadQueue;
    while (tid) {
	if (tid == t) {
	    if (prev) {
		prev->next = tid->next;
	    } else {
		ThreadQueue = tid->next;
	    }
	    tid->next = 0;
	    break;
	}
	prev = tid;
	tid = tid->next;
    }
    QUEUE_UNLOCK();
}

/*
 * The primordial thread is not allowed to fall off the end of start_java()
 * until the Java system is about to exit.  This is allowed only when the
 * primordial thread is the sole "user" thread remaining.
 */
void
WaitToDie() {
    QUEUE_LOCK();
    while (UserThreadCount > 1) {  /* Primordial thread considered user */
	QUEUE_WAIT();
    }
    QUEUE_UNLOCK();
}

#ifdef DEBUG

/*
 * Diagnostic routines:
 *
 * If a thread is really dead should not appear in:
 *
 *  - the thread queue
 *  - the alarm queue
 *  - any monitor's waitq, condvar_waitq or suspend_waitq
 *  - the runnable_queue
 *
 * Note that this can only be tested between the time that a thread
 * is cleaned up after and the time that its stack is deleted!  This
 * makes it applicable in only a very small window in sysThreadExit(),
 * after which point the thread's sys_thread_t is deallocated.  Note
 * also that the thread may be checking itself, and so it is valid for
 * the thread being checked to be the current thread or to have certain
 * monitors at certain times, like _registry_lock when enumerating
 * registered monitors.
 */
static void
checkNotInQueue(sys_thread_t *tid, sys_thread_t *t)
{
    for (; t != 0; t = t->waitq) {
	sysAssert(t != tid);
    }
}

static void
notInMonitor(sys_thread_t *tid, sys_mon_t *mid)
{
    sysAssert(mid->monitor_owner != tid);

    checkNotInQueue(tid, mid->monitor_waitq);
    checkNotInQueue(tid, mid->condvar_waitq);
    checkNotInQueue(tid, mid->suspend_waitq);
}

static void
monitorNotInMonitorHelper(monitor_t *monitor, void *arg)
{
    sys_thread_t *tid = (sys_thread_t *) arg;
    sys_mon_t *mid = (sys_mon_t *) &monitor->mid;

    notInMonitor(tid, mid);
}

static void
threadNotInMonitors(sys_thread_t *tid)
{
    extern monitor_t **monHashTable;
    extern reg_mon_t *MonitorRegistry;  /* Root of registered monitors */
    extern int monHashTableBuckets;
    monitor_t *mid;
    int	i;

    for (i = monHashTableBuckets; --i >= 0; ) {
	mid = monHashTable[i];
	while (mid != 0) {
	    notInMonitor(tid, (sys_mon_t *) &mid->mid);
	    mid = mid->next;
	}
    }

    {
	reg_mon_t *reg;

	reg = MonitorRegistry;
	while (reg) {
	    notInMonitor(tid, reg->mid);
	    reg = reg->next;
	}
    }
}

static void
threadNotOnThreadQueue(sys_thread_t *tid)
{
    sys_thread_t *t;

    /* Check the thread queue */
    t = ThreadQueue;
    while (t) {
	sysAssert(t != tid);
	t = t->next;
    }
}

static void
threadNotOnRunQueue(sys_thread_t *tid)
{
    checkNotInQueue(tid, runnable_queue);
}

static void
threadNotOnAlarmQueue(sys_thread_t *tid)
{
    extern sys_thread_t *threadAlarmQ;

    checkNotInQueue(tid, threadAlarmQ);
}

static void
threadReallyDead(sys_thread_t *tid)
{
    sysAssert(SCHED_LOCKED());
    threadNotOnThreadQueue(tid);
    threadNotOnAlarmQueue(tid);
    threadNotOnRunQueue(tid);
    threadNotInMonitors(tid);
}

#endif

/*
 * The mechanics of actually signalling an exception (in the future,
 * and Alarm or Interrupt) depend upon what thread implementation you
 * are using.
 */
void
sysThreadPostException(sys_thread_t *tid, void *exc)
{
    struct execenv *ee;  

    ee = (ExecEnv *) THREAD((TID) tid->cookie)->eetop;
    if (ee != (ExecEnv *) 0) {   
	exceptionThrow(ee, exc);
    }
}
