/*
 * @(#)async_gc.c	1.9 95/11/29
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Machine-specific support for asynchronous garbage collection: creation
 * of the idle thread and async GC thread.
 */

#include "threads.h"
#include "async_gc.h"
#include "log.h"

#include "java_lang_Thread.h"

extern ClassClass *Thread_classblock;

void out_of_memory(void);

/*
 * Idle thread
 */

static volatile int idle_count = 0;

static void
idle_thread(TID tid)
{
    int sb;
    extern void FreeJavaStackMemory();

    threadInit(tid, (stackp_t) &sb);
    while (1) {
	Log1(2, "Idle loop is running, %d active threads\n", 
	     ActiveThreadCount);
	FreeJavaStackMemory();	/* Free Java stacks from dead threads */
	idle_count++;
	pause();	/* Suspend idle thread */
    }
}

/*
 * Create the idle thread.
 */
void
InitializeIdle()
{
    TID tid;
    char *name = "Idle thread";

    tid = (TID)execute_java_constructor(NULL, NULL, Thread_classblock, "()");
    if (threadCreate(tid, THR_SYSTEM, ProcStackSize,
		     (void * (*)(void *)) idle_thread) != 0) {
	SignalError(0, JAVAPKG "OutOfMemoryError", 0);
    } else {
	THREAD(tid)->name = MakeString(name, strlen(name));
	THREAD(tid)->priority = MinimumPriority - 1;
	THREAD(tid)->daemon = 1;
	threadSetPriority(tid, MinimumPriority - 1);
	threadResume(tid);
    }
}


/*
 * Asynchronous GC thread
 *
 * Asynchronous GC is decided upon as follows: the idle thread is running
 * at a priority lower than anything else, including the async GC thread.
 * If the async GC thread wakes up, it remembers the current idle count
 * and then goes to sleep for a second.  (The idle count essentially counts
 * the number of times the idle thread has run.)  If when the GC thread
 * wakes up again the idle count is higher than it had been, it means that
 * the idle thread ran while the GC thread slept.  That's a reasonably good
 * indication that nothing was going on during that time, and thus that
 * nothing will be going on in the near future, so the garbage collector is
 * run.  If the idle thread has not been run while the GC thread, it goes
 * back to sleep again.
 *
 * Note that there is nothing to keep the system from garbage collecting
 * even though nothing has happened since it last GC'ed, and this will in
 * fact happen when the system is truly idle.  This isn't really desirable
 * as it takes CPU cycles and, if one of these useless GCs is interrupted,
 * may mean worse system responsiveness.  We could avoid this by, say, having
 * a flag that is set when GC completes and is cleared when the system does
 * something significant, like runs some Java code.
 */
static void
gc_loop(TID tid)
{
    extern int heap_memory_changes;
    extern void asyncGC();
    int previous_idle_value;

    threadInit(tid, (stackp_t) &tid);
    while (1) {
	previous_idle_value = idle_count;
	threadSleep(1000);	/* 1000 milliseconds */

	/* run GC on gc'able heap */
	if (idle_count > previous_idle_value && heap_memory_changes) {
	    asyncGC();
	}
    }
}

void
InitializeGCThread()
{
    TID tid;
    char *name = "Async Garbage Collector";

    tid = (TID)execute_java_constructor(NULL, NULL, Thread_classblock, "()");
    if (threadCreate(tid, THR_SYSTEM, ProcStackSize,
		     (void * (*)(void *)) gc_loop) != 0) {
	out_of_memory();		/* fatal error */
    }
    /* Set expected tid fields... */
    THREAD(tid)->name = MakeString(name, strlen(name));
    THREAD(tid)->priority = MinimumPriority;
    THREAD(tid)->daemon = 1;
    threadSetPriority(tid, MinimumPriority);
    threadResume(tid);
}
