/*
 * @(#)memory_md.c	1.9 95/11/29
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Implementation of primitive memory allocation.
 *
 * The only thing machine dependent about this allocator is how it 
 * initially finds all of the possible memory, and how it implements
 * mapChunk() and unmapChunk().
 *
 * This is all pretty simple stuff.  It is not likely to be banged on
 * frequently enough to be a performance issue, unless the underlying
 * primitives are.  Implementing things:
 *
 * HPI function      Solaris   "malloc"    Win32 
 * --------------------------------------------------------------------
 * sysMapMem()	     mmap()     malloc()   VirtualAlloc(...MEM_RESERVE...)
 * sysUnMapMem()     munmap()   free()     VirtualFree(...MEM_RESERVE...)
 * sysCommitMem()    no-op      no-op      VirtualAlloc(...MEM_COMMIT...)
 * sysDecommitMem()  no-op      no-op      VirtualFree(...MEM_COMMIT...)
 *
 * Memory mapping is the default, but compiling with -DUSE_MALLOC gives
 * a system based on malloc().
 */

#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>	/* For perror() */

#include "log.h"
#include "sys_api.h"

#ifndef USE_MALLOC

#include <sys/mman.h>
#include <fcntl.h>
static int devZeroFD;

#endif /* !USE_MALLOC */

static unsigned int memGrainSize;	/* A page for Solaris */

/*
 * Mem size rounding is done at this level.  The calling code asks
 * these routines for literally what it thinks it wants.  The size is
 * rounded up to the first multiple of memGrainSize that contains 
 * requestedSize bytes.
 */

static long
roundUpToGrain(long value)
{
    return (value + memGrainSize - 1) & ~(memGrainSize - 1);
}

static long
roundDownToGrain(long value)
{
    return value & ~(memGrainSize - 1);
}

void
InitializeMem()
{
    static int init = 0;
    extern unsigned int memGrainSize;

    if (init) {
	return;		/* Subsequent calls are no-ops */
    }

    /*
     * Set system-specific variables used by mem allocator
     */
    if (memGrainSize == 0) {
	memGrainSize = sysconf(_SC_PAGESIZE);
    }

#ifndef USE_MALLOC
    devZeroFD = open("/dev/zero", O_RDWR);
    if (devZeroFD == -1) {
	perror("devzero");
	sysExit(0);
    }
#endif /* !USE_MALLOC */

    init = 1;		/* We're initialized now */
}


#ifndef USE_MALLOC

#define PROT_ALL (PROT_READ|PROT_WRITE|PROT_EXEC)

/*
 * Map a chunk of memory.  Return the address of the base if successful,
 * 0 otherwise.  We do not care where the mapped memory is, and can't
 * even express a preference in the current HPI.  If any platforms
 * require us to manage addresses of mapped chunks explicitly, that
 * must be done below the HPI.
 */
static char *
mapChunk(long length)
{
    char *ret;

    ret = (char *) mmap(0, length, PROT_ALL, MAP_NORESERVE|MAP_PRIVATE,
		   devZeroFD, (off_t) 0);
    return (ret == MAP_FAILED ? 0 : ret);
}

/*
 * Map a chunk of memory at a specific address and reserve swap space
 * for it.  This is currently only used to remap space previously mapped
 * MAP_NORESERVE, reserving swap and getting native error handling.  We
 * assume that all alignment and rounding has been done by the caller.
 * Return 1 if successful and 0 otherwise.
 */
static char *
mapChunkReserve(char *addr, long length)
{
    char *ret;
 
    ret = (char *) mmap(addr, length, PROT_ALL, MAP_FIXED|MAP_PRIVATE,
                        devZeroFD, (off_t) 0);
    return (ret == MAP_FAILED ? 0 : ret);
}

/*
 * Unmap a chunk of memory.  Return 1 if successful, 0 otherwise.  We
 * currently don't do any alignment or rounding, assuming that we only
 * will unmap chunks that have previously been returned by mapChunk().
 */
static int
unmapChunk(void *addr, long length)
{
    return (munmap(addr, length) == 0);
}

#endif /* !USE_MALLOC */


/* HPI Functions: */

/*
 * Map a range of virtual memory.  Note that the size asked for here
 * is literally what the upper level has asked for.  We need to do
 * any rounding, etc. here.  If mapping fails return 0, otherwise
 * return the address of the base of the mapped memory.
 */
void *
sysMapMem(long requestedSize, long *mappedSize)
{
    void *mappedAddr;

    *mappedSize = roundUpToGrain(requestedSize);
#ifdef USE_MALLOC
    mappedAddr = (void *) malloc(*mappedSize);	/* Returns 0 on failure */
#else
    mappedAddr = mapChunk(*mappedSize);		/* Returns 0 on failure */
#endif /* USE_MALLOC */
    if (mappedAddr) {
	Log3(2, "sysMapMem: 0x%x bytes at 0x%x (request: 0x%x bytes)\n",
	     *mappedSize, mappedAddr, requestedSize);
    } else {
	Log1(2, "sysMapMem failed: (request: 0x%x bytes)\n", requestedSize);
    } 
    return mappedAddr;
}

/*
 * Unmap a range of virtual memory.  Note that the size asked for here
 * is literally what the upper level has asked for.  We need to do any
 * rounding, etc. here.  If unmapping fails return 0, otherwise return
 * the address of the base of the unmapped memory.
 */
void *
sysUnmapMem(void *requestedAddr, long requestedSize, long *unmappedSize)
{
    void *unmappedAddr;
    int ret;

    *unmappedSize = roundUpToGrain(requestedSize);
#ifdef USE_MALLOC
    free(requestedAddr);
    ret = 1;
#else
    ret = unmapChunk(requestedAddr, *unmappedSize);
#endif /* USE_MALLOC */
    if (ret) {
	unmappedAddr = requestedAddr;
        Log4(2,
	     "sysUnmapMem: 0x%x bytes at 0x%x (request: 0x%x bytes at 0x%x)\n",
	     *unmappedSize, unmappedAddr, requestedSize, requestedAddr);
    } else {
	unmappedAddr = 0;
	Log2(2, "sysUnmapMem failed: (request: 0x%x bytes at 0x%x)\n",
	     requestedSize, requestedAddr);
    }
    return unmappedAddr;
}

/*
 * Commit/decommit backing store to a range of virtual memory.  This range
 * needs not be identical to a mapped range, but must be a subset of one.  
 * On Solaris, we remap the range to reserve swap for the space on
 * commit.  We don't strictly need to do this, as Solaris will demand
 * page pages that we've mapped when we want to access them.  But by
 * reserving swap we get reasonable error handling for free where we'd
 * otherwise end up getting a SIGBUS on a random write when we run out
 * of swap.  It also emphasizes the general need for shared code to
 * postpone committing to mapped memory for as long as is feasible.
 * When Java really needs space (the thread stacks excepted), it will
 * soon write over it (heap, markbits), so we don't really get much from
 * demand paging. 
 *
 * We do not validate that commitment requests cover already-mapped
 * memory, although in principle we could.  The size asked for here
 * is what the upper level has asked for.  We need to do any platform-
 * dependent rounding here.
 *
 * When you commit, you commit to the entire page (or whatever quantum
 * your O/S requires) containing the pointer, and return the beginning of
 * that page.  When you decommit, you decommit starting at the next page
 * *up* from that containing the pointer, except that decommitting from
 * a pointer to the beginning of the page operates on that page.
 */

/*
 * Return the address of the base of the newly committed memory, or 0
 * if committing failed.
 */
void *
sysCommitMem(void *requestedAddr, long requestedSize, long *committedSize)
{
    void *committedAddr;
    char *ret;

    *committedSize = roundUpToGrain(requestedSize);
    committedAddr = (void *) roundDownToGrain((long) requestedAddr);
    ret = mapChunkReserve(committedAddr, *committedSize);
    if (ret) {
	committedAddr = ret;
	Log4(2,
    "sysCommitMem: 0x%x bytes at 0x%x (request: 0x%x bytes at 0x%x)\n",
	     *committedSize, committedAddr, requestedSize, requestedAddr);
    } else {
	committedAddr = 0;
	Log2(2, "sysCommitMem failed: (request: 0x%x bytes at 0x%x)\n",
	     requestedSize, requestedAddr);
    }

    return committedAddr;
} 

/*
 * Return the address of the base of the newly decommitted memory, or 0
 * if decommitting failed.  This is currently a no-op for Solaris -- I
 * don't know if it's possible to un-reserve swap.
 */
void *
sysDecommitMem(void *requestedAddr, long requestedSize, long *decommittedSize)
{
    void *decommittedAddr;

    *decommittedSize = roundUpToGrain(requestedSize);
    decommittedAddr = (void *) roundUpToGrain((long) requestedAddr);
    Log4(2,
	 "sysDecommitMem: 0x%x bytes at 0x%x (request: 0x%x bytes at 0x%x)\n",
	 *decommittedSize, decommittedAddr, requestedSize, requestedAddr);

    return decommittedAddr;	/* Always succeeds on Solaris */
}
