/*
 * @(#)classloader.c	1.81 96/03/05  
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * modified by Nigel Atkinson for ST20, 17/6/96
 */

/*-
 *      Routines for loading and resolving class definitions.
 *      These routines should not depending up on the interpreter or
 *      garbage collector.
 *
 */

#include <string.h>
#include <stddef.h>
#include <iocntrl.h>
#include "stat_fix.h"
#include <stdlib.h>
#include <setjmp.h>
#include "st20misc.h"
#include "oobj.h"
#include "path.h"
#include "interpreter.h"
#include "tree.h"
#include "bool.h"
#include "signature.h"

#include "sys_api.h"
#include "byteorder_md.h"
#include "io_md.h"

int LoadFile(char *fn, char *dir, char *SourceHint);
int DoImport(char *name, char *SourceHint);
char *stat_source(ClassClass *cb, struct stat *s, char *pathbuf, int maxlen);

void
AddBinClass(ClassClass * cb)
{
    register int i = -1;
    if (cb->loader == 0) {
	for (i = (int) nbinclasses; --i >= 0 && strcmp(classname(binclasses[i]), classname(cb)) != 0;)
          ;
    }
    if (i < 0) {
	if (nbinclasses >= sizebinclasses){
	    if (binclasses == 0)
		binclasses = (ClassClass **)sysMalloc(sizeof(ClassClass *) * (int) (sizebinclasses = 50));
	    else
		binclasses = (ClassClass **)realloc(binclasses, sizeof(ClassClass *) * (int) (sizebinclasses = nbinclasses * 2));
        }
	if (binclasses == 0)
	    return;
	i = (int) nbinclasses;
	nbinclasses++;
    }
    binclasses[i] = cb;
}

void
DelBinClass(ClassClass * cb)
{
    register int i;
    for (i = (int) nbinclasses; --i >= 0; )
	if (binclasses[i] == cb) {
	    binclasses[i] = binclasses[--nbinclasses];
	    break;
	}
}

/* LoadFile is shared by the compiler and the interpreter.  It does
   different things for each.

   When the compiler askes to load a file for a particular class,
   in the end all it's looking for is a classinfo structure to be
   created for that class.  It can get that either by loading a
   .class file, and creating the classinfo from that, OR, by
   reading a source file and compiling it.

   When the interpreter askes for a file, in the end it's looking
   for a classblock structure to be created.  The only way it can
   get one of those is by loading a compiled class.

   OpenCode tries to open a .class file first.  -2 means it failed
   because it couldn't open the file, -1 means the requested file
   is being compiled.  Of course, in the case of the interpreter,
   OpenCode never returns -1, only a valid fd or -2.

   If OpenCode returns a valid fd, that means it is a .class file.
   If SkipCheckSource is false (meaning we're checking source
   files), and the source file as specified in the compiled .class
   file is more recent than the .class file, then the loaded class
   is thrown away, and OpenCode is called again, only this time,
   with arguments which will cause it to recompile the source.  In
   the case of java, it forks javac to do this.  When this happens,
   it has to reload the resulting .class file.

   Returns 1 when it succeeds, 0 when it fails. */

extern int OpenCode(char *, char *, char *, struct stat*);
int
LoadFile(char *fn, char *dir, char *SourceHint)
{
    struct stat st;
    ClassClass *cb = 0;
    int codefd = -1;
    int	retryCount = 0;
    unsigned char *external_class;

    codefd = OpenCode(fn, SourceHint, dir, &st);

again:
    switch (codefd) {
      case -2:		    /* open failed */
	return 0;

      case -1:
	return 1;       /* already compiled (or compiling) source */
    }

    /* Snarf the file into memory. */
    external_class = (unsigned char *)sysMalloc(st.st_size);
    if (external_class == 0)
        goto failed;
    if (sysRead(codefd, (char *)external_class, st.st_size) != st.st_size)
        goto failed;
    sysClose(codefd);
    codefd = -1;
    cb = sysCalloc(sizeof(ClassClass), 1);
    if (!createInternalClass(external_class, external_class + st.st_size, cb)) {
	sysFree(external_class);
	goto failed;
    }
    sysFree(external_class);
    /* Valid class - let's put it in the class table. */
    AddBinClass(cb);
    
    /* Now, if we're checking sources, and there's a source file
       name in the .class file we just loaded, check to see if we
       should recompile this class.  If so, call try again. */

    if (!SkipSourceChecks && cb->source_name) {
	struct stat st2;
	char pathbuf[255];

	if (cb->major_version != JAVA_VERSION) {
	    fprintf(stderr, "Warning: Not using %s because it was\n"
		    "\t compiled with version %d.%d of javac.  "
		    "Current version is %d.%d.\n",
		    classname(cb),
		    cb->major_version, cb->minor_version, 
		    JAVA_VERSION, JAVA_MINOR_VERSION);
	}
	if (stat_source(cb, &st2, pathbuf, sizeof(pathbuf)) != 0
	    && (st2.st_mtime > st.st_mtime || cb->major_version != JAVA_VERSION)
	    && ++retryCount < 2) {
	    if (cb->major_version != JAVA_VERSION) {
		fprintf(stderr, "Warning: Attempting to recompile from %s...\n",
			classsrcname(cb));
		unlink(fn);
	    }
	    /* remove the class we just defined - we're going again;  */
	    codefd = OpenCode(fn, pathbuf, dir, &st);
	    DelBinClass(cb);
	    FreeClass(cb);
	    goto again;
	} else if (cb->major_version != JAVA_VERSION) {
	    /* remove the class we just defined - we're going
	       to try again (once) */
	    DelBinClass(cb);
	    goto failed;
	}
    }
    if (verbose)
	fprintf(stderr, "[Loaded %s]\n", fn);
    return 1;
failed:
    if (codefd >= 0)
	sysClose(codefd);
    if (cb != 0)
        FreeClass(cb);
    return 0;
}

int
LoadFileZip(char *fn, zip_t *zip)
{
    struct stat st;
    ClassClass *cb;
    unsigned char *clazz;

    if (!zip_stat(zip, fn, &st)) {
	return 0;
    }
    if ((clazz = sysMalloc(st.st_size)) == 0) {
	return 0;
    }
    if (!zip_get(zip, fn, clazz, st.st_size)) {
	sysFree(clazz);
	return 0;
    }
    if ((cb = sysCalloc(1, sizeof(ClassClass))) == 0) {
	sysFree(clazz);
	return 0;
    }
    if (!createInternalClass(clazz, clazz + st.st_size, cb)) {
	sysFree(clazz);
	sysFree(cb);
	return 0;
    }
    sysFree(clazz);
    AddBinClass(cb);
    if (verbose) {
	fprintf(stderr, "[Loaded %s from %s]\n", fn, zip->fn);
    }
    return 1;
}

int 
DoImport(char *name, char *SourceHint)
{
    cpe_t **cpp;

    if (name[0] == SIGNATURE_ARRAY) {
	AddBinClass(createFakeArrayClass(name));
	return 1;
    }
    if (name[0] == DIR_SEPARATOR) {
	return 0;
    }
    for (cpp = sysGetClassPath(); cpp && *cpp != 0; cpp++) {
	cpe_t *cpe = *cpp;
	char path[255];
	if (cpe->type == CPE_DIR) {
	    if (jio_snprintf(path, sizeof(path), "%s%c%s." JAVAOBJEXT, cpe->u.dir,
			 LOCAL_DIR_SEPARATOR, name) == -1) {
		return 0;
	    }
	    if (LoadFile(sysNativePath(path), cpe->u.dir, SourceHint)) {
		return 1;
	    }
	} else if (cpe->type == CPE_ZIP) {
	    if (jio_snprintf(path, sizeof(path), "%s." JAVAOBJEXT, name) == -1) {
		return 0;
	    }
	    if (LoadFileZip(path, cpe->u.zip)) {
		return 1;
	    }
	}
    }
    return 0;
}

char *
stat_source(ClassClass *cb, struct stat *s, char *pathbuf, int maxlen)
{
#define NAMEBUFLEN 255
    char nm[NAMEBUFLEN];
    char *p, *q, *lp;
    cpe_t **cpp;

    /* don't bother searching if absolute */
    /* REMIND: only here for compatibility */
    if (sysIsAbsolute(classsrcname(cb))) {
	if (sysStat(classsrcname(cb), s) == 0) {
	    if (jio_snprintf(pathbuf, maxlen, "%s", classsrcname(cb)) == -1) {
		return 0;
	    }
	    return pathbuf;
	} else {
	    return 0;
	}
    }

    /* parse the package name */
    p = classname(cb);
    if (strlen(p) > NAMEBUFLEN - 1) {
	return 0;
    }
    for (q = lp = nm ; *p ; p++) {
	if (*p == DIR_SEPARATOR) {
	    *q++ = LOCAL_DIR_SEPARATOR;
	    lp = q;
	} else {
	    *q++ = *p;
	}
    }

    /* append the source file name */
    p = classsrcname(cb);
    if (strlen(p) + (lp - nm) > NAMEBUFLEN - 1) {
	return 0;
    }
    for (; *p ; p++) {
	*lp++ = (*p == DIR_SEPARATOR ? LOCAL_DIR_SEPARATOR : *p);
    }
    *lp = '\0';

    /* search the class path */
    for (cpp = sysGetClassPath() ; cpp && *cpp != 0 ; cpp++) {
	cpe_t *cpe = *cpp;
	if (cpe->type == CPE_DIR) {
	    if (jio_snprintf(pathbuf, maxlen, "%s%c%s",
			 cpe->u.dir, LOCAL_DIR_SEPARATOR, nm) == -1) {
		return 0;
	    }
	    if (sysStat(pathbuf, s) == 0) {
		return pathbuf;
	    }
	}
    }
    return 0;
}

struct CICcontext {
    unsigned char *ptr;
    unsigned char *end_ptr;
    ClassClass *cb;
    jmp_buf jump_buffer;
    char **detail;
};

typedef struct CICcontext CICcontext;

static unsigned long get1byte(CICcontext *);
static unsigned long get2bytes(CICcontext *);
static unsigned long get4bytes(CICcontext *);
static char *getAsciz(CICcontext *);
static char *getAscizFromClass(CICcontext *, int i);
static void getNbytes(CICcontext *, int count, char *buffer);
static void LoadConstantPool(CICcontext *);
static void ReadInCode(CICcontext *, struct methodblock *);
static void ReadLineTable(CICcontext *, struct methodblock *mb);
static void ReadLocalVars(CICcontext *, struct methodblock *mb);
static void ReadInCompiledCode(CICcontext *context, struct methodblock *mb);
static void InitializeStaticVar(struct fieldblock *fb, CICcontext *context);


#undef  ERROR		/* Avoid conflict on PC */
#define ERROR(name) { *(context->detail) = name; \
                        longjmp(context->jump_buffer, 1); }

bool_t 
createInternalClass(unsigned char *ptr, unsigned char *end_ptr, ClassClass *cb)
{
    int i, j;
    union cp_item_type *constant_pool;
    unsigned char *type_table;
    int attribute_count;
    unsigned fields_count;
    struct methodblock *mb;
    struct fieldblock *fb;
    struct CICcontext context_block;
    struct CICcontext *context = &context_block;
    char *detail = 0;

    context->ptr = ptr;
    context->end_ptr = end_ptr;
    context->cb = cb;
    context->detail = &detail;
    if (setjmp(context->jump_buffer)) {
	/* Indicate an error of some sort */
	return FALSE;
    }
    
    if (get4bytes(context) != JAVA_CLASSFILE_MAGIC) 
	ERROR("Bad magic number");

    cb->minor_version = (unsigned short) get2bytes(context);
    cb->major_version = (unsigned short) get2bytes(context);
    if (cb->major_version != JAVA_VERSION) 
	ERROR("Bad major version number");

    LoadConstantPool(context);
    constant_pool = cb->constantpool;
    type_table = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;

    cb->access = (unsigned short) get2bytes(context) & ACC_WRITTEN_FLAGS;

    /* Get the name of the class */
    i = (int) get2bytes(context);	/* index in constant pool of class */
    cb->name = getAscizFromClass(context, i);
    constant_pool[i].p = cb;
    CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);

    /* Get the super class name. */
    i = (int) get2bytes(context);	/* index in constant pool of class */
    if (i > 0) 
	cb->super_name = getAscizFromClass(context, i);


    i = cb->implements_count = (unsigned short) get2bytes(context);
    if (i > 0) {
	unsigned j;
	cb->implements = sysCalloc(i, sizeof(short));
	for (j = 0; j < i; j++) {
	    cb->implements[j] = (short) get2bytes(context);
	}
    }

    fields_count = cb->fields_count = (unsigned short) get2bytes(context);
    if (fields_count > 0) 
        cb->fields = (struct fieldblock *)
	       sysCalloc(cb->fields_count, sizeof(struct fieldblock));
    for (i = fields_count, fb = cb->fields; --i >= 0; fb++) {
	fieldclass(fb) = cb;
	fb->access = (unsigned short) get2bytes(context) & ACC_WRITTEN_FLAGS;
	fb->name = getAsciz(context);
	fb->signature = getAsciz(context);
	attribute_count = (int)get2bytes(context);
	for (j = 0; j < (int)attribute_count; j++) {
	    char *name = getAsciz(context);
	    int length = (int) get4bytes(context);
	    if ((fb->access & ACC_STATIC) 
		 && strcmp(name, "ConstantValue") == 0) {
		if (length != 2) 
		    ERROR("Wrong size for VALUE attribute");
		fb->access |= ACC_VALKNOWN;
		/* we'll change this below */
		fb->u.offset = get2bytes(context); 
	    } else {
		getNbytes(context, length, NULL);
	    }
	}
	if (fb->access & ACC_STATIC) {
	    InitializeStaticVar(fb, context);
	}
    }

    if ((cb->methods_count = (unsigned short) get2bytes(context)) > 0) 
	cb->methods = (struct methodblock *)
	            sysCalloc(cb->methods_count, sizeof(struct methodblock));
    for (i = cb->methods_count, mb = cb->methods; --i >= 0; mb++) {
	bool_t seenCompiled = FALSE;
	fieldclass(&mb->fb) = cb;
	mb->fb.access = (unsigned short) get2bytes(context) & ACC_WRITTEN_FLAGS;
	mb->fb.name = getAsciz(context);
	mb->fb.signature = getAsciz(context);
	mb->args_size = Signature2ArgsSize(mb->fb.signature) 
	                + ((mb->fb.access & ACC_STATIC) ? 0 : 1);
	attribute_count = (int) get2bytes(context);
	for (j = 0; j < attribute_count; j++) {
	    extern char *CompiledCodeAttribute;
	    char *name = getAsciz(context);
	    if (CompiledCodeAttribute != NULL 
		  && !strcmp(name, CompiledCodeAttribute) 
		  && cbLoader(cb) == 0) {
		ReadInCompiledCode(context, mb);
		seenCompiled = TRUE;
	    } else if ((strcmp(name, "Code") == 0) 
		       && !seenCompiled 
		       && ((mb->fb.access & (ACC_NATIVE | ACC_ABSTRACT))==0)) {
		ReadInCode(context, mb);
	    } else {
		int length = (int) get4bytes(context);
		getNbytes(context, length, NULL);
	    }
	}
    }

    /* See if there are class attributes */
    attribute_count = (int) get2bytes(context); 
    for (j = 0; j < attribute_count; j++) {
	char *name = getAsciz(context);
	int length = (int) get4bytes(context);
	if (strcmp(name, "SourceFile") == 0) {
	    if (length != 2) {
		ERROR("Wrong size for VALUE attribute");
	    }
	    cb->source_name = getAsciz(context);
	} else {
	    getNbytes(context, length, NULL);
	}
    }
    return TRUE;
}

static void InitializeStaticVar(struct fieldblock *fb, CICcontext *context) {
    Java8 t1, t2;
    char isig = fb->signature[0];
    if (fb->access & ACC_VALKNOWN) { 
	ClassClass *cb = context->cb;
	int index = (int) fb->u.offset; /* set by "VALUE" attribute */
	union cp_item_type *constant_pool = cb->constantpool;
	unsigned char *type_table = 
	      constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;

	if (index <= 0 || index >= cb->constantpool_count) 
	    ERROR("Bad initial value");
	switch (isig) { 
	    case SIGNATURE_DOUBLE: 
		if (type_table[index] != 
		    (CONSTANT_Double | CONSTANT_POOL_ENTRY_RESOLVED))
		    ERROR("Bad index into constant pool");
		twoword_static_address(fb) = sysMalloc(2 * sizeof(long));
		SET_DOUBLE(t1, twoword_static_address(fb),
			   GET_DOUBLE(t2, &constant_pool[index]));
	    break;

	    case SIGNATURE_LONG: 
		if (type_table[index] != 
		    (CONSTANT_Long | CONSTANT_POOL_ENTRY_RESOLVED))
		    ERROR("Bad index into constant pool");
		twoword_static_address(fb) = sysMalloc(2 * sizeof(long));
		SET_INT64(t1, twoword_static_address(fb),
			  GET_INT64(t2, &constant_pool[index]));
		break;

	    case SIGNATURE_CLASS:
		if (strcmp(fb->signature, "L" JAVAPKG "String;") != 0) {
		    ERROR("Cannot set initial value for object");
		} else if (!ResolveClassStringConstant(cb, index, 0)) { 
		    ERROR("Unable to resolve string");
		} else {
		    *(void **)normal_static_address(fb) = 
			   constant_pool[index].p;
		}
		break;

	    case SIGNATURE_BYTE: 
		if (type_table[index] != 
		    (CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED))
		    ERROR("Bad index into constant pool");
		*(int *)normal_static_address(fb) = 
		        (signed char)(constant_pool[index].i);
		break;

	    case SIGNATURE_CHAR:
		if (type_table[index] != 
		    (CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED))
		    ERROR("Bad index into constant pool");
		*(int *)normal_static_address(fb) = 
		        (unsigned short)(constant_pool[index].i);
		break;

	    case SIGNATURE_SHORT:
		if (type_table[index] != 
		    (CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED))
		    ERROR("Bad index into constant pool");
		*(int *)normal_static_address(fb) = 
		        (signed short)(constant_pool[index].i);
		break;
	    case SIGNATURE_BOOLEAN:
		if (type_table[index] != 
		    (CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED))
		    ERROR("Bad index into constant pool");
		*(int *)normal_static_address(fb) = 
		       (constant_pool[index].i != 0);
		break;

	    case SIGNATURE_INT:  
		if (type_table[index] != 
		    (CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED))
		    ERROR("Bad index into constant pool");
		*(int *)normal_static_address(fb) = constant_pool[index].i;
		break;
		
	    case SIGNATURE_FLOAT:
		if (type_table[index] != 
		        (CONSTANT_Float | CONSTANT_POOL_ENTRY_RESOLVED))
		    ERROR("Bad index into constant pool");
		*(float *)normal_static_address(fb) = constant_pool[index].f;
		break;
			
	    default: 
		ERROR("Unable to set initial value");
	}
    } else { 
	switch (isig) { 
	    default:
		*(int *)normal_static_address(fb) = 0;
		break;

	    case SIGNATURE_FLOAT:
		*(float *)normal_static_address(fb) = (float)0.0;
		break;

	    case SIGNATURE_LONG: 
		twoword_static_address(fb) = sysMalloc(2 * sizeof(long));
		SET_INT64(t1, twoword_static_address(fb), ll_zero_const);
		break;

	    case SIGNATURE_DOUBLE: 
		twoword_static_address(fb) = sysMalloc(2 * sizeof(long));
		SET_DOUBLE(t1, twoword_static_address(fb), 0.0);
		break;

	    case SIGNATURE_CLASS:
		*normal_static_address(fb) = 0;
		break;
	}
    }
}




static void LoadConstantPool(CICcontext *context) 
{
    ClassClass *cb = context->cb;
    int nconstants = (int) get2bytes(context);
    unsigned char *initial_ptr = context->ptr;
    cp_item_type *constant_pool;
    unsigned char *type_table;
    char *malloced_space;
    int malloc_size;
    int i;
    Java8 t1;
    
    if (nconstants < CONSTANT_POOL_UNUSED_INDEX) {
	ERROR("Illegal constant pool size");
    }
    
    constant_pool = (cp_item_type *)sysCalloc(nconstants, sizeof(cp_item_type));
    type_table = sysCalloc(nconstants, sizeof(char));
    cb->constantpool = constant_pool;
    cb->constantpool_count = nconstants;

    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p = type_table;
    for (malloc_size = 0, i = CONSTANT_POOL_UNUSED_INDEX; i < nconstants; i++) {
	int type = (int) get1byte(context);
	switch (type) {
	  case CONSTANT_Utf8: {
	      int length = (int) get2bytes(context);
	      malloc_size += length + 1;
	      getNbytes(context, length, NULL);
	      break;
	  }
	    
	  case CONSTANT_Class:
	  case CONSTANT_String:
	      getNbytes(context, 2, NULL);
	      break;

	  case CONSTANT_Fieldref:
	  case CONSTANT_Methodref:
	  case CONSTANT_InterfaceMethodref:
	  case CONSTANT_NameAndType:
	  case CONSTANT_Integer:
	  case CONSTANT_Float:
	      getNbytes(context, 4, NULL);
	      break;

	  case CONSTANT_Long:
	  case CONSTANT_Double: 
	      i++;
	      getNbytes(context, 8, NULL);
	      break;

	  default: 
	      ERROR("Illegal constant pool type");
	}
    }

    malloced_space = (char *)sysMalloc(malloc_size);
    
    context->ptr = initial_ptr;

    for (i = CONSTANT_POOL_UNUSED_INDEX; i < nconstants; i++) {
	int type = (int) get1byte(context);
	CONSTANT_POOL_TYPE_TABLE_PUT(type_table, i, type);
	switch (type) {
	  case CONSTANT_Utf8: {
	      int length = (int) get2bytes(context);
	      char *result = malloced_space;
	      malloced_space += (length + 1);
	      getNbytes(context, length, result);
	      result[length] = '\0';
	      constant_pool[i].p = result;
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	      break;
	  }
	
	  case CONSTANT_Class:
	  case CONSTANT_String:
	    constant_pool[i].i = (int) get2bytes(context);
	    break;

	  case CONSTANT_Fieldref:
	  case CONSTANT_Methodref:
	  case CONSTANT_InterfaceMethodref:
	  case CONSTANT_NameAndType:
	    constant_pool[i].i = (int) get4bytes(context);
	    break;
	    
	  case CONSTANT_Integer:
	  case CONSTANT_Float:
	    constant_pool[i].i = (int) get4bytes(context);
	    CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	    break;

	  case CONSTANT_Long:
	  case CONSTANT_Double: {
	      unsigned high = (unsigned int) get4bytes(context);
	      unsigned low = (unsigned int) get4bytes(context);
	      int64_t value;
	      value = ll_add(ll_shl(uint2ll(high), 32), uint2ll(low));
	      SET_INT64(t1, &constant_pool[i], value);
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	      i++;		/* increment i for loop, too */
	      /* Indicate that the next object in the constant pool cannot
	       * be accessed independently.    */
	      CONSTANT_POOL_TYPE_TABLE_PUT(type_table, i, 0);
	      CONSTANT_POOL_TYPE_TABLE_SET_RESOLVED(type_table, i);
	      break;
	  }

	  default:
	      ERROR("Illegal constant pool type");
	}
    }
}



static void ReadInCode(CICcontext *context, struct methodblock *mb)
{
    int attribute_length = (int) get4bytes(context);
    unsigned char *end_ptr = context->ptr + attribute_length;
    int attribute_count;
    int code_length;
    int i;

    if (context->cb->minor_version <= 2) { 
	mb->maxstack = (unsigned short) get1byte(context);
	mb->nlocals = (unsigned short) get1byte(context);
	code_length = (int) (mb->code_length = (unsigned long) get2bytes(context));
    } else { 
	mb->maxstack = (unsigned short) get2bytes(context);
	mb->nlocals = (unsigned short) get2bytes(context);
	code_length = (int) (mb->code_length = (unsigned long) get4bytes(context));
    }
    mb->code = sysMalloc(code_length);

    getNbytes(context, code_length, (char *)mb->code);
    if ((mb->exception_table_length = get2bytes(context)) > 0) {
	unsigned exception_table_size = (unsigned int) mb->exception_table_length 
	                                  * sizeof(struct CatchFrame);
	mb->exception_table = sysMalloc(exception_table_size);
	for (i = 0; i < (int)mb->exception_table_length; i++) {
	    mb->exception_table[i].start_pc = get2bytes(context);
	    mb->exception_table[i].end_pc = get2bytes(context);		
	    mb->exception_table[i].handler_pc = get2bytes(context); 
	    mb->exception_table[i].catchType = (short) get2bytes(context);
	}
    }
    attribute_count = (int) get2bytes(context);
    for (i = 0; i < attribute_count; i++) {
	char *name = getAsciz(context);
	if (strcmp(name, "LineNumberTable") == 0) {
	    ReadLineTable(context, mb);
	} else if (strcmp(name, "LocalVariableTable") == 0) {
	    ReadLocalVars(context, mb);
	} else {
	    int length = (int) get4bytes(context);
	    getNbytes(context, length, NULL);
	}
    }
    if (context->ptr != end_ptr) 
	ERROR("Code segment was wrong length");
}


static void ReadLineTable(CICcontext *context, struct methodblock *mb)
{
    int attribute_length = (int) get4bytes(context);
    unsigned char *end_ptr = context->ptr  + attribute_length;
    int i;
    if ((mb->line_number_table_length = get2bytes(context)) > 0) {
	struct lineno *ln = 
	    sysMalloc((int) mb->line_number_table_length * sizeof(struct lineno));
	mb->line_number_table = ln;
	for (i = (int) mb->line_number_table_length; --i >= 0; ln++) {
	    ln->pc = get2bytes(context);
	    ln->line_number = get2bytes(context);
	}
    }
    if (context->ptr != end_ptr)
	ERROR("Line number table was wrong length?");
}	

static void ReadLocalVars(CICcontext *context, struct methodblock *mb)
{
    int attribute_length = (int) get4bytes(context);
    unsigned char *end_ptr = context->ptr  + attribute_length;
    int i;
    if ((mb->localvar_table_length = get2bytes(context)) > 0) {
	struct localvar *lv = 
	        sysMalloc((int) mb->localvar_table_length * sizeof(struct localvar));
	mb->localvar_table = lv;
	for (i = (int) mb->localvar_table_length; --i >= 0; lv++) {
	    lv->pc0 = get2bytes(context);
	    lv->length = get2bytes(context);
	    lv->nameoff = (short) get2bytes(context);
	    lv->sigoff = (short) get2bytes(context);
	    lv->slot = get2bytes(context);
	}
    }
    if (context->ptr != end_ptr)
	ERROR("Local variables table was wrong length?");
}	

/* Read in Compiled code.  The code has a very strict format:
 *    4 bytes:  length of code
 *    < code >
 *    4 bytes:  length of patches
 *    < patches>
 */
static void 
ReadInCompiledCode(CICcontext *context, struct methodblock *mb) {
    int attribute_length = (int) get4bytes(context);
    unsigned char *end_ptr = context->ptr + attribute_length;
    int code_length = (int) get4bytes(context);
    char *code = sysMalloc(code_length); 
    int patch_length;
    unsigned char *patch;
    mb->code = 0;
    mb->CompiledCode = code;
    mb->code_length = code_length;
    getNbytes(context, code_length, code);

    mb->maxstack = mb->nlocals = 0;
    patch_length = (int) get4bytes(context);
    mb->code = patch = sysMalloc(patch_length + 1);
    getNbytes(context, patch_length, (char *)patch);
    patch[patch_length] = 0;
    if (context->ptr != end_ptr) 
	ERROR("Compiled Code segment was wrong length");
    mb->fb.access |= ACC_MACHINE_COMPILED;
}

unsigned Signature2ArgsSize(char *method_signature)
{
    char *p;
    int args_size = 0;
    for (p = method_signature; *p != SIGNATURE_ENDFUNC; p++) {
	switch (*p) {
	  case SIGNATURE_BOOLEAN:
	  case SIGNATURE_BYTE:
	  case SIGNATURE_CHAR:
	  case SIGNATURE_SHORT:
	  case SIGNATURE_INT:
	  case SIGNATURE_FLOAT:
	    args_size += 1;
	    break;
	  case SIGNATURE_CLASS:
	    args_size += 1;
	    while (*p != SIGNATURE_ENDCLASS) p++;
	    break;
	  case SIGNATURE_ARRAY:
	    args_size += 1;
	    while ((*p == SIGNATURE_ARRAY)) p++;
	    /* If an array of classes, skip over class name, too. */
	    if (*p == SIGNATURE_CLASS) { 
		while (*p != SIGNATURE_ENDCLASS) 
		  p++;
	    } 
	    break;
	  case SIGNATURE_DOUBLE:
	  case SIGNATURE_LONG:
	    args_size += 2;
	    break;
	  case SIGNATURE_FUNC:	/* ignore initial (, if given */
	    break;
	  default:
	    /* Indicate an error. */
	    return 0;
	}
    }
    return args_size;
}


ClassClass*
createFakeArrayClass(char *name) 
{
    char *name_p;
    ClassClass *cb = sysCalloc(sizeof(ClassClass), 1);
    int depth, base_type;
    char buffer_space[256];
    char *buffer = buffer_space;
    cp_item_type *constant_pool = sysCalloc(CONSTANT_POOL_ARRAY_LENGTH, 
					    sizeof(cp_item_type));
    unsigned char *type_table = sysCalloc(CONSTANT_POOL_ARRAY_LENGTH,
					  sizeof(char));
    
    sysAssert(name[0] == SIGNATURE_ARRAY);

    if (strlen(name) + 1 > sizeof(buffer_space)) { 
	/* don't allow buffer overflow  */
	buffer = sysMalloc(strlen(name));
    }

    for (name_p = name, depth = 0; *name_p == SIGNATURE_ARRAY; 
	 name_p++, depth++);
    switch(*name_p) {
        case SIGNATURE_INT:    base_type = T_INT; break;
	case SIGNATURE_LONG:   base_type = T_LONG; break;
	case SIGNATURE_FLOAT:  base_type = T_FLOAT; break;
	case SIGNATURE_DOUBLE: base_type = T_DOUBLE; break;
	case SIGNATURE_BOOLEAN:base_type = T_BOOLEAN; break; 
	case SIGNATURE_BYTE:   base_type = T_BYTE; break;
	case SIGNATURE_CHAR:   base_type = T_CHAR; break;
	case SIGNATURE_SHORT:  base_type = T_SHORT; break;
	case SIGNATURE_CLASS: {
	    char *bp;
	    for (bp = buffer, name_p++; /* skip over SIGNATURE_ARRAY */
		 *name_p != SIGNATURE_ENDCLASS; *bp++ = *name_p++);
	    *bp = '\0';
	    base_type = T_CLASS;
	}
    }
    cb->major_version = JAVA_VERSION;
    cb->minor_version = JAVA_MINOR_VERSION;
    cb->access = ACC_FINAL | ACC_ABSTRACT | ACC_PUBLIC;
    cb->name = strdup(name);
    cb->super_name = JAVAPKG "Object";
    cb->constantpool = constant_pool;
    cb->constantpool_count = CONSTANT_POOL_ARRAY_LENGTH;

    constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p = type_table;
    constant_pool[CONSTANT_POOL_ARRAY_DEPTH_INDEX].i = depth;     
    constant_pool[CONSTANT_POOL_ARRAY_TYPE_INDEX].i = base_type;
    type_table[CONSTANT_POOL_ARRAY_DEPTH_INDEX] = 
	type_table[CONSTANT_POOL_ARRAY_TYPE_INDEX] = 
	    CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED;
    
    if (base_type == T_CLASS) {
	constant_pool[CONSTANT_POOL_ARRAY_CLASS_INDEX].i = 
	      CONSTANT_POOL_ARRAY_CLASSNAME_INDEX;
	constant_pool[CONSTANT_POOL_ARRAY_CLASSNAME_INDEX].cp = strdup(buffer);
	type_table[CONSTANT_POOL_ARRAY_CLASS_INDEX] = CONSTANT_Class;
	type_table[CONSTANT_POOL_ARRAY_CLASSNAME_INDEX] = 
	      CONSTANT_Utf8 | CONSTANT_POOL_ENTRY_RESOLVED;
    } else {
	type_table[CONSTANT_POOL_ARRAY_CLASS_INDEX] = 
	    type_table[CONSTANT_POOL_ARRAY_CLASSNAME_INDEX] = 
		CONSTANT_Integer | CONSTANT_POOL_ENTRY_RESOLVED;
    }
    if (buffer != buffer_space) { 
	/* free buffer, if we malloc'ed it */
	sysFree(buffer);
    } 
    return cb;
}


void FreeClass(ClassClass *cb) 
{
    int i;
    unsigned char *type_table;
    struct methodblock *mb;
    struct fieldblock *fb;

    sysFree(cb->implements);
    for (i = cb->fields_count, fb = cb->fields; --i >= 0; fb++) {
	if (fb->access & ACC_STATIC) {
	    if (fb->signature[0] == SIGNATURE_DOUBLE || fb->signature[0] == SIGNATURE_LONG) {
		sysFree(twoword_static_address(fb));
	    }
	}
    }
    sysFree(cb->fields);
    for (i = cb->methods_count, mb = cb->methods; --i >= 0; mb++) {
	if (mb->code_length) /* not external */
	    sysFree(mb->code);
	sysFree(mb->exception_table);
	sysFree(mb->localvar_table);
	sysFree(mb->line_number_table);
    }
    sysFree(cb->methods);
    if (cb->constantpool) {
	type_table = cb->constantpool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
	for (i = CONSTANT_POOL_UNUSED_INDEX; 
	         i < (int)cb->constantpool_count; i++) {
	    /* The first item in the constant pool of type CONSTANT_Utf8
	     * has the malloc of all the strings. */
	    if ((type_table[i] & CONSTANT_POOL_ENTRY_TYPEMASK) == 
		    CONSTANT_Utf8) {
		sysFree(cb->constantpool[i].p);
		break;
	    }
	}
	sysFree(cb->constantpool);
    }
    sysFree(cb);
}
	
static unsigned long get1byte(CICcontext *context)
{
    unsigned char *ptr = context->ptr;
    if (context->end_ptr - ptr < 1) {
	ERROR("Truncated class file");
    } else {
	unsigned char *ptr = context->ptr;
	unsigned long value = ptr[0];
	(context->ptr) += 1;
	return value;
    }
}

static unsigned long get2bytes(CICcontext *context)
{
    unsigned char *ptr = context->ptr;
    if (context->end_ptr - ptr < 2) {
	ERROR("Truncated class file");
    } else {
	unsigned long value = (ptr[0] << 8) + ptr[1];
	(context->ptr) += 2;
	return value;
    }
}


static unsigned long get4bytes(CICcontext *context)
{
    unsigned char *ptr = context->ptr;
    if (context->end_ptr - ptr < 4) {
	ERROR("Truncated class file");
    } else {
	unsigned long value = (ptr[0] << 24) + (ptr[1] << 16) + 
	                             (ptr[2] << 8) + ptr[3];
	(context->ptr) += 4;
	return value;
    }
}

static void getNbytes(CICcontext *context, int count, char *buffer) 
{
    unsigned char *ptr = context->ptr;
    if (context->end_ptr - ptr < count)
	ERROR("Truncated class file");
    if (buffer != NULL) 
	memcpy(buffer, ptr, count);
    (context->ptr) += count;
}


static char *getAsciz(CICcontext *context) 
{
    ClassClass *cb = context->cb;
    union cp_item_type *constant_pool = cb->constantpool;
    int nconstants = cb->constantpool_count;
    unsigned char *type_table = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;

    int value = (int) get2bytes(context);
    if ((value == 0) || (value >= nconstants) || 
	   type_table[value] != (CONSTANT_Utf8 | CONSTANT_POOL_ENTRY_RESOLVED))
	ERROR("Illegal constant pool index");
    return constant_pool[value].cp;
}

static char *getAscizFromClass(CICcontext *context, int value) 
{
    ClassClass *cb = context->cb;
    union cp_item_type *constant_pool = cb->constantpool;
    int nconstants = cb->constantpool_count;
    unsigned char *type_table = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
    if ((value <= 0) || (value >= nconstants) 
	|| (type_table[value] != CONSTANT_Class))
	ERROR("Illegal constant pool index");
    value = constant_pool[value].i;
    if ((value <= 0) || (value >= nconstants) ||
	 (type_table[value] != (CONSTANT_Utf8 | CONSTANT_POOL_ENTRY_RESOLVED)))
	ERROR("Illegal constant pool index");
    return constant_pool[value].cp;
}

/* In certain cases, we only need a class's name.  We don't want to resolve
 * the class reference if it isn't, already. 
 */

char *GetClassConstantClassName(cp_item_type *constant_pool, int index)
{
    unsigned char *type_table = constant_pool[CONSTANT_POOL_TYPE_TABLE_INDEX].p;
    switch(type_table[index]) {
        case CONSTANT_Class | CONSTANT_POOL_ENTRY_RESOLVED: {
	    ClassClass *cb = constant_pool[index].p;
	    return classname(cb);
	}

        case CONSTANT_Class: {
	    int name_index = constant_pool[index].i;
	    return constant_pool[name_index].cp;
	}
	    
	default:
	    return (char *)0;
    }
}
