/*
 * @(#)javai.c	1.52 95/12/22
 *
 * Copyright (c) 1994 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Interpret an java class file
 */

#include <windows.h>
#include <memory.h>
#include <process.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include "threads.h"
#include "threads_md.h"
#include "monitor.h"
#include "monitor_md.h"
#include "tree.h"
#include "oobj.h"
#include "interpreter.h"
#include "javaString.h"
#include "log.h"
#include "bool.h"
#include "path_md.h"

#include "java_lang_Thread.h"

/*
 * Default values for minimum and maximum amounts of memory to devote
 * to the Java heap.  The current abstraction, which may not hold up
 * forever, is that the initial size of the heap is the minimum heap.
 *
 * The maximum memory to map: As we're mmapping the memory, it looks
 * like we should ask swapctl(2) (or equivalent on Win32) how much swap
 * we have, and thus what the maximum we've got is.
 *
 * These values can be overridden on the java command line.
 */
#ifdef USE_MALLOC
#define MINHEAPSIZE ((int) ( 3 * (1024*1024)))
#define MAXHEAPSIZE MINHEAPSIZE
#else
#define MINHEAPSIZE ((int) ( 3 * (1024*1024)))
#define MAXHEAPSIZE ((int) (64 * (1024*1024)))
#endif

/*
 * The default thread stack size.  This is machine-dependent, because it
 * is based on a platform's C stack usage.
 */
#define PROCSTACKSIZE (128 * 1024)     /* Default size of a thread C stack */
long ProcStackSize = PROCSTACKSIZE;

#define AGENT_CLASS	"sun/tools/debug/Agent"

#ifdef DEBUG
#define JAVAC_EXE	"javac.exe"
#else
#define JAVAC_EXE	"javac_g.exe"
#endif

void usage(void);

static long
mtoi(char *cp)
{
    long r = strtoul(cp, &cp, 10);
    switch (*cp) {
    case 'M': case 'm':
	r *= 1024;
    case 'K': case 'k':
	r *= 1024;
    }
    return r;
}

/* add a user property */
static void add_user_prop(char *def)
{
    extern char **user_props;
    static int nprops = 0;
    static int max_props = 0;
    char *p, *str;

    /* scan to the '=' */
    for (p = def ; *p && *p != '=' ; p++);

    /* grow the array (if needed) */
    if (nprops + 2 > max_props) {
        if (user_props == 0) {
            user_props = (char **)calloc(16, sizeof(char *));
            max_props = 16;
        } else {
            char **new_props = (char **)calloc(2 * max_props, sizeof(char *));
            memcpy(new_props, user_props, nprops * sizeof(char *));
            free(user_props);
            user_props = new_props;
            max_props *= 2;
        }
    }

    /* malloc the key */
    str = (char *)malloc((p - def) + 1);
    strncpy(str, def, p - def);
    str[p - def] = '\0';
    user_props[nprops++] = str;

    /* malloc the value */
    if (*p == '=') {
        p++;
    }
    str = (char *)malloc(strlen(p) + 1);
    strcpy(str, p);
    user_props[nprops++] = str;
}

static struct methodblock *
FindMethod(ClassClass *cb, char *name, char *sig)
{
    int i = cb->methods_count;
    struct methodblock *mb = cbMethods(cb);

    while (--i >= 0) {
	struct fieldblock *fb = &mb->fb;
	if (strcmp(fieldname(fb), name) == 0 &&
	    strcmp(fieldsig(fb), sig) == 0) {
	    return mb;
	}
	mb++;
    }
    return NULL;
}

/*
 * Build argument array to be passed to "main"
 */
static HArrayOfString *
BuildArguments(int argc, char **argv, struct execenv *ee)
{
    HArrayOfString *args = (HArrayOfString *)ArrayAlloc(T_CLASS, argc);

    if (args == NULL) {
	SignalError(ee, JAVAPKG "OutOfMemoryError", 0);
	return NULL;
    }
    /* Give the array a "type" */
    unhand((HArrayOfObject *)args)->body[argc] = 
                        (JHandle *)FindClass(ee, JAVAPKG "String", TRUE);
    while (--argc >= 0) {
	char *s = argv[argc];
	if ((unhand(args)->body[argc] = makeJavaString(s, strlen(s))) == NULL) {
	    SignalError(ee, JAVAPKG "OutOfMemoryError", 0);
	    return NULL;
	}
    }
    return args;
}

/*
 * This function is included primarily as a debugging aid. If Java is
 * running in a console window, then pressing <CTRL-BREAK> will cause
 * the current state of all active threads and monitors to be written
 * to the console window.
 */
static void
BreakHandler(int junk)
{
    DumpThreads();
    DumpMonitors();
    /* We need to reinstate the signal handler each time... */
    signal(SIGBREAK, BreakHandler);
}

int
java_main(int argc, char *argv[])
{
    DWORD regs[N_TRACED_REGS];
    int MinHeapSize = MINHEAPSIZE;
    int MaxHeapSize = MAXHEAPSIZE;
    HThread *self;
    struct methodblock *mb;
    ExecEnv ee;
    char *errmsg, *classname, *cp;
    long debugport = 0;
    ClassClass *cb;
    HArrayOfString *args;

    if ((progname = strrchr(argv[0], '\\')) != NULL) {
	progname++;
    } else {
	progname = argv[0];
    }

    SkipSourceChecks = TRUE;
    verifyclasses = VERIFY_REMOTE;

    /*
     * Read command line options
     */
    while (--argc > 0 && (++argv)[0][0] == '-') {
#ifndef trace
	if (strcmp(*argv, "-t") == 0) {
	    trace++;
	} else if (strcmp(*argv, "-tm") == 0) {
	    tracem++;
	} else
#endif
	if (strcmp(*argv, "-v") == 0 || strcmp(*argv, "-verbose") == 0) {
	    verbose++;
	} else if (strcmp(*argv, "-debug") == 0) {
	    debugagent++;
	} else if (strcmp(*argv, "-debugport") == 0) {
	    if (argc > 1) {
		debugport = atol(argv[1]);
		--argc; argv++;
	    } else {
		fprintf(stderr, "-debugport requires port number\n");
		usage();
		return 1;
	    }
	} else if (strcmp(*argv, "-noasyncgc") == 0) {
	    noasyncgc++;
	} else if (strcmp(*argv, "-verbosegc") == 0) {
	    verbosegc++;
	} else if (strcmp(*argv, "-verify") == 0) {
	    verifyclasses = VERIFY_ALL;
	} else if (strcmp(*argv, "-verifyremote") == 0) {
	    verifyclasses = VERIFY_REMOTE;
	} else if (strcmp(*argv, "-noverify") == 0) {
	    verifyclasses = VERIFY_NONE;
	} else if (strcmp(*argv, "-tracegc") == 0) {
	    tracegc++;
	} else if (strcmp(*argv, "-prof") == 0) {
	    sysAtexit(java_mon_dump);
	    javamon(1);
	} else if (strcmp(*argv, "-cs") == 0 ||
		   strcmp(*argv, "-checksource") == 0) {
	    SkipSourceChecks = 0;
	} else if (strncmp(*argv, "-ss", 3) == 0) {
	    long n = mtoi(&argv[0][3]);
	    if (n > 1000) {
		ProcStackSize = n;
	    }
	} else if (strncmp(*argv, "-oss", 4) == 0) {
	    long n = mtoi(&argv[0][4]);
	    if (n > 1000) {
		JavaStackSize = n;
	    }
	} else if (strncmp(*argv, "-ms", 3) == 0) {
	    long n = mtoi(&argv[0][3]);
	    if (n > 1000) {
		MinHeapSize = n;
	    }
	} else if (strncmp(*argv, "-mx", 3) == 0) {
	    long n = mtoi(&argv[0][3]);
	    if (n > 1000) {
		MaxHeapSize = n;
	    }
	} else if (strncmp(*argv, "-D", 2) == 0) {
	    add_user_prop(*argv + 2);
#ifdef LOGGING
	} else if (strncmp(*argv, "-l", 2) == 0) {
	    logging_level = mtoi(&argv[0][2]);
#endif
	} else if (strcmp(*argv, "-classpath") == 0) {
	    if (argc > 1) {
		char *buf = (char *)malloc(strlen(argv[1]) + 32);
		sprintf(buf, "CLASSPATH=%s", argv[1]);
		putenv(buf);
		--argc; argv++;
	    } else {
		fprintf(stderr,
			"-classpath requires class path specification\n");
		usage();
		return 1;
	    }
	} else if (strcmp(*argv, "-h") == 0 || strcmp(*argv, "-help") == 0||
		   strcmp(*argv, "-?") == 0) {
	    usage();
	    return 0;
        } else if (strcmp(*argv, "-version") == 0) {
            fprintf(stderr, "%s version \"%s\"\n", progname, RELEASE);
            return 0;
	} else {
	    fprintf(stderr, "%s: illegal argument\n", *argv);
	    usage();
	    return 1;
	}
    }

    if (--argc < 0) {
	usage();
	return 1;
    }
    classname = *argv++;

    if (strchr(classname, '\\') != NULL) {
	fprintf(stderr, "Invalid class name '%s'\n", classname);
	usage();
	return 1;
    }

    /* Allow use of '.' in class names (convert to '/') */
    for (cp = classname; *cp; cp++) {
	if (*cp == '.') *cp = '/';
    }

    /*
     * Set the base of the primordial thread's stack, which must also
     * mark the end of the register save area.
     */
    mainstktop = (stackp_t)(regs + N_TRACED_REGS);

    monitorRegistryInit();	/* Initialize the monitor registry */
    monitorCacheInit();		/* Create the monitor cache */

    InitializeExecEnv(&ee, 0);  /* Call before InitializeMem() */
    InitializeMem();		/* Could throw exception on open() */
#ifdef USE_MALLOC
    InitializeAlloc(MaxHeapSize, MaxHeapSize);
#else
    InitializeAlloc(MaxHeapSize, MinHeapSize);
#endif
    InitializeInterpreter();	/* Call before InitializeClassThread() */
    
    self = InitializeClassThread(&ee, &errmsg);
    if (self == 0) {
	fprintf(stderr, "Unable to initialize threads: %s\n", errmsg);
	sysExit(1);
    }
    setThreadName(self, MakeString("main", 4));

    cb = FindClass(&ee, classname, TRUE);
    if (cb == NULL) {
	fprintf(stderr, "Can't find class %s\n", classname);
	sysExit(1);
    }

    if (debugagent) {
	ClassClass *cb = FindClass(&ee, AGENT_CLASS, TRUE);
	if (cb == NULL) {
	    fprintf(stderr, "Can't find class " AGENT_CLASS "\n");
	    sysExit(1);
	}
	execute_java_static_method(0, cb, "boot", "(I)V", debugport);
    }

    InitializeMainThread();	/* Finish initialization of the main thread */

    /* try to load the compiler package since it loads some hooks for
       jit compilers. If it fails it's ok. */
    FindClass(&ee, JAVAPKG "Compiler", TRUE);

    mb = FindMethod(cb, "main", "([Ljava/lang/String;)V");
    if (mb == NULL) {
	fprintf(stderr, "In class %s: void main(String argv[]) is undefined\n",
		classname);
	sysExit(1);
    }

    if ((mb->fb.access & (ACC_STATIC|ACC_PUBLIC)) != (ACC_STATIC|ACC_PUBLIC)) {
	fprintf(stderr, "In class %s: main must be public and static\n",
		classname);
	sysExit(1);
    }

    args = BuildArguments(argc, argv, &ee);
    if (args == NULL) {
	sysExit(1);
    }

    signal(SIGBREAK, BreakHandler);

    do_execute_java_method(&ee, cb, 0, 0, mb, TRUE, (JHandle *)args);

    if (exceptionOccurred(&ee)) {
	if (ee.thread && THREAD(ee.thread)->group) {
	    void *t = (void *)ee.exception.exc;
	    exceptionClear(&ee);
	    execute_java_dynamic_method(&ee, (void *)THREAD(ee.thread)->group, 
		"uncaughtException", "(Ljava/lang/Thread;Ljava/lang/Throwable;)V", ee.thread, t);
	}
    }

    /*
     * The primordial thread is not allowed to exit until all user
     * threads have exited.  Note that the primordial thread's
     * resources (e.g. its Java stack) have not been reclaimed at this
     * point, even though it's got no chance of running again.  The
     * thread is still on the active thread queue and will be scanned
     * on GC, so has to be left intact.  This should be revisited.
     */
    QUEUE_LOCK();
    while (UserThreadCount > 1) {
	QUEUE_WAIT();
    }

    sysExit(0);
}

void usage(void)
{
    fprintf(stderr, "usage: %s [-options] class\n", progname);
    fprintf(stderr, "\n");
    fprintf(stderr, "where options include:\n");
    fprintf(stderr, "    -help             print out this message\n");
    fprintf(stderr, "    -version          print out the build version\n");
#ifndef trace
    fprintf(stderr, "    -t                turn on instruction tracing\n");
    fprintf(stderr, "    -tm               turn on method tracing\n");
#endif
#ifdef MONITOR
    fprintf(stderr, "    -m                turn on monitoring\n");
#endif
    fprintf(stderr, "    -v -verbose       turn on verbose mode\n");
    fprintf(stderr, "    -debug            enable JAVA debugging\n");
    fprintf(stderr, "    -noasyncgc        don't allow asynchronous gc's\n");
    fprintf(stderr, "    -verbosegc        print a message when GCs occur\n");
    fprintf(stderr, "    -cs -checksource  check if source is newer when loading classes\n");
    fprintf(stderr, "    -ss<number>       set the C stack size of a process\n");
    fprintf(stderr, "    -oss<number>      set the JAVA stack size of a process\n");
    fprintf(stderr, "    -ms<number>       set the initial Java heap size\n");
    fprintf(stderr, "    -mx<number>       set the maximum Java heap size\n");
#ifdef LOGGING
    fprintf(stderr, "    -l<number>        set the logging level\n");
#endif
    fprintf(stderr, "    -classpath <directories separated by colons>\n");
    fprintf(stderr, "                      list directories in which to look for classes\n");
    fprintf(stderr, "    -prof             output profiling data to ./java.prof\n");
    fprintf(stderr, "    -verify           verify all classes when read in\n");
    fprintf(stderr, "    -verifyremote     verify classes read in over "
                                              "the network [default]\n");
    fprintf(stderr, "    -noverify         do not verify any class\n");
}

/* OpenCode for the interpreter.

   Tries to open a code file (fn).  That is, a .class file.  This
   returns -1 if the specified file can be opened but cannot be
   stat'd.  REMIND: When can that happen?  It returns -2 if the
   file cannot be opened.  Otherwise, it returns a valid file
   descriptor. If SkipSourceChecks is true, then we return at that
   point.

   Otherwise, we try to recompile the source.  If fn == 0, then we
   use sfn as the source file name to use.  sfn is specified by
   LoadFile by using the source hint associated with the import
   statement, or by the source file name as extracted from the
   previously recompiled version of this class.  That only happens
   if LoadFile already successfully loaded the class, and then
   noticed that the source is more recent than the .class file it
   opened.

   If sfn is 0 then a source file name is constructed by stripping
   the ".class" off of fn and replacing it with .java.  This often
   doesn't work, because .class files can be put in difference
   directories from their source files, but it's a good guess
   sometimes.  Then javac is forked to recompile the class, and
   the resulting .class file is opened.

   In summary, OpenCode for the interpreter returns a file
   descriptor for a compiler .class file.  If it can't open the
   .class file, it forks the compiler to compile some source so
   that the .class file will be there.  If after all that there is
   still no .class file, it gives up and returns -2.

   If st == 0, then doesn't even bother trying to open the file.
   Instead, it just does the source file thing.
*/

/* THE FOLLOWING CODE HAS BEEN COMMENTED OUT IN THE INTEREST OF COMPLETING
   THIS AS QUICKLY AS POSSIBLE...

static void
CompileFile(char *fn, char *dir)
{
    char *argv[8], **ap = argv;
    int r;

    *ap++ = JAVAC_EXE;
    if (verbose) {
	*ap++ = "-verbose";
    }
    if (dir) {
	*ap++ = "-d"; *ap++ = dir;
    }
    *ap++ = fn;
    *ap = NULL;

    if (verbose) {
	printf("[Executing");
	for (ap = argv; *ap != NULL; ap++) {
	    printf(" %s", *ap);
	}
	printf("]\n");
    }

    if ((r = spawnvp(P_WAIT, JAVAC_EXE, argv)) == -1) {
	char path[MAX_PATH];
	sprintf(path, "%s\\bin\\" JAVAC_EXE, java_home);
	if ((r = spawnv(P_WAIT, path, argv)) == -1) {
	    perror(argv[0]);
	    return FALSE;
	}
    }

    if (r != 0) {
	printf("%s: failed (%X)\n", argv[0], r);
	return FALSE;
    }

    if (verbose) {
	printf("[Finished %s]\n", argv[0]);
    }

    return TRUE;
}
*/

int
OpenCode(char *fn, char *sfn, char *dir, struct stat *st)
{
    long fd = -1;
    struct stat sst;

    if (fn != NULL) {
	/* First, try and open the class file */
	fd = open(fn, O_RDONLY | O_BINARY);
	if (fd >= 0 && st != NULL && stat(fn, st) < 0) {
	    close(fd);
	    fd = -1;
	}
    }

#if 0
/* THE FOLLOWING CODE HAS BEEN TEMPORARILY COMMENTED OUT IN THE INTEREST OF
   COMPLETING THIS AS QUICKLY AS POSSIBLE...

    if (SkipSourceChecks) {
	return fd < 0 ? -2 : fd;
    }

    if (sfn == NULL) {
	/* 
	 * Derive source file name from class file name by replacing the
	 * ".class" extension with ".java"
	 */
	char *cp;
	sysAssert(fn != NULL);
	strcpy(cbuf, fn);
	cp = cbuf + strlen(cbuf) - JAVAOBJEXTLEN - 2;
	if (cp < cbuf || stricmp(cp, "." JAVAOBJEXT) != 0) {
	    return fd < 0 ? -2 : fd;
	}
	strcpy(cp, "." JAVASRCEXT);
	sfn = strdup(cbuf);
    }

    if (stat(sfn, &sst) < 0) {
	return fd < 0 ? -2 : fd;    /* Source file does not exist */
    }

    if (fn != NULL && sst.st_mtime <= st->st_mtime) {
	return fd < 0 ? -2 : fd;    /* Class file is newer */
    }

    /* 
     *	Otherwise, source requires recompilation
     */
    CompileFile(sfn);

    if (fn != NULL) {
	/* Attempt to open generated class file */
	fd = open(fn, O_RDONLY | O_BINARY);
	if (fd >= 0 && static != NULL && stat(fn, st) < 0) {
	    close(fd);
	    fd = -1;
	}
    }

*/
#endif
    return fd < 0 ? -2 : fd;
}

void
ClassDefHook(ClassClass *cb)
{
}
