/*
 * @(#)condvar_md.c     1.11 95/11/18
 *
 * Copyright (c) 1995 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Win32 implementation of condition variables
 */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>

#include "condvar_md.h"

/*
 * Create and initialize condition variable.
 */
bool_t
condvarInit(condvar_t *condvar)
{
    HANDLE e = CreateEvent(NULL, TRUE, FALSE, NULL);
    if (e == 0) {
	return FALSE;
    }
    condvar->event = e;
    return TRUE;
}

/*
 * Delete the condition variable by closing the event object handle.
 */
void
condvarDestroy(condvar_t *condvar)
{
    CloseHandle(condvar->event);
}

/*
 * Block on a condition variable until signaled.
 */
void
condvarWait(condvar_t *condvar, mutex_t *mutex)
{
    unsigned int c = condvar->counter;

    condvar->waiters++;
    /*
     * Loop until condition variable is signaled. The event object is
      * set whenever the condition variable is signaled, and tickets will
     * reflect the number of threads which have been notified. The counter
     * field is used to make sure we don't respond to notifications that
     * have occurred *before* we started waiting, and is incremented each
     * time the condition variable is signaled.
     */
    do {
	LeaveCriticalSection(mutex);
	WaitForSingleObject(condvar->event, INFINITE);
	EnterCriticalSection(mutex);
    } while (condvar->tickets == 0 || condvar->counter == c);
    --condvar->waiters;
    /*
     * If this was the last thread to be notified, then we need to reset
     * the event object.
     */
    if (--condvar->tickets == 0) {
	ResetEvent(condvar->event);
    }
}

/*
 * Block on a condition variable until signaled or timeout has expired.
 */
void
condvarTimedWait(condvar_t *condvar, mutex_t *mutex, unsigned long millis)
{
    unsigned int c = condvar->counter;
    unsigned long t0, t1, elapsed;

    condvar->waiters++;
    while (TRUE) {
	LeaveCriticalSection(mutex);
	t0 = GetTickCount();
	WaitForSingleObject(condvar->event, millis);
	t1 = GetTickCount();
	EnterCriticalSection(mutex);
	if (condvar->tickets > 0 && condvar->counter != c) {
	    --condvar->waiters;
	    if (--condvar->tickets == 0) {
		ResetEvent(condvar->event);
	    }
	    return;
	}
	/* Need to check for wraparound of elapsed time */
	elapsed = t1 >= t0 ? t1 - t0 : ULONG_MAX - t1 + t0;
	if (elapsed >= millis) { /* timed out */
	    --condvar->waiters;
	    return;
	}
	millis -= elapsed;
    }
}

/*
 * Notify single thread waiting on a condition variable.
 */
bool_t
condvarSignal(condvar_t *condvar)
{
    if (condvar->waiters > condvar->tickets) {
	if (!SetEvent(condvar->event)) {
	    return FALSE;
	}
	condvar->tickets++;
	condvar->counter++;
    }
    return TRUE;
}

/*
 * Notify all threads waiting on a condition variable.
 */
bool_t
condvarBroadcast(condvar_t *condvar)
{
    if (condvar->waiters > 0) {
	if (!SetEvent(condvar->event)) {
	    return FALSE;
	}
	condvar->tickets = condvar->waiters;
	condvar->counter++;
    }
    return TRUE;
}
