/*
 * @(#)linker_md.c	1.17 95/11/20
 *
 * Copyright (c) 1995 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Maintains a list of currently loaded DLLs (Dynamic Link Libraries)
 * and their associated handles. Library names are case-insensitive.
 */

#include <windows.h>
#include <stdio.h>
#include <string.h>

#include "sys_api.h"
#include "path_md.h"

typedef struct {
    const char *fn;
    HANDLE h;
} dllentry_t;

static dllentry_t * dllarray;
static int	    dllarraysize;
static int	    dllcount;
static bool_t	    initialized;

/*
 * Build a machine dependent library name out of a path and file name.
 */
void
sysBuildLibName(char *holder, int holderlen, char *pname, char *fname)
{
    const int pnamelen = strlen(pname);

    /* Quietly truncates on buffer overflow. Should be an error. */
    if (pnamelen == 0) {
	jio_snprintf(holder, holderlen, "%s.dll", fname);
    } else {
	const char c = pname[pnamelen-1];
	if (c == ':' || c == '\\') {
	    jio_snprintf(holder, holderlen, "%s%s.dll", pname, fname);
	} else {
	    jio_snprintf(holder, holderlen, "%s\\%s.dll", pname, fname);
	}
    }
}

/*
 * Initialize the linker and return empty DLL search path. This will cause
 * the default Win32 search path to be used when a new DLL is loaded.
 */
char *
sysInitializeLinker(void)
{
    if (!initialized) {
	/* Make sure that runtime DLL is included first */
	if (!sysAddDLSegment("javai.dll")) {
	    fprintf(stderr, "*** Linker initialization failed ***\n");
	    exit(1);
	}
	initialized = TRUE;
    }
    return "";
}

/* 
 * Add a shared library to the list. DLL names are case-insensitive.
 */
bool_t
sysAddDLSegment(char *fname)
{
    HANDLE h;
    int i;

#ifdef DEBUG
    {
        char tmp[MAX_PATH+1];
        strcpy(tmp, fname);
        strcpy(tmp + strlen(tmp) - 4, "_g.dll");
        fname = strdup(tmp);
    }
#else
    fname = strdup(fname);
#endif

    if (fname == NULL) {
        out_of_memory();
    }

    for (i = 0; i < dllcount; i++) {
        if (stricmp(fname, dllarray[i].fn) == 0) {
	    return TRUE;    // dll is already loaded
	}
    }

    if ((h = LoadLibrary(fname)) == 0) {
	fprintf(stderr, "Unable to load dll '%s' (errcode = %x)\n", fname,
		GetLastError());
	return FALSE;
    }

    if (dllcount >= dllarraysize) {
	dllarraysize += 16;
	dllarray = realloc(dllarray, sizeof(dllentry_t) * dllarraysize);
	if (dllarray == 0) {
	    out_of_memory();
	}
    }

    dllarray[dllcount].fn = fname;
    dllarray[dllcount++].h = h;

    return TRUE;
}

/*
 * Look up symbol name in currently loaded DLLs and return it's handle.
 */
long
sysDynamicLink(char *sym)
{
    int i;

    /*
     * The following is necessary since the interpreter may invoke this
     * function before sysInitializeLinker() is called.
     */
    if (!initialized) {
	sysInitializeLinker();
    }

    for (i = 0; i < dllcount; i++) {
	FARPROC *f = (FARPROC *)GetProcAddress(dllarray[i].h, sym);
	if (f != NULL) {
	    return (long)f;
	}
    }

    return 0;
}
