/*
 * @(#)monitor_md.c	1.19 95/11/29
 *
 * Copyright (c) 1995 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Win32 implementation of Java monitors
 */

#include <windows.h>

#include "sys_api.h"
#include "threads.h"
#include "threads_md.h"
#include "monitor.h"
#include "monitor_md.h"

/*
 * Create and initialize monitor. This can be called before threads have
 * been initialized.
 */
int
sysMonitorInit(sys_mon_t *mid, bool_t in_cache)
{
    mid->owner = 0;
    mid->depth = 0;
    mutexInit(&mid->mutex);
    if (!condvarInit(&mid->condvar)) {
	mutexDestroy(&mid->mutex);
	return SYS_ERR;
    }
    return SYS_OK;
}

/*
 * Take ownership of monitor. This can be called before threads have
 * been initialized, in which case we do nothing since locked monitors
 * are not yet needed.
 */
int
sysMonitorEnter(sys_mon_t *mid)
{
    sys_thread_t *self = sysThreadSelf();

    if (self == 0) {
	return ThreadsInitialized ? SYS_ERR : SYS_OK;
    }
    sysAssert(ThreadsInitialized);
    self->state = MONITOR_WAIT;
    mutexLock(&mid->mutex);
    self->state = RUNNABLE;
    if (mid->owner == 0) {
        mid->owner = self;
    } else {
        sysAssert(mid->owner == self);
	sysAssert(mid->depth > 0);
	/* Don't lock mutex again if we already own it */
        mutexUnlock(&mid->mutex);
    }
    mid->depth++;
    return SYS_OK;
}

/*
 * Return TRUE if this thread currently owns the monitor. This can be
 * called before threads have been initialized, in which case we always
 * return TRUE.
 */
bool_t
sysMonitorEntered(sys_mon_t *mid)
{
    sys_thread_t *self = sysThreadSelf();
    return !ThreadsInitialized || self != 0 && mid->owner == self;
}

/*
 * Release ownership of monitor. This can be called before threads have
 * been initialized, in which case we do nothing as locked monitors are
 * not yet needed.
 */
int
sysMonitorExit(sys_mon_t *mid)
{
    sys_thread_t *self = sysThreadSelf();

    if (self == 0) {
	return ThreadsInitialized ? SYS_ERR : SYS_OK;
    }
    sysAssert(ThreadsInitialized);
    if (mid->owner != self) {
        return SYS_ERR;
    }
    sysAssert(mid->depth > 0);
    if (--mid->depth == 0) {
        mid->owner = 0;
	if (mid->condvar.waiters == 0) {
	    mutexUnlock(&mid->mutex);
	    return SYS_DESTROY;
	}
	mutexUnlock(&mid->mutex);
    }
    return SYS_OK;
}

/*
 * Notify single thread waiting on condition variable.
 */
int
sysMonitorNotify(sys_mon_t *mid)
{
    if (!sysMonitorEntered(mid)) {
	return SYS_ERR;
    }
    return condvarSignal(&mid->condvar) ? SYS_OK : SYS_ERR;
}

/*
 * Notify all threads waiting on condition variable.
 */
int
sysMonitorNotifyAll(sys_mon_t *mid)
{
    if (!sysMonitorEntered(mid)) {
	return SYS_ERR;
    }
    return condvarBroadcast(&mid->condvar) ? SYS_OK : SYS_ERR;
}

/*
 * Atomically drop mutex and wait for notification.
 */
int
sysMonitorWait(sys_mon_t *mid, int millis)
{
    sys_thread_t *self = sysThreadSelf();
    int depth = mid->depth;

    if (self == 0) {
        return SYS_ERR;
    }
    sysAssert(ThreadsInitialized);
    if (mid->owner != self) {
        return SYS_ERR;
    }
    mid->owner = 0;
    mid->depth = 0;
    self->state = CONDVAR_WAIT;
    if (millis == SYS_TIMEOUT_INFINITY) {
        condvarWait(&mid->condvar, &mid->mutex);
    } else {
        condvarTimedWait(&mid->condvar, &mid->mutex, millis);
    }
    self->state = RUNNABLE;
    mid->depth = depth;
    mid->owner = self;
    return SYS_OK;
}

/*
 * If monitor owned by thread then release monitor and notify all other
 * waiting threads. If no other threads are waiting for the monitor then
 * return SYS_DESTROY so that the monitor can be freed up from the cache.
 */
int
sysMonitorDestroy(sys_mon_t *mid, sys_thread_t *tid)
{
    if (mid->owner == tid) {
	mid->owner = 0;
	mid->depth = 0;
	if (mid->condvar.waiters > 0) {
	    condvarBroadcast(&mid->condvar);
	    mutexUnlock(&mid->mutex);
	    return SYS_OK;
	}
	mutexUnlock(&mid->mutex);
	return SYS_DESTROY;
    }
    return SYS_OK;
}

/*
 * Print monitor information.
 */
void
sysMonitorDumpInfo(sys_mon_t *mid)
{
    if (mid->owner != 0) {
        TID tid = mid->owner->cookie;
	fprintf(stderr, "%s\n",
	        Object2CString((JHandle *)THREAD(tid)->name));
    } else {
	fprintf(stderr, "<unowned>\n");
    }
    if (mid->condvar.waiters > 0) {
	fprintf(stderr, "        waiters = %d\n", mid->condvar.waiters);
    }
    if (mid->condvar.tickets > 0) {
	fprintf(stderr, "        tickets = %d\n", mid->condvar.tickets);
    }
}

/*
 * Return size of system-dependent monitor structure.
 */
int
sysMonitorSizeof(void)
{
    return sizeof(struct sys_mon);
}
