/*
 * @(#)threads_md.c	1.35 96/01/11
 *
 * Copyright (c) 1995 Sun Microsystems, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for NON-COMMERCIAL purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies. Please refer to the file "copyright.html"
 * for further important copyright and licensing information.
 *
 * SUN MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF
 * THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
 * TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE FOR
 * ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR
 * DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

/*
 * Win32 implementation of Java threads
 */

#include <windows.h>
#include <stdlib.h>

#include "sys_api.h"
#include "monitor.h"
#include "monitor_md.h"
#include "monitor_cache.h"
#include "threads.h"
#include "threads_md.h"

#include "java_lang_Thread.h"

/*
 * Set to TRUE once threads have been bootstrapped
 */
bool_t ThreadsInitialized = FALSE;

/*
 * Thread local storage index used for looking up sys_thread_t struct
 * (tid) associated with the current thread.
 */
#define TLS_INVALID_INDEX 0xffffffffUL
static unsigned long tls_index = TLS_INVALID_INDEX;

/*
 * Queue of active Java threads
 */
static sys_thread_t *ThreadQueue;

/*
 * Add thread to queue of active threads.
 */
static void
queueInsert(sys_thread_t *tid)
{
    QUEUE_LOCK();
    ActiveThreadCount++;
    tid->next = ThreadQueue;
    ThreadQueue = tid;
    if (!tid->system_thread) {
	UserThreadCount++;
    }
    QUEUE_UNLOCK();
}

/*
 * Remove thread from queue of active threads.
 */
static void
queueRemove(sys_thread_t *tid)
{
    QUEUE_LOCK();
    --ActiveThreadCount;
    if (!tid->system_thread) {
	--UserThreadCount;
/* 
 * Temporarily removing this fix as it introduces another nasty bug
 * that we need to investigate and fix for 1.0.1.   --dac 1/11/96
 * (scrivener mrm)
 */
/*	QUEUE_NOTIFY(); */
    }
    if (ThreadQueue == tid) {
	ThreadQueue = tid->next;
    } else {
	sys_thread_t *p;
	for (p = ThreadQueue; p->next != 0; p = p->next) {
	    if (p->next == tid) {
		p->next = tid->next;
		break;
	    }
	}
    }
    QUEUE_UNLOCK();
}

/* 
 * Bootstrap the Java thread system by making the current thread the
 * "primordial" thread.
 */
int
sysThreadBootstrap(sys_thread_t **ptid)
{
    HANDLE hnd = GetCurrentProcess();
    sys_thread_t *tid;

    /*
     * Allocate TLS index for thread-specific data.
     */
    tls_index = TlsAlloc();
    if (tls_index == TLS_INVALID_INDEX) {
        fprintf(stderr, "TlsAlloc failed (errcode = %x)\n", GetLastError());
	return SYS_INVALID;
    }
    
    /*
     * Allocate and initialize "thread private" structure.
     */
    if ((tid = malloc(sizeof(sys_thread_t))) == 0) {
        out_of_memory();
    }
    memset(tid, 0, sizeof(sys_thread_t));
    tid->stack_base = mainstktop;
    tid->state = RUNNABLE;
    tid->id = GetCurrentThreadId();
    DuplicateHandle(hnd, GetCurrentThread(), hnd, &tid->handle, 0, FALSE,
		    DUPLICATE_SAME_ACCESS);
    queueInsert(tid);
    TlsSetValue(tls_index, tid);

    ThreadsInitialized = TRUE;
    *ptid = tid;
    return SYS_OK;
}

/*
 * Return current stack pointer of specified thread.
 */
void *
sysThreadStackPointer(sys_thread_t *tid)
{
    CONTEXT context;

    /* REMIND: Need to fix this for Win95 */
    context.ContextFlags = CONTEXT_CONTROL;
    if (!GetThreadContext(tid->handle, &context)) {
        fprintf(stderr, "GetThreadContext failed (errcode = %x)\n",
		GetLastError());
	return 0;
    }
    return (void *)context.Esp;
}

/*
 * Return stack base of specified thread.
 */
void *
sysThreadStackBase(sys_thread_t *tid)
{
    return tid->stack_base;
}

/*
 * Thread start routine for new Java threads
 */
static unsigned __stdcall
_start(sys_thread_t *tid)
{
    DWORD regs[N_TRACED_REGS];

    /* Should thread suspend itself at this point? */

    tid->state = RUNNABLE;
    tid->stack_base = (void *)(regs + N_TRACED_REGS);
    tid->start_proc(tid->start_parm);
    return 0;
}

/* 
 * Create a new Java thread. The thread is initially suspended.
 */
int
sysThreadCreate(long stack_size, unsigned int flags, void *(*proc)(void *),
		sys_thread_t **tidp, void *parm)
{
    sys_thread_t *tid;

    /* 
     * Allocate and initialize thread structure.
     */
    if ((tid = malloc(sizeof(sys_thread_t))) == 0) {
	out_of_memory();
    }
    memset(tid, 0, sizeof(sys_thread_t));
    tid->state = SUSPENDED;
    tid->start_proc = proc;
    tid->start_parm = parm;
    if (flags == THR_SYSTEM) {
	tid->system_thread = TRUE;
    }

    /*
     * Start the new thread.
     */
    tid->handle = (HANDLE)_beginthreadex(NULL, 0, _start, tid,
					 CREATE_SUSPENDED, &tid->id);
    if (tid->handle == 0) {
	free(tid);
	return SYS_NORESOURCE;
    }
    queueInsert(tid);

    *tidp = tid;
    return SYS_OK;
}

/*
 * First thing called by a newly created Java thread.
 */
void
sysThreadInit(sys_thread_t *tid, stackp_t stack)
{
    /*
     * The setting of the stack base is now handled by the thread start
     * routine, so the 'stack' parameter is ignored here.
     */
    TlsSetValue(tls_index, tid);
}

/* 
 * Destroy calling thread (function never returns!).
 */
void
sysThreadExit(void)
{
    sys_thread_t *tid = sysThreadSelf();
    monitor_t *mon;
    TID Hthread ;
    ExecEnv *ee;

    CACHE_LOCK();
    REGISTRY_LOCK();

    queueRemove(tid);

    /*
     * Put the Java stack segments on the to-be-freed list.
     */
    Hthread = (TID)tid->cookie;
    ee = (ExecEnv *)THREAD(Hthread)->eetop;
    if (ee != 0) {  /* ee == 0 does happen... */
        DeleteExecEnv(ee, (JHandle *)Hthread);
    }

    /*
     * Notify anybody who is waiting for us to die.
     */
    mon = lookupMonitor(obj_monitor(tid->cookie));
    if (mon != 0) {
        sys_mon_t *mid = &sysmon(mon);
	sysMonitorEnter(mid);
	sysMonitorNotifyAll(mid);
	sysMonitorExit(mid);
    }

    /*
     * Clean up all monitors that we currently own.  monitorCleanup()
     * exits _moncache_lock but not _registry_lock
     */
    monitorCleanup(Hthread);
    REGISTRY_UNLOCK();

    THREAD(Hthread)->PrivateInfo = 0;

    /*
     * Close the thread handle and free the sys_thread_t structure.
     */
    CloseHandle(tid->handle);
    free(tid);
    _endthreadex(0);
    /* NOT REACHED */
}

/* 
 * Yield control to another thread.
 */
void
sysThreadYield(void)
{
    Sleep(0);
}

/* 
 * Suspend execution of the specified thread.
 */
int
sysThreadSuspend(sys_thread_t *tid)
{
    /* REMIND: Fix for Win95 */
    if (SuspendThread(tid->handle) == 0xffffffffUL) {
	return SYS_ERR;
    }
    tid->state = SUSPENDED;
    return SYS_OK;
}

/* 
 * Continue execution of the specified thread.
 */
int
sysThreadResume(sys_thread_t *tid)
{
    unsigned long n;

    /* Decrement thread's suspend count until no longer suspended */
    tid->state = RUNNABLE;
    while ((n = ResumeThread(tid->handle)) > 1) {
	if (n == 0xffffffffUL) {
	    return SYS_ERR;
	}
    }
    return SYS_OK;
}

/* 
 * Return priority of specified thread.
 */
int
sysThreadGetPriority(sys_thread_t *tid, int *pp)
{
    switch (GetThreadPriority(tid->handle)) {
    case THREAD_PRIORITY_IDLE:
	*pp = 0; break;
    case THREAD_PRIORITY_LOWEST:
	*pp = 2; break;
    case THREAD_PRIORITY_BELOW_NORMAL:
	*pp = 3; break;
    case THREAD_PRIORITY_NORMAL:
	*pp = 5; break;
    case THREAD_PRIORITY_ABOVE_NORMAL:
	*pp = 7; break;
    case THREAD_PRIORITY_HIGHEST:
	*pp = 8; break;
    case THREAD_PRIORITY_TIME_CRITICAL:
	*pp = 10; break;
    case THREAD_PRIORITY_ERROR_RETURN:
	return SYS_ERR;
    }
    return SYS_OK;
}

/*
 * Set priority of specified thread.
 */
int
sysThreadSetPriority(sys_thread_t *tid, int p)
{
    int priority;

    switch (p) {
    case 0:
	priority = THREAD_PRIORITY_IDLE;
	break;
    case 1: case 2:
	priority = THREAD_PRIORITY_LOWEST;
	break;
    case 3:
	priority = THREAD_PRIORITY_BELOW_NORMAL;
	break;
    case 4: case 5: case 6:
	priority = THREAD_PRIORITY_NORMAL;
	break;
    case 7:
	priority = THREAD_PRIORITY_ABOVE_NORMAL;
	break;
    case 8: case 9:
	priority = THREAD_PRIORITY_HIGHEST;
	break;
    case 10:
	priority = THREAD_PRIORITY_TIME_CRITICAL;
	break;
    default:
	return SYS_ERR;
    }
    return SetThreadPriority(tid->handle, priority) ? SYS_OK : SYS_ERR;
}

/* 
 * Return the thread information block of the calling thread.
 */
sys_thread_t *
sysThreadSelf(void)
{
    return tls_index == 0xffffffffUL ? 0 : TlsGetValue(tls_index);
}

/*
 * Enumerate over all threads in active queue calling a function for
 * each one.
 */
int
sysThreadEnumerateOver(int (*func)(sys_thread_t *, void *), void *arg)
{
    sys_thread_t *tid;
    int ret = SYS_OK;

    QUEUE_LOCK();
    for (tid = ThreadQueue; tid != 0; tid = tid->next) {
	if ((ret = (*func)(tid, arg)) != SYS_OK) {
	    break;
	}
    }
    QUEUE_UNLOCK();
    return ret;
}

/*
 * Helper function for sysThreadSingle()
 */
static int
threadSingleHelper(sys_thread_t *tid, void *self)
{
    return tid == self ? SYS_OK : sysThreadSuspend(tid);
}

/*
 * Puts each thread in the active thread queue to sleep except for the
 * calling thread. The threads must be later woken up with a corresponding
 * call to 'sysThreadMulti'. Returns SYS_OK on success, or SYS_ERR if any
 * of the threads could not be suspended.
 */
int
sysThreadSingle(void)
{
    return sysThreadEnumerateOver(threadSingleHelper, sysThreadSelf());
}

/*
 * Helper function for sysThreadMulti()
 */
static int
threadMultiHelper(sys_thread_t *tid, void *self)
{
    return tid == self ? SYS_OK : sysThreadResume(tid);
}

/*
 * Wakes up each thread in active thread queue except for the calling
 * thread. Returns SYS_ERR if not all threads could be woken up.
 */
void
sysThreadMulti(void)
{
    sysThreadEnumerateOver(threadMultiHelper, sysThreadSelf());
}

/*
 * Dump system-specific information about threads.
 */
void
sysThreadDumpInfo(sys_thread_t *tid)
{
    char *state;

    switch(tid->state) {
	case RUNNABLE: 
	    state = "RUNNABLE";
	    break;
	case SUSPENDED:
	    state = "SUSPENDED";
	    break;
	case SLEEPING:
	    state = "SLEEPING";
	    break;
	case MONITOR_WAIT:
	    state = "MONITOR_WAIT";
	    break;
	case CONDVAR_WAIT:
	    state = "CONDVAR_WAIT";
	    break;
	default:
	    state = "UNKNOWN";
    }
    fprintf(stderr, ", Win32ID:0x%x, state:%s", tid->id, state);
}

int
sysInterruptsPending(void)
{
    return FALSE;
}

void
sysThreadSetBackPtr(sys_thread_t *t, void *p)
{
    t->cookie = p;
}

void *
sysThreadGetBackPtr(sys_thread_t *t)
{
    return t->cookie;
}

void
sysThreadInitializeSystemThreads(void)
{
    extern void InitializeFinalizerThread(void);

    /* Create the finalization thread */
    InitializeFinalizerThread();
}

int
sysThreadCheckStack(void)
{
    return 1;
}

/*
 * The mechanics of actually signalling an exception (in the future,
 * and Alarm or Interrupt) depend upon what thread implementation you
 * are using.
 */
void
sysThreadPostException(sys_thread_t *tid, void *exc)
{
    struct execenv *ee;

    ee = (ExecEnv *) THREAD((TID)tid->cookie)->eetop;
    if (ee != 0) {
        exceptionThrow(ee, exc);
    }
}

unsigned sleep(unsigned seconds)
{
    Sleep(seconds * 1000);
    return 0;
}
